(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[     19660,        676]
NotebookOptionsPosition[     16205,        558]
NotebookOutlinePosition[     16618,        576]
CellTagsIndexPosition[     16575,        573]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["QuantumGroups`Algebra` package", "Subtitle",
 InitializationCell->True],

Cell["\<\
A subpackage for QuantumGroups v2.
Version 2.0, June 20, 2005, Scott Morrison\
\>", "Text",
 InitializationCell->True],

Cell[CellGroupData[{

Cell["Introduction", "Section",
 InitializationCell->True],

Cell["", "Text",
 InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implementation", "Section",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"BeginPackage", "[", 
  RowBox[{"\"\<QuantumGroups`Algebra`\>\"", ",", 
   RowBox[{"{", 
    RowBox[{
    "\"\<QuantumGroups`\>\"", ",", "\"\<QuantumGroups`RootSystems`\>\"", ",", 
     "\"\<QuantumGroups`WeylGroups`\>\""}], "}"}]}], "]"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"X", ";", "K", ";", "\[ScriptOne]", ";", "\[ScriptZero]", 
  ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4307651702196703`*^9, 3.4307651704600163`*^9}}],

Cell[BoxData[
 RowBox[{"SuperPlus", ";", "SuperMinus", ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"PositiveGenerators", ";", "NegativeGenerators", ";", 
  "CartanGenerators", ";", "Generators", ";", "NonCommutativePower", ";", 
  "\[CapitalDelta]", ";", "\[CapitalDelta]op", ";", "OperatorWeight", ";", 
  "OperatorLength", ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"Begin", "[", "\"\<`Private`\>\"", "]"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PositiveGenerators", "[", "\[CapitalGamma]_", "]"}], ":=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"SuperPlus", "[", 
      SubscriptBox["X", "i"], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "1", ",", 
       RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "}"}]}], "]"}]}], 
  ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4307652235763936`*^9, 3.430765228223075*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"NegativeGenerators", "[", "\[CapitalGamma]_", "]"}], ":=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"SuperMinus", "[", 
      SubscriptBox["X", "i"], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "1", ",", 
       RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "}"}]}], "]"}]}], 
  ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4307652301358256`*^9, 3.4307652346322913`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CartanGenerators", "[", "\[CapitalGamma]_", "]"}], ":=", 
   RowBox[{"Table", "[", 
    RowBox[{
     SubscriptBox["K", "i"], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "1", ",", 
       RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "}"}]}], "]"}]}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"Generators", "[", "\[CapitalGamma]_", "]"}], ":=", 
  RowBox[{
   RowBox[{"CartanGenerators", "[", "\[CapitalGamma]", "]"}], "~", "Join", 
   "~", 
   RowBox[{"PositiveGenerators", "[", "\[CapitalGamma]", "]"}], "~", "Join", 
   "~", 
   RowBox[{"NegativeGenerators", "[", "\[CapitalGamma]", "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[CellGroupData[{

Cell["Composition", "Subsubsection",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"Unprotect", "[", "NonCommutativeMultiply", "]"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"0", "**", "b_"}], ":=", "0"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"a_", "**", "0"}], ":=", "0"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"a_", "**", "b_Plus"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"a", "**", "#"}], "&"}], "/@", "b"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"a_Plus", "**", "b_"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"#", "**", "b"}], "&"}], "/@", "a"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"\[Alpha]_", "?", "qNumberQ"}], " ", "a_"}], ")"}], "**", "b_"}],
   ":=", 
  RowBox[{"\[Alpha]", 
   RowBox[{"(", 
    RowBox[{"a", "**", "b"}], ")"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"a_", "**", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"\[Alpha]_", "?", "qNumberQ"}], " ", "b_"}], ")"}]}], ":=", 
  RowBox[{"\[Alpha]", 
   RowBox[{"(", 
    RowBox[{"a", "**", "b"}], ")"}]}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[ScriptOne]", "**", "Z_"}], ":=", "Z"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Z_", "**", "\[ScriptOne]"}], ":=", "Z"}]}], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"a_", "\[CircleTimes]", "b_"}], ")"}], "**", 
   RowBox[{"(", 
    RowBox[{"c_", "\[CircleTimes]", "d_"}], ")"}]}], ":=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"a", "**", "c"}], ")"}], "\[CircleTimes]", 
   RowBox[{"(", 
    RowBox[{"b", "**", "d"}], ")"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"NonCommutativePower", "[", 
   RowBox[{"x_", ",", 
    RowBox[{"n_", "?", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"#", "\[GreaterEqual]", "2"}], "&"}], ")"}]}]}], "]"}], ":=", 
  RowBox[{"NonCommutativeMultiply", "@@", 
   RowBox[{"Table", "[", 
    RowBox[{"x", ",", 
     RowBox[{"{", "n", "}"}]}], "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"NonCommutativePower", "[", 
   RowBox[{"x_", ",", "1"}], "]"}], ":=", "x"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"NonCommutativePower", "[", 
   RowBox[{"x_", ",", "0"}], "]"}], ":=", "\[ScriptOne]"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"NonCommutativePower", "[", 
   RowBox[{"x_", ",", 
    RowBox[{"n_", "?", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"#", "<", "0"}], "&"}], ")"}]}]}], "]"}], ":=", 
  RowBox[{"NonCommutativePower", "[", 
   RowBox[{
    SuperscriptBox["x", 
     RowBox[{"-", "1"}]], ",", 
    RowBox[{"-", "n"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"Protect", "[", "NonCommutativeMultiply", "]"}], ";"}]], "Input",
 InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell["Comultiplication", "Subsubsection",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"\[CapitalDelta]", "[", "\[ScriptOne]", "]"}], "=", 
   RowBox[{"\[ScriptOne]", "\[CircleTimes]", "\[ScriptOne]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", 
   SubscriptBox["K", "i_"], "]"}], ":=", 
  RowBox[{
   SubscriptBox["K", "i"], "\[CircleTimes]", 
   SubscriptBox["K", "i"]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", 
   SuperscriptBox[
    SubscriptBox["K", "i_"], 
    RowBox[{"-", "1"}]], "]"}], ":=", 
  RowBox[{
   SuperscriptBox[
    SubscriptBox["K", "i"], 
    RowBox[{"-", "1"}]], "\[CircleTimes]", 
   SuperscriptBox[
    SubscriptBox["K", "i"], 
    RowBox[{"-", "1"}]]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", 
   RowBox[{"SuperPlus", "[", 
    SubscriptBox["X", "i_"], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"SuperPlus", "[", 
     SubscriptBox["X", "i"], "]"}], "\[CircleTimes]", 
    SubscriptBox["K", "i"]}], "+", 
   RowBox[{"\[ScriptOne]", "\[CircleTimes]", 
    RowBox[{"SuperPlus", "[", 
     SubscriptBox["X", "i"], "]"}]}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", 
   RowBox[{"SuperMinus", "[", 
    SubscriptBox["X", "i_"], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"SuperMinus", "[", 
     SubscriptBox["X", "i"], "]"}], "\[CircleTimes]", "\[ScriptOne]"}], "+", 
   RowBox[{
    SuperscriptBox[
     SubscriptBox["K", "i"], 
     RowBox[{"-", "1"}]], "\[CircleTimes]", 
    RowBox[{"SuperMinus", "[", 
     SubscriptBox["X", "i"], "]"}]}]}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4307652424435234`*^9, 3.430765282771512*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", 
   RowBox[{
    RowBox[{"\[Alpha]_", "?", "qNumberQ"}], " ", "A_"}], "]"}], ":=", 
  RowBox[{"\[Alpha]", " ", 
   RowBox[{"\[CapitalDelta]", "[", "A", "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", 
   RowBox[{"A_", "**", "B_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"\[CapitalDelta]", "[", "A", "]"}], "**", 
   RowBox[{"\[CapitalDelta]", "[", "B", "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[CapitalDelta]op", "[", "Z_", "]"}], ":=", 
  RowBox[{
   RowBox[{"\[CapitalDelta]", "[", "Z", "]"}], "/.", 
   RowBox[{
    RowBox[{"X_", "\[CircleTimes]", "Y_"}], "\[RuleDelayed]", 
    RowBox[{"Y", "\[CircleTimes]", "X"}]}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", "X_Plus", "]"}], ":=", 
  RowBox[{"\[CapitalDelta]", "/@", "X"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CapitalDelta]", "[", "X_CircleTimes", "]"}], ":=", 
  RowBox[{
   RowBox[{"\[CapitalDelta]", "[", 
    RowBox[{"First", "[", "X", "]"}], "]"}], "\[CircleTimes]", 
   RowBox[{"Rest", "[", "X", "]"}]}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.630698091406953*^9, 3.630698096134737*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Weights of quantum group generators", "Subsubsection",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   "\[ScriptZero]", "]"}], ":=", 
  RowBox[{"ZeroVector", "[", 
   RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   "\[ScriptOne]", "]"}], ":=", 
  RowBox[{"ZeroVector", "[", 
   RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   SubscriptBox["K", "i_"], "]"}], ":=", 
  RowBox[{"ZeroVector", "[", 
   RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   SuperscriptBox[
    SubscriptBox["K", "i_"], 
    RowBox[{"-", "1"}]], "]"}], ":=", 
  RowBox[{"ZeroVector", "[", 
   RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}], "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"SuperPlus", "[", 
    SubscriptBox["X", "i_"], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"SuperPlus", "[", 
     SubscriptBox["X", "i"], "]"}], "]"}], "=", 
   RowBox[{
    RowBox[{"SimpleRoots", "[", "\[CapitalGamma]", "]"}], 
    "\[LeftDoubleBracket]", "i", 
    "\[RightDoubleBracket]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"SuperMinus", "[", 
    SubscriptBox["X", "i_"], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"SuperMinus", "[", 
     SubscriptBox["X", "i"], "]"}], "]"}], "=", 
   RowBox[{"-", 
    RowBox[{
     RowBox[{"SimpleRoots", "[", "\[CapitalGamma]", "]"}], 
     "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.430765289941822*^9, 3.430765327976514*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"A_", "**", "B_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", "]"}], 
   "+", 
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "B", 
    "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"A_", "\[CircleTimes]", "B_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", "]"}], 
   "+", 
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "B", 
    "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
    RowBox[{
     RowBox[{"\[Alpha]_", "?", "qNumberQ"}], " ", "A_"}], "]"}], ":=", 
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
    "]"}]}], "\[IndentingNewLine]", 
  RowBox[{"(*", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", 
      RowBox[{"A_", "+", "B_"}], "]"}], "/;", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
        "]"}], "\[Equal]", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "B", 
        "]"}]}], ")"}]}], ":=", 
    RowBox[{
     RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
     "]"}]}], "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"OperatorWeight", "[", "\[CapitalGamma]_", "]"}], "[", "A_Plus", 
    "]"}], "/;", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"Length", "[", 
      RowBox[{"Union", "[", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "/@", 
        RowBox[{"(", 
         RowBox[{"List", "@@", "A"}], ")"}]}], "]"}], "]"}], "\[Equal]", 
     "1"}], ")"}]}], ":=", 
  RowBox[{
   RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
   RowBox[{"First", "[", "A", "]"}], "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"OperatorLength", "[", 
   RowBox[{"\[CapitalGamma]_", ",", "A_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"OperatorLength", "[", 
    RowBox[{"\[CapitalGamma]", ",", "A"}], "]"}], "=", 
   RowBox[{"With", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"\[Lambda]", "=", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
        "]"}]}], "}"}], ",", 
     SqrtBox[
      RowBox[{
       RowBox[{"KillingForm", "[", "\[CapitalGamma]", "]"}], "[", 
       RowBox[{"\[Lambda]", ",", "\[Lambda]"}], "]"}]]}], "]"}]}]}]], "Input",
 InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 SubscriptBox["K", "\[Rho]"]], "Subsubsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   SubscriptBox["K", 
    RowBox[{
     SubscriptBox["A", "1"], ",", "\[Rho]"}]], "=", 
   SubscriptBox["K", "1"]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  SubscriptBox["K", 
   RowBox[{"\[CapitalGamma]_", ",", "\[Rho]"}]], ":=", 
  RowBox[{"Inner", "[", 
   RowBox[{"NonCommutativePower", ",", 
    RowBox[{"CartanGenerators", "[", "\[CapitalGamma]", "]"}], ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"Plus", "@@", 
       RowBox[{"PositiveRoots", "[", "\[CapitalGamma]", "]"}]}], ")"}], ".", 
     RowBox[{"Inverse", "[", 
      RowBox[{"SimpleRoots", "[", "\[CapitalGamma]", "]"}], "]"}]}], ",", 
    "NonCommutativeMultiply"}], "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"End", "[", "]"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"EndPackage", "[", "]"}], ";"}]], "Input",
 InitializationCell->True]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
AutoGeneratedPackage->Automatic,
WindowSize->{832, 558},
WindowMargins->{{Automatic, 46}, {Automatic, -63}},
ShowSelection->True,
FrontEndVersion->"10.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (September 9, \
2014)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 77, 1, 51, "Subtitle",
 InitializationCell->True],
Cell[647, 25, 128, 4, 49, "Text",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[800, 33, 58, 1, 64, "Section",
 InitializationCell->True],
Cell[861, 36, 43, 1, 30, "Text",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[941, 42, 60, 1, 64, "Section",
 InitializationCell->True],
Cell[1004, 45, 300, 7, 46, "Input",
 InitializationCell->True],
Cell[1307, 54, 197, 4, 28, "Input",
 InitializationCell->True],
Cell[1507, 60, 98, 2, 28, "Input",
 InitializationCell->True],
Cell[1608, 64, 292, 5, 63, "Input",
 InitializationCell->True],
Cell[1903, 71, 119, 3, 28, "Input",
 InitializationCell->True],
Cell[2025, 76, 463, 13, 28, "Input",
 InitializationCell->True],
Cell[2491, 91, 466, 13, 28, "Input",
 InitializationCell->True],
Cell[2960, 106, 353, 11, 31, "Input",
 InitializationCell->True],
Cell[3316, 119, 374, 9, 31, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[3715, 132, 63, 1, 28, "Subsubsection",
 InitializationCell->True],
Cell[3781, 135, 128, 3, 31, "Input",
 InitializationCell->True],
Cell[3912, 140, 102, 3, 31, "Input",
 InitializationCell->True],
Cell[4017, 145, 102, 3, 31, "Input",
 InitializationCell->True],
Cell[4122, 150, 176, 6, 31, "Input",
 InitializationCell->True],
Cell[4301, 158, 176, 6, 31, "Input",
 InitializationCell->True],
Cell[4480, 166, 515, 18, 46, "Input",
 InitializationCell->True],
Cell[4998, 186, 202, 5, 46, "Input",
 InitializationCell->True],
Cell[5203, 193, 365, 12, 28, "Input",
 InitializationCell->True],
Cell[5571, 207, 393, 12, 28, "Input",
 InitializationCell->True],
Cell[5967, 221, 148, 4, 28, "Input",
 InitializationCell->True],
Cell[6118, 227, 159, 4, 28, "Input",
 InitializationCell->True],
Cell[6280, 233, 387, 13, 35, "Input",
 InitializationCell->True],
Cell[6670, 248, 126, 3, 28, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[6833, 256, 68, 1, 35, "Subsubsection",
 InitializationCell->True],
Cell[6904, 259, 1693, 51, 116, "Input",
 InitializationCell->True],
Cell[8600, 312, 257, 7, 28, "Input",
 InitializationCell->True],
Cell[8860, 321, 256, 7, 28, "Input",
 InitializationCell->True],
Cell[9119, 330, 304, 8, 28, "Input",
 InitializationCell->True],
Cell[9426, 340, 480, 11, 46, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[9943, 356, 87, 1, 35, "Subsubsection",
 InitializationCell->True],
Cell[10033, 359, 1041, 30, 90, "Input",
 InitializationCell->True],
Cell[11077, 391, 1117, 30, 73, "Input",
 InitializationCell->True],
Cell[12197, 423, 393, 12, 28, "Input",
 InitializationCell->True],
Cell[12593, 437, 405, 12, 28, "Input",
 InitializationCell->True],
Cell[13001, 451, 1497, 44, 97, "Input",
 InitializationCell->True],
Cell[14501, 497, 631, 18, 59, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[15169, 520, 61, 1, 34, "Subsubsection"],
Cell[15233, 523, 727, 20, 68, "Input",
 InitializationCell->True],
Cell[15963, 545, 96, 3, 31, "Input",
 InitializationCell->True],
Cell[16062, 550, 103, 3, 31, "Input",
 InitializationCell->True]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
