(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    210346,       5151]
NotebookOptionsPosition[     89141,       2590]
NotebookOutlinePosition[    204676,       4976]
CellTagsIndexPosition[    204633,       4973]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["MatrixPresentations package", "Subtitle",
 InitializationCell->True],

Cell["\<\
A subpackage for QuantumGroups v2.
Version 2.0, July 30, 2005, Scott Morrison\
\>", "Text",
 InitializationCell->True],

Cell[CellGroupData[{

Cell["Introduction", "Section",
 InitializationCell->True],

Cell["\<\
This package gives explicit matrices for the generators of a quantum group, \
in an arbitrary representation.\
\>", "Text",
 InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implementation", "Section",
 InitializationCell->True],

Cell[CellGroupData[{

Cell["Start of package", "Subsection"],

Cell["Specify package dependencies:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"BeginPackage", "[", 
   RowBox[{"\"\<QuantumGroups`MatrixPresentations`\>\"", ",", 
    RowBox[{"{", 
     RowBox[{
     "\"\<QuantumGroups`\>\"", ",", 
      "\"\<QuantumGroups`Utilities`IntersectSubspaces`\>\"", ",", 
      "\"\<QuantumGroups`Utilities`MatrixWrapper`\>\"", ",", 
      "\"\<QuantumGroups`Utilities`Debugging`\>\"", ",", 
      "\"\<QuantumGroups`Utilities`DataPackage`\>\"", ",", 
      "\"\<QuantumGroups`RootSystems`\>\"", ",", 
      "\"\<QuantumGroups`Algebra`\>\"", ",", 
      "\"\<QuantumGroups`Representations`\>\""}], "}"}]}], "]"}], 
  ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{3.430404551120432*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Usage messages", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"MatrixPresentation", "::", "usage"}], "=", "\"\<\>\""}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"HighWeightVectors", "::", "usage"}], "=", "\"\<\>\""}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"HighWeightVectorQ", ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"HighWeights", "::", "usage"}], "=", "\"\<\>\""}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"AllHighWeightVectors", "::", "usage"}], "=", "\"\<\>\""}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"TensorProductWeightSpaceInclusion", ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"ChangeOfBasisMatrix", ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"ShortRootBasis", ";"}]], "Input",
 InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell["Internals", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Begin", "[", "\"\<`Private`\>\"", "]"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"q", "=", "Global`q"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"(*", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", "A_", 
     "]"}], "[", 
    RowBox[{
     RowBox[{"CircleTimes", "[", "V_", "]"}], ",", "\[Beta]_", ",", 
     "\[Lambda]_"}], "]"}], ":=", 
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A", 
     "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}]}], "*)"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.630697696717904*^9, 3.630697697740766*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A_", "**", "B_"}], "]"}], "[", 
    RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{"!", 
    RowBox[{"MemberQ", "[", 
     RowBox[{
      RowBox[{"Weights", "[", 
       RowBox[{"\[CapitalGamma]", ",", "V"}], "]"}], ",", 
      RowBox[{"\[Lambda]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
        RowBox[{"A", "**", "B"}], "]"}]}]}], "]"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    RowBox[{"A_", "**", "B_"}], "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A", 
     "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", 
     RowBox[{"\[Lambda]", "+", 
      RowBox[{
       RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "B", 
       "]"}]}]}], "]"}], ".", 
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "B", 
     "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    "\[ScriptZero]", "]"}], "[", 
   RowBox[{"V_", ",", "_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"m", "=", 
      RowBox[{"WeightMultiplicity", "[", 
       RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}]}], 
     "}"}], ",", 
    RowBox[{"ZeroesMatrix", "[", 
     RowBox[{"m", ",", "m"}], "]"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    "\[ScriptOne]", "]"}], "[", 
   RowBox[{"V_", ",", "_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"identityMatrix", "[", 
   RowBox[{"WeightMultiplicity", "[", 
    RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    RowBox[{
     RowBox[{"\[Alpha]_", "?", "qNumberQ"}], " ", "A_"}], "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"\[Alpha]", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A", 
     "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     "A_Plus", "]"}], "[", 
    RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"Length", "[", 
      RowBox[{"Union", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "#", 
          "]"}], "&"}], "/@", 
        RowBox[{"(", 
         RowBox[{"List", "@@", "A"}], ")"}]}], "]"}], "]"}], "\[Equal]", 
     "1"}], ")"}]}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "#", 
      "]"}], "[", 
     RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}], "&"}], "/@", 
   "A"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    SubscriptBox["K", "i_"], "]"}], 
   RowBox[{"[", 
    RowBox[{"V_", ",", "_", ",", "\[Lambda]_"}], "]"}]}], ":=", 
  RowBox[{
   SuperscriptBox["q", 
    RowBox[{
     RowBox[{
      RowBox[{"CartanFactors", "[", "\[CapitalGamma]", "]"}], 
      "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
     RowBox[{
     "\[Lambda]", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]], 
   RowBox[{"identityMatrix", "[", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    SuperscriptBox[
     SubscriptBox["K", "i_"], 
     RowBox[{"-", "1"}]], "]"}], "[", 
   RowBox[{"V_", ",", "_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   SuperscriptBox["q", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       RowBox[{"CartanFactors", "[", "\[CapitalGamma]", "]"}], 
       "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], 
     RowBox[{
     "\[Lambda]", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]], 
   RowBox[{"identityMatrix", "[", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}], 
    "]"}]}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.409502623477992*^9, 3.4095026340131407`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A", ":", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"SuperPlus", "[", 
         SubscriptBox["X", "_"], "]"}], "|", 
        RowBox[{"SuperMinus", "[", 
         SubscriptBox["X", "_"], "]"}]}], ")"}]}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Mu]_", "]"}], 
     ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"ZeroVectorQ", "[", "\[Mu]", "]"}], "\[And]", 
     RowBox[{
      RowBox[{"Length", "[", "\[Mu]", "]"}], "\[Equal]", 
      RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "\[And]", 
     RowBox[{
      RowBox[{"Length", "[", "\[Lambda]", "]"}], "\[Equal]", 
      RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}]}], ")"}]}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{"ZeroVectorQ", "[", "\[Lambda]", "]"}], ",", 
    RowBox[{"ZeroesMatrix", "[", 
     RowBox[{"0", ",", "1"}], "]"}], ",", 
    RowBox[{"ZeroesMatrix", "[", 
     RowBox[{
      RowBox[{"WeightMultiplicity", "[", 
       RowBox[{"\[CapitalGamma]", ",", 
        RowBox[{
         RowBox[{"Irrep", "[", "\[CapitalGamma]", "]"}], "[", "\[Mu]", "]"}], 
        ",", 
        RowBox[{"\[Lambda]", "+", 
         RowBox[{
          RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
          "]"}]}]}], "]"}], ",", "0"}], "]"}]}], "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4095018173388205`*^9, 3.40950183395271*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A", ":", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"SuperPlus", "[", 
         SubscriptBox["X", "_"], "]"}], "|", 
        RowBox[{"SuperMinus", "[", 
         SubscriptBox["X", "_"], "]"}]}], ")"}]}], "]"}], "[", 
    RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{"MinusculeRepresentationQ", "[", 
    RowBox[{"\[CapitalGamma]", ",", "V"}], "]"}]}], ":=", 
  RowBox[{"OnesMatrix", "[", 
   RowBox[{
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", "V", ",", 
      RowBox[{"\[Lambda]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
        "]"}]}]}], "]"}], ",", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.409501838318989*^9, 3.409501849495059*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A_", "\[CircleTimes]", "B_"}], "]"}], "[", 
    RowBox[{
     RowBox[{"V_", "\[CircleTimes]", "W_"}], ",", "\[Beta]_", ",", 
     "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{"!", 
    RowBox[{"MemberQ", "[", 
     RowBox[{
      RowBox[{"Weights", "[", 
       RowBox[{"\[CapitalGamma]", ",", 
        RowBox[{"V", "\[CircleTimes]", "W"}]}], "]"}], ",", 
      RowBox[{"\[Lambda]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
        RowBox[{"A", "\[CircleTimes]", "B"}], "]"}]}]}], "]"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", 
      RowBox[{"V", "\[CircleTimes]", "W"}], ",", "\[Lambda]"}], "]"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A_", "\[CircleTimes]", "B_"}], "]"}], "[", 
    RowBox[{
     RowBox[{"V_", "\[CircleTimes]", "W_"}], ",", "\[Beta]_", ",", 
     "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{
    RowBox[{"OperatorLength", "[", 
     RowBox[{"\[CapitalGamma]", ",", "A"}], "]"}], ">", 
    RowBox[{"WeightDiameter", "[", 
     RowBox[{"\[CapitalGamma]", ",", "V"}], "]"}]}]}], ":=", 
  RowBox[{"ZeroesMatrix", "[", 
   RowBox[{
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", 
      RowBox[{"V", "\[CircleTimes]", "W"}], ",", 
      RowBox[{"\[Lambda]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
        RowBox[{"A", "\[CircleTimes]", "B"}], "]"}]}]}], "]"}], ",", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", 
      RowBox[{"V", "\[CircleTimes]", "W"}], ",", "\[Lambda]"}], "]"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A_", "\[CircleTimes]", "B_"}], "]"}], "[", 
    RowBox[{
     RowBox[{"V_", "\[CircleTimes]", "W_"}], ",", "\[Beta]_", ",", 
     "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{
    RowBox[{"OperatorLength", "[", 
     RowBox[{"\[CapitalGamma]", ",", "B"}], "]"}], ">", 
    RowBox[{"WeightDiameter", "[", 
     RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}]}]}], ":=", 
  RowBox[{"ZeroesMatrix", "[", 
   RowBox[{
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", 
      RowBox[{"V", "\[CircleTimes]", "W"}], ",", 
      RowBox[{"\[Lambda]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
        RowBox[{"A", "\[CircleTimes]", "B"}], "]"}]}]}], "]"}], ",", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", 
      RowBox[{"V", "\[CircleTimes]", "W"}], ",", "\[Lambda]"}], "]"}]}], 
   "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
    RowBox[{"A_", "\[CircleTimes]", "B_"}], "]"}], "[", 
   RowBox[{
    RowBox[{"V_", "\[CircleTimes]", "W_"}], ",", "\[Beta]_", ",", 
    "\[Lambda]_"}], "]"}], ":=", "\[IndentingNewLine]", 
  RowBox[{"AppendRows", "@@", 
   RowBox[{"Table", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"TensorProductWeightSpaceInclusion", "[", 
       RowBox[{"\[CapitalGamma]", ",", 
        RowBox[{"{", 
         RowBox[{"V", ",", "W"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"\[Lambda]", "+", 
           RowBox[{
            RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A",
             "]"}], "-", 
           RowBox[{
            RowBox[{"Weights", "[", 
             RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}], 
            "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], ",", 
          RowBox[{
           RowBox[{
            RowBox[{"Weights", "[", 
             RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}], 
            "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "+", 
           RowBox[{
            RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "B",
             "]"}]}]}], "}"}]}], "]"}], ".", 
      RowBox[{"Together", "[", 
       RowBox[{"MatrixKroneckerProduct", "[", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
           "A", "]"}], "[", 
          RowBox[{"V", ",", "\[Beta]", ",", 
           RowBox[{"\[Lambda]", "-", 
            RowBox[{
             RowBox[{"Weights", "[", 
              RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}], 
             "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}], 
          "]"}], ",", 
         RowBox[{
          RowBox[{
           RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
           "B", "]"}], "[", 
          RowBox[{"W", ",", "\[Beta]", ",", 
           RowBox[{
            RowBox[{"Weights", "[", 
             RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}], 
            "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], "]"}]}],
         "]"}], "]"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "1", ",", 
       RowBox[{"Length", "[", 
        RowBox[{"Weights", "[", 
         RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}], "]"}]}], "}"}]}], 
    "]"}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{3.409501695273299*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", "A_", 
     "]"}], "[", 
    RowBox[{
     RowBox[{"V_", "\[CircleTimes]", "W_"}], ",", "\[Beta]_", ",", 
     "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{"!", 
    RowBox[{"MemberQ", "[", 
     RowBox[{
      RowBox[{"Weights", "[", 
       RowBox[{"\[CapitalGamma]", ",", 
        RowBox[{"V", "\[CircleTimes]", "W"}]}], "]"}], ",", 
      RowBox[{"\[Lambda]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
        RowBox[{"\[CapitalDelta]", "[", "A", "]"}], "]"}]}]}], "]"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", 
    RowBox[{"WeightMultiplicity", "[", 
     RowBox[{"\[CapitalGamma]", ",", 
      RowBox[{"V", "\[CircleTimes]", "W"}], ",", "\[Lambda]"}], "]"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", "A_", 
    "]"}], "[", 
   RowBox[{
    RowBox[{"V_", "\[CircleTimes]", "W_"}], ",", "\[Beta]_", ",", 
    "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"\[CapitalDelta]", "[", "A", "]"}], "]"}], "[", 
   RowBox[{
    RowBox[{"V", "\[CircleTimes]", "W"}], ",", "\[Beta]", ",", "\[Lambda]"}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell["\<\
Watch out-due to the crappiness of my implementation of BlockDiagonalMatrix, \
this will puke with more than 256 direct summands.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", "A_", 
    "]"}], "[", 
   RowBox[{"V_DirectSum", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"BlockDiagonalMatrix", "@@", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A",
         "]"}], "[", 
       RowBox[{"#", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}], "&"}], "/@", 
     RowBox[{"(", 
      RowBox[{"List", "@@", "V"}], ")"}]}], ")"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"PadWithZeroRows", "[", 
   RowBox[{"m_Matrix", ",", 
    RowBox[{"initial_", "?", "NaturalQ"}], ",", 
    RowBox[{"total_", "?", "NaturalQ"}]}], "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"rows", " ", "=", " ", 
       RowBox[{
        RowBox[{"Dimensions", "[", "m", "]"}], "\[LeftDoubleBracket]", "1", 
        "\[RightDoubleBracket]"}]}], ",", 
      RowBox[{"cols", "=", 
       RowBox[{
        RowBox[{"Dimensions", "[", "m", "]"}], "\[LeftDoubleBracket]", "2", 
        "\[RightDoubleBracket]"}]}]}], "}"}], ",", 
    RowBox[{"Matrix", "[", 
     RowBox[{"total", ",", "cols", ",", 
      RowBox[{"Join", "[", 
       RowBox[{
        RowBox[{"ConstantArray", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{"initial", ",", "cols"}], "}"}]}], "]"}], ",", 
        RowBox[{"MatrixData", "[", "m", "]"}], ",", 
        RowBox[{"ConstantArray", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"total", "-", "initial", "-", "rows"}], ",", "cols"}], 
           "}"}]}], "]"}]}], "]"}]}], "]"}]}], "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4312431590618114`*^9, 3.431243173402432*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"WeightMultiplicityComponents", "[", 
   RowBox[{"\[CapitalGamma]_", ",", "V1_", ",", "V2_", ",", "\[Lambda]_"}], 
   "]"}], ":=", 
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"WeightMultiplicity", "[", 
      RowBox[{"\[CapitalGamma]", ",", "V1", ",", 
       RowBox[{"\[Lambda]", "-", 
        RowBox[{
         RowBox[{"Weights", "[", 
          RowBox[{"\[CapitalGamma]", ",", "V2"}], "]"}], 
         "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}], "]"}], 
     RowBox[{"WeightMultiplicity", "[", 
      RowBox[{"\[CapitalGamma]", ",", "V2", ",", 
       RowBox[{
        RowBox[{"Weights", "[", 
         RowBox[{"\[CapitalGamma]", ",", "V2"}], "]"}], 
        "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], "]"}]}], 
    ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "1", ",", 
      RowBox[{"Length", "[", 
       RowBox[{"Weights", "[", 
        RowBox[{"\[CapitalGamma]", ",", "V2"}], "]"}], "]"}]}], "}"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"WeightMultiplicityPartialSums", "[", 
   RowBox[{"\[CapitalGamma]_", ",", "V1_", ",", "V2_", ",", "\[Lambda]_"}], 
   "]"}], ":=", 
  RowBox[{
   RowBox[{"WeightMultiplicityPartialSums", "[", 
    RowBox[{"\[CapitalGamma]", ",", "V1", ",", "V2", ",", "\[Lambda]"}], 
    "]"}], "=", 
   RowBox[{"Drop", "[", 
    RowBox[{
     RowBox[{"FoldList", "[", 
      RowBox[{"Plus", ",", "0", ",", 
       RowBox[{"WeightMultiplicityComponents", "[", 
        RowBox[{"\[CapitalGamma]", ",", "V1", ",", "V2", ",", "\[Lambda]"}], 
        "]"}]}], "]"}], ",", 
     RowBox[{"-", "1"}]}], "]"}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4468361949740953`*^9, 3.446836207081067*^9}}],

Cell[TextData[{
 "Gives the map from ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["V", "\[Lambda]"], "\[CircleTimes]", 
    SubscriptBox["W", "\[Mu]"]}], TraditionalForm]]],
 " into ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    RowBox[{"(", 
     RowBox[{"V", "\[CircleTimes]", "W"}], ")"}], 
    RowBox[{"\[Lambda]", "+", "\[Mu]"}]], TraditionalForm]]],
 "."
}], "Text",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"TensorProductWeightSpaceInclusion", "[", 
   RowBox[{"\[CapitalGamma]_", ",", 
    RowBox[{"{", 
     RowBox[{"V_", ",", "W_"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"\[Lambda]_", ",", "\[Mu]_"}], "}"}]}], "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"m0", "=", 
       RowBox[{
        RowBox[{"WeightMultiplicity", "[", 
         RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}], 
        RowBox[{"WeightMultiplicity", "[", 
         RowBox[{"\[CapitalGamma]", ",", "W", ",", "\[Mu]"}], "]"}]}]}], ",", 
      RowBox[{"m1", "=", 
       RowBox[{"WeightMultiplicity", "[", 
        RowBox[{"\[CapitalGamma]", ",", 
         RowBox[{"V", "\[CircleTimes]", "W"}], ",", 
         RowBox[{"\[Lambda]", "+", "\[Mu]"}]}], "]"}]}], ",", 
      RowBox[{"pos", "=", 
       RowBox[{"Position", "[", 
        RowBox[{
         RowBox[{"Weights", "[", 
          RowBox[{"\[CapitalGamma]", ",", "W"}], "]"}], ",", "\[Mu]"}], 
        "]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"pos", "\[Equal]", 
        RowBox[{"{", "}"}]}], "\[Or]", 
       RowBox[{"m0", "\[Equal]", "0"}]}], ",", "\[IndentingNewLine]", 
      RowBox[{"Matrix", "[", 
       RowBox[{"m1", ",", "m0"}], "]"}], ",", "\[IndentingNewLine]", 
      RowBox[{"PadWithZeroRows", "[", 
       RowBox[{
        RowBox[{"identityMatrix", "[", "m0", "]"}], ",", 
        RowBox[{
         RowBox[{"WeightMultiplicityPartialSums", "[", 
          RowBox[{"\[CapitalGamma]", ",", "V", ",", "W", ",", 
           RowBox[{"\[Lambda]", "+", "\[Mu]"}]}], "]"}], 
         "\[LeftDoubleBracket]", 
         RowBox[{"pos", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}], 
         "\[RightDoubleBracket]"}], ",", "m1"}], "]"}]}], 
     "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"HighWeightVectorQ", "[", 
   RowBox[{"\[CapitalGamma]_", ",", "V_", ",", "b_", ",", "\[Lambda]_"}], 
   "]"}], ":=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"And", "@@", 
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"ZeroVectorQ", "[", 
        RowBox[{"Together", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
            RowBox[{"SuperPlus", "[", 
             SubscriptBox["X", "i"], "]"}], "]"}], "[", 
           RowBox[{"V", ",", "b", ",", "\[Lambda]"}], "]"}], ".", "#"}], 
         "]"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", 
         RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "}"}]}], "]"}]}], 
    ")"}], "&"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.409501989977062*^9, 3.4095020020344*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"HighWeightVectors", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Lambda]_", 
     "]"}], ",", "_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{"{", "1", "}"}], "}"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"HighWeightVectors", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Lambda]_", 
     "]"}], ",", "_", ",", "_"}], "]"}], ":=", 
  RowBox[{"{", "}"}]}]}], "Input",
 InitializationCell->True],

Cell["\<\
In order to allow branching calculations, I changed (2007-11-06) the pattern \
from:\
\>", "Text"],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"HighWeightVectors", "[", "\[CapitalGamma]_", "]"}], "[", 
    RowBox[{
     RowBox[{"V", ":", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "_", "]"}], 
        "\[CircleTimes]", 
        RowBox[{
         RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "_", "]"}]}], 
       ")"}]}], ",", "b_", ",", "\[Lambda]_"}], "]"}], ":="}], " ", 
  "*)"}]], "Input",
 CellChangeTimes->{{3.40950215935061*^9, 3.4095021635466433`*^9}}],

Cell["to the less restrictive:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"HighWeightVectors", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_", ",", "b_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"HighWeightVectors", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"V", ",", "b", ",", "\[Lambda]"}], "]"}], "=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"T", ",", "r"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"T", "=", 
       RowBox[{"AppendColumns", "@@", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
            RowBox[{"SuperPlus", "[", 
             SubscriptBox["X", "i"], "]"}], "]"}], "[", 
           RowBox[{"V", ",", "b", ",", "\[Lambda]"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "1", ",", 
            RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "}"}]}], 
         "]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Dimensions", "[", "T", "]"}], "\[LeftDoubleBracket]", "1", 
          "\[RightDoubleBracket]"}], "\[Equal]", "0"}], ",", 
        RowBox[{"Return", "[", 
         RowBox[{"IdentityMatrix", "[", 
          RowBox[{"WeightMultiplicity", "[", 
           RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}], 
          "]"}], "]"}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"DebugPrint", "[", 
       RowBox[{"\"\<About to find the null space of a \>\"", ",", 
        RowBox[{
         RowBox[{"Dimensions", "[", "T", "]"}], "\[LeftDoubleBracket]", "2", 
         "\[RightDoubleBracket]"}], ",", "\"\< by \>\"", ",", 
        RowBox[{
         RowBox[{"Dimensions", "[", "T", "]"}], "\[LeftDoubleBracket]", "2", 
         "\[RightDoubleBracket]"}], ",", "\"\< matrix.\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"r", "=", 
       RowBox[{"Together", "[", 
        RowBox[{"NullSpace", "[", 
         RowBox[{"T", ",", 
          RowBox[{"Method", "\[Rule]", "\"\<OneStepRowReduction\>\""}]}], 
         "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
      "DebugPrint", "[", "\"\<Finished finding null space.\>\"", "]"}], ";", 
      "\[IndentingNewLine]", "r"}]}], "\[IndentingNewLine]", 
    "]"}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.409501949578973*^9, 3.409501955367296*^9}, {
  3.4304045658616285`*^9, 3.430404567724307*^9}}],

Cell["\<\
Changing PositiveWeights to the more sensible HighWeights in the next line \
seems to break things. Weight ordering issues?\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"AllHighWeightVectors", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_"}], "]"}], ":=", 
  RowBox[{"Flatten", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Function", "[", 
      RowBox[{
       RowBox[{"{", "\[Lambda]", "}"}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"{", 
          RowBox[{"\[Lambda]", ",", "#"}], "}"}], "&"}], "/@", 
        RowBox[{
         RowBox[{"HighWeightVectors", "[", "\[CapitalGamma]", "]"}], "[", 
         RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}]}]}], "]"}], "/@", 
     RowBox[{"PositiveWeights", "[", 
      RowBox[{"\[CapitalGamma]", ",", "V"}], "]"}]}], ",", "1"}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"HighWeights", "[", 
   RowBox[{"\[CapitalGamma]_", ",", "V_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"SortWeights", "[", "\[CapitalGamma]", "]"}], "[", 
   RowBox[{"Union", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"DecomposeRepresentation", "[", "\[CapitalGamma]", "]"}], "[", 
       "V", "]"}], "/.", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"DirectSum", "[", "v__", "]"}], "\[RuleDelayed]", 
         RowBox[{"{", "v", "}"}]}], ",", 
        RowBox[{
         RowBox[{"i", ":", 
          RowBox[{
           RowBox[{"Irrep", "[", "_", "]"}], "[", "_", "]"}]}], 
         "\[RuleDelayed]", 
         RowBox[{"{", "i", "}"}]}]}], "}"}]}], "/.", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", "\[CapitalGamma]", "]"}], "[", "\[Lambda]_", 
       "]"}], "\[RuleDelayed]", "\[Lambda]"}]}], "]"}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"AddOneVectorToSpanningSet", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", 
    RowBox[{"v_", "?", "VectorQ"}]}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{"ZeroVectorQ", "[", "v", "]"}], ",", 
    RowBox[{"{", "}"}], ",", 
    RowBox[{"{", "v", "}"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"AddOneVectorToSpanningSet", "[", 
   RowBox[{
    RowBox[{"m_", "?", "MatrixQ"}], ",", 
    RowBox[{"v_", "?", "VectorQ"}]}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"MatrixRank", "[", 
      RowBox[{"m", "~", "Join", "~", 
       RowBox[{"{", "v", "}"}]}], "]"}], "\[Equal]", 
     RowBox[{"Length", "[", "m", "]"}]}], ",", "m", ",", 
    RowBox[{"m", "~", "Join", "~", 
     RowBox[{"{", "v", "}"}]}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"SpanningSet", "[", 
    RowBox[{"{", "}"}], "]"}], ":=", 
   RowBox[{"{", "}"}]}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"SpanningSet", "[", 
   RowBox[{"m_", "?", "MatrixQ"}], "]"}], ":=", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"Global`spanningSetArgument", "=", "m"}], ";", 
    RowBox[{"Fold", "[", 
     RowBox[{"AddOneVectorToSpanningSet", ",", 
      RowBox[{"{", "}"}], ",", "m"}], "]"}]}], ")"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"SpanningSet", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"_", ",", "_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"SpanningSet", "[", "data", "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"LazyAddOneVectorToSpanningSet", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", 
    RowBox[{"v_", "?", "VectorQ"}]}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{"ZeroVectorQ", "[", "v", "]"}], ",", 
    RowBox[{"{", "}"}], ",", 
    RowBox[{"{", "v", "}"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"LazyAddOneVectorToSpanningSet", "[", 
   RowBox[{
    RowBox[{"m_", "?", "MatrixQ"}], ",", 
    RowBox[{"v_", "?", "VectorQ"}]}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"MatrixRank", "[", 
      RowBox[{
       RowBox[{"m", "~", "Join", "~", 
        RowBox[{"{", "v", "}"}]}], "/.", 
       RowBox[{"q", "\[Rule]", "1"}]}], "]"}], "\[Equal]", 
     RowBox[{"Length", "[", "m", "]"}]}], ",", "m", ",", 
    RowBox[{"m", "~", "Join", "~", 
     RowBox[{"{", "v", "}"}]}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"LazySpanningSet", "[", 
    RowBox[{"{", "}"}], "]"}], ":=", 
   RowBox[{"{", "}"}]}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"LazySpanningSet", "[", 
   RowBox[{"m_", "?", "MatrixQ"}], "]"}], ":=", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"Global`spanningSetArgument", "=", "m"}], ";", 
    RowBox[{"Fold", "[", 
     RowBox[{"LazyAddOneVectorToSpanningSet", ",", 
      RowBox[{"{", "}"}], ",", "m"}], "]"}]}], ")"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"LazySpanningSet", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"_", ",", "_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"LazySpanningSet", "[", "data", "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"CarefulSpanningSetNewTime", "=", 
   RowBox[{"CarefulSpanningSetOldTime", "=", "0"}]}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"CarefulSpanningSet", "[", "m_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"tnew", ",", "rnew", ",", "told", ",", "rold"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"{", 
       RowBox[{"tnew", ",", "rnew"}], "}"}], "=", 
      RowBox[{"AbsoluteTiming", "[", 
       RowBox[{"LazySpanningSet", "[", "m", "]"}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"told", ",", "rold"}], "}"}], "=", 
      RowBox[{"AbsoluteTiming", "[", 
       RowBox[{"SpanningSet", "[", "m", "]"}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"rnew", "=!=", "rold"}], ",", 
       RowBox[{"Print", "[", "\"\<LazySpanningSet failed!\>\"", "]"}]}], 
      "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"CarefulSpanningSetNewTime", "+=", "tnew"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"CarefulSpanningSetOldTime", "+=", "told"}], ";", 
     "\[IndentingNewLine]", "rnew"}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{
   "V_", ",", "\[Beta]_", ",", "\[Lambda]_", ",", "v_", ",", "\[Lambda]_"}], 
   "]"}], ":=", 
  RowBox[{"{", "v", "}"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]_", "]"}], "[", 
    RowBox[{
    "V_", ",", "\[Beta]_", ",", "\[Lambda]_", ",", "v_", ",", "\[Mu]_"}], 
    "]"}], "/;", 
   RowBox[{"InWeylPolytopeQ", "[", 
    RowBox[{"\[CapitalGamma]", ",", "\[Lambda]", ",", "\[Mu]"}], "]"}]}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]", ",", "v", ",", "\[Mu]"}], 
    "]"}], "=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"c", ",", "r"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"DebugPrintHeld", "[", 
       RowBox[{"\"\<Calculating \>\"", ",", 
        RowBox[{
         RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]", "]"}], "[", 
         RowBox[{
         "V", ",", "\[Beta]", ",", "\[Lambda]", ",", "v", ",", "\[Mu]"}], 
         "]"}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"c", "=", 
       RowBox[{"Join", "@@", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{
               RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], 
               "[", 
               RowBox[{"SuperMinus", "[", 
                SubscriptBox["X", "i"], "]"}], "]"}], "[", 
              RowBox[{"V", ",", "\[Beta]", ",", 
               RowBox[{"\[Mu]", "-", 
                RowBox[{
                 RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
                 RowBox[{"SuperMinus", "[", 
                  SubscriptBox["X", "i"], "]"}], "]"}]}]}], "]"}], ".", "#"}],
             "&"}], "/@", 
           RowBox[{
            RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]", "]"}], "[", 
            RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]", ",", "v", ",", 
             RowBox[{"\[Mu]", "-", 
              RowBox[{
               RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
               RowBox[{"SuperMinus", "[", 
                SubscriptBox["X", "i"], "]"}], "]"}]}]}], "]"}]}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "1", ",", 
            RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], "}"}]}], 
         "]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"DebugPrint", "[", "\"\< ... prepared spanning set.\>\"", "]"}],
       ";", "\[IndentingNewLine]", 
      RowBox[{"r", "=", 
       RowBox[{"LazySpanningSet", "[", 
        RowBox[{"Together", "[", "c", "]"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"DebugPrintHeld", "[", 
       RowBox[{"\"\<Finished calculating \>\"", ",", 
        RowBox[{
         RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]", "]"}], "[", 
         RowBox[{
         "V", ",", "\[Beta]", ",", "\[Lambda]", ",", "v", ",", "\[Mu]"}], 
         "]"}]}], "]"}], ";", "\[IndentingNewLine]", "r"}]}], 
    "\[IndentingNewLine]", "]"}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.409502225836211*^9, 3.4095022479580207`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{
   "V_", ",", "\[Beta]_", ",", "\[Lambda]_", ",", "v_", ",", "\[Mu]_"}], 
   "]"}], ":=", 
  RowBox[{"{", "}"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"FullSubIrrepWeightBasis", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Mu]_"}], "]"}], ":=", 
  RowBox[{"Flatten", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"SubIrrepWeightBasis", "[", "\[CapitalGamma]", "]"}], "[", 
       RowBox[{"V", ",", "\[Beta]", ",", 
        RowBox[{"#", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
        ",", 
        RowBox[{"#", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], 
        ",", "\[Mu]"}], "]"}], "&"}], "/@", 
     RowBox[{
      RowBox[{"AllHighWeightVectors", "[", "\[CapitalGamma]", "]"}], "[", 
      RowBox[{"V", ",", "\[Beta]"}], "]"}]}], ",", "1"}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DirectSumDecomposition", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"DirectSumDecomposition", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}], "=", 
   "\[IndentingNewLine]", 
   RowBox[{"With", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"data", "=", 
        RowBox[{
         RowBox[{"FullSubIrrepWeightBasis", "[", "\[CapitalGamma]", "]"}], 
         "[", 
         RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}]}], ",", 
       RowBox[{"d", "=", 
        RowBox[{"WeightMultiplicity", "[", 
         RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}]}]}], 
      "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Length", "[", "data", "]"}], "\[NotEqual]", "d"}], ",", 
        RowBox[{
         RowBox[{"Print", "[", 
          RowBox[{
          "\"\<Direct sum decomposition failed, in weight space \>\"", ",", 
           "\[Lambda]", ",", "\"\< of \>\"", ",", " ", "V"}], "]"}], ";", 
         RowBox[{"Return", "[", "$Failed", "]"}]}]}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"Transpose", "[", 
       RowBox[{"Matrix", "[", 
        RowBox[{
         RowBox[{"Length", "[", "data", "]"}], ",", "d", ",", "data"}], "]"}],
        "]"}]}]}], "\[IndentingNewLine]", "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"InverseDirectSumDecomposition", "[", "\[CapitalGamma]_", "]"}], 
   "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"InverseDirectSumDecomposition", "[", "\[CapitalGamma]", "]"}], 
    "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}], "=", 
   RowBox[{"Inverse", "[", 
    RowBox[{
     RowBox[{"DirectSumDecomposition", "[", "\[CapitalGamma]", "]"}], "[", 
     RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], "]"}], 
    "]"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DirectSumProjection", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_DirectSum", ",", "index_Integer", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"a", "=", 
       RowBox[{"WeightMultiplicity", "[", 
        RowBox[{"\[CapitalGamma]", ",", 
         RowBox[{"Take", "[", 
          RowBox[{"V", ",", 
           RowBox[{"index", "-", "1"}]}], "]"}], ",", "\[Lambda]"}], "]"}]}], 
      ",", 
      RowBox[{"b", "=", 
       RowBox[{"WeightMultiplicity", "[", 
        RowBox[{"\[CapitalGamma]", ",", 
         RowBox[{
         "V", "\[LeftDoubleBracket]", "index", "\[RightDoubleBracket]"}], ",",
          "\[Lambda]"}], "]"}]}], ",", 
      RowBox[{"c", "=", 
       RowBox[{"WeightMultiplicity", "[", 
        RowBox[{"\[CapitalGamma]", ",", 
         RowBox[{"Drop", "[", 
          RowBox[{"V", ",", "index"}], "]"}], ",", "\[Lambda]"}], "]"}]}]}], 
     "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{"AppendRows", "[", 
     RowBox[{
      RowBox[{"ZeroesMatrix", "[", 
       RowBox[{"b", ",", "a"}], "]"}], ",", 
      RowBox[{"identityMatrix", "[", "b", "]"}], ",", 
      RowBox[{"ZeroesMatrix", "[", 
       RowBox[{"b", ",", "c"}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DirectSumProjection", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_DirectSum", ",", 
    RowBox[{"indexes", ":", 
     RowBox[{"{", "___Integer", "}"}]}], ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"AppendColumns", "@@", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"DirectSumProjection", "[", "\[CapitalGamma]", "]"}], "[", 
       RowBox[{"V", ",", "#", ",", "\[Lambda]"}], "]"}], "&"}], "/@", 
     "indexes"}], ")"}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DirectSumInclusion", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_DirectSum", ",", "index_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"Transpose", "[", 
   RowBox[{
    RowBox[{"DirectSumProjection", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"V", ",", "index", ",", "\[Lambda]"}], "]"}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"ChangeOfBasisMatrix", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{
    RowBox[{"V", ":", 
     RowBox[{
      RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "_", "]"}]}], ",",
     "\[Beta]_", ",", "FundamentalBasis", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"DirectSumDecomposition", "[", "\[CapitalGamma]", "]"}], "[", 
   RowBox[{"V", ",", "\[Beta]", ",", "\[Lambda]"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ChangeOfBasisMatrix", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{
    RowBox[{"V", ":", 
     RowBox[{
      RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "_", "]"}]}], ",",
     "FundamentalBasis", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"Inverse", "[", 
   RowBox[{
    RowBox[{"ChangeOfBasisMatrix", "[", "\[CapitalGamma]", "]"}], "[", 
    RowBox[{"V", ",", "\[Beta]", ",", "FundamentalBasis", ",", "\[Lambda]"}], 
    "]"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ChangeOfBasisMatrix", "[", "\[CapitalGamma]_", "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}],
   ":=", 
  RowBox[{"identityMatrix", "[", 
   RowBox[{"WeightMultiplicity", "[", 
    RowBox[{"\[CapitalGamma]", ",", "V", ",", "\[Lambda]"}], "]"}], 
   "]"}]}]}], "Input",
 InitializationCell->True],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A", ":", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"SuperPlus", "[", 
         SubscriptBox["X", "_"], "]"}], "|", 
        RowBox[{"SuperMinus", "[", 
         SubscriptBox["X", "_"], "]"}]}], ")"}]}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Lambda]_", 
      "]"}], ",", "ShortRootBasis", ",", "\[Kappa]_"}], "]"}], "/;", 
   RowBox[{
    RowBox[{"UnitVectorQ", "[", "\[Lambda]", "]"}], "\[And]", 
    RowBox[{"ShortDominantRootQ", "[", 
     RowBox[{"\[CapitalGamma]", ",", "\[Lambda]"}], "]"}]}]}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A", 
     "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", "\[CapitalGamma]", "]"}], "[", "\[Lambda]", "]"}],
      ",", "ShortRootBasis", ",", "\[Mu]"}], "]"}], "=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
      "x", ",", "h", ",", "s", ",", "d", ",", "kf", ",", "action", ",", 
       "basis"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       SubscriptBox["s", "i_"], ":=", 
       RowBox[{
        RowBox[{"SimpleRoots", "[", "\[CapitalGamma]", "]"}], 
        "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{
       SubscriptBox["d", "i_"], ":=", 
       RowBox[{
        RowBox[{"CartanFactors", "[", "\[CapitalGamma]", "]"}], 
        "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"kf", "=", 
       RowBox[{"KillingForm", "[", "\[CapitalGamma]", "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"action", "[", 
         RowBox[{"SuperPlus", "[", 
          SubscriptBox["X", "i_"], "]"}], "]"}], "[", 
        SubscriptBox["x", "\[Mu]_"], "]"}], ":=", 
       RowBox[{"Switch", "[", 
        RowBox[{
         RowBox[{"2", 
          FractionBox[
           RowBox[{"kf", "[", 
            RowBox[{"\[Mu]", ",", 
             SubscriptBox["s", "i"]}], "]"}], 
           RowBox[{"kf", "[", 
            RowBox[{
             SubscriptBox["s", "i"], ",", 
             SubscriptBox["s", "i"]}], "]"}]]}], ",", 
         RowBox[{"2", "|", "0", "|", "1"}], ",", "0", ",", 
         RowBox[{"-", "1"}], ",", 
         SubscriptBox["x", 
          RowBox[{"\[Mu]", "+", 
           SubscriptBox["s", "i"]}]], ",", 
         RowBox[{"-", "2"}], ",", 
         SubscriptBox["h", 
          SubscriptBox["s", "i"]]}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"action", "[", 
         RowBox[{"SuperMinus", "[", 
          SubscriptBox["X", "i_"], "]"}], "]"}], "[", 
        SubscriptBox["x", "\[Mu]_"], "]"}], ":=", 
       RowBox[{"Switch", "[", 
        RowBox[{
         RowBox[{"2", 
          FractionBox[
           RowBox[{"kf", "[", 
            RowBox[{"\[Mu]", ",", 
             SubscriptBox["s", "i"]}], "]"}], 
           RowBox[{"kf", "[", 
            RowBox[{
             SubscriptBox["s", "i"], ",", 
             SubscriptBox["s", "i"]}], "]"}]]}], ",", 
         RowBox[{
          RowBox[{"-", "2"}], "|", "0", "|", 
          RowBox[{"-", "1"}]}], ",", "0", ",", "1", ",", 
         SubscriptBox["x", 
          RowBox[{"\[Mu]", "-", 
           SubscriptBox["s", "i"]}]], ",", "2", ",", 
         SubscriptBox["h", 
          SubscriptBox["s", "i"]]}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"action", "[", 
         RowBox[{"SuperPlus", "[", 
          SubscriptBox["X", "i_"], "]"}], "]"}], "[", 
        SubscriptBox["h", "\[Mu]_"], "]"}], ":=", 
       RowBox[{
        RowBox[{"Switch", "[", 
         RowBox[{
          RowBox[{"2", 
           FractionBox[
            RowBox[{"kf", "[", 
             RowBox[{"\[Mu]", ",", 
              SubscriptBox["s", "i"]}], "]"}], 
            RowBox[{"kf", "[", 
             RowBox[{
              SubscriptBox["s", "i"], ",", 
              SubscriptBox["s", "i"]}], "]"}]]}], ",", "2", ",", 
          RowBox[{
           RowBox[{"qInteger", "[", "2", "]"}], "[", 
           SuperscriptBox["q", 
            SubscriptBox["d", "i"]], "]"}], ",", 
          RowBox[{"-", "1"}], ",", "1", ",", "_", ",", "0"}], "]"}], 
        SubscriptBox["x", 
         SubscriptBox["s", "i"]]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"action", "[", 
         RowBox[{"SuperMinus", "[", 
          SubscriptBox["X", "i_"], "]"}], "]"}], "[", 
        SubscriptBox["h", "\[Mu]_"], "]"}], ":=", 
       RowBox[{
        RowBox[{"Switch", "[", 
         RowBox[{
          RowBox[{"2", 
           FractionBox[
            RowBox[{"kf", "[", 
             RowBox[{"\[Mu]", ",", 
              SubscriptBox["s", "i"]}], "]"}], 
            RowBox[{"kf", "[", 
             RowBox[{
              SubscriptBox["s", "i"], ",", 
              SubscriptBox["s", "i"]}], "]"}]]}], ",", "2", ",", 
          RowBox[{
           RowBox[{"qInteger", "[", "2", "]"}], "[", 
           SuperscriptBox["q", 
            SubscriptBox["d", "i"]], "]"}], ",", 
          RowBox[{"-", "1"}], ",", "1", ",", "_", ",", "0"}], "]"}], 
        SubscriptBox["x", 
         RowBox[{"-", 
          SubscriptBox["s", "i"]}]]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"basis", "[", "\[Mu]_", "]"}], ":=", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{"MemberQ", "[", 
          RowBox[{
           RowBox[{"ShortRoots", "[", "\[CapitalGamma]", "]"}], ",", 
           "\[Mu]"}], "]"}], ",", 
         RowBox[{"{", 
          SubscriptBox["x", "\[Mu]"], "}"}], ",", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"ZeroVectorQ", "[", "\[Mu]", "]"}], ",", 
           RowBox[{
            RowBox[{
             SubscriptBox["h", "#"], "&"}], "/@", 
            RowBox[{"ShortSimpleRoots", "[", "\[CapitalGamma]", "]"}]}], ",", 
           RowBox[{"{", "}"}]}], "]"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"With", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"\[Beta]1", "=", 
           RowBox[{"basis", "[", "\[Kappa]", "]"}]}], ",", 
          RowBox[{"\[Beta]2", "=", 
           RowBox[{"basis", "[", 
            RowBox[{"\[Kappa]", "+", 
             RowBox[{
              RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
              "A", "]"}]}], "]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
        RowBox[{"Matrix", "[", 
         RowBox[{
          RowBox[{"Length", "[", "\[Beta]2", "]"}], ",", 
          RowBox[{"Length", "[", "\[Beta]1", "]"}], ",", 
          RowBox[{
           RowBox[{
            RowBox[{"Coefficient", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"action", "[", "A", "]"}], "/@", "\[Beta]1"}], ",", 
              "#"}], "]"}], "&"}], "/@", "\[Beta]2"}]}], "]"}]}], 
       "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", 
    "]"}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4095018675309935`*^9, 3.409501877395178*^9}, {
  3.409502276058427*^9, 3.4095022985107117`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"General", "::", "\<\"spell1\"\>"}], ":", 
  " ", "\<\"Possible spelling error: new symbol name \\\"\\!\\(action\\)\\\" \
is similar to existing symbol \\\"\\!\\(Action\\)\\\". \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::spell1\\\"]\\)\"\>"}]], \
"Message"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A", ":", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"SuperPlus", "[", 
         SubscriptBox["X", "_"], "]"}], "|", 
        RowBox[{"SuperMinus", "[", 
         SubscriptBox["X", "_"], "]"}]}], ")"}]}], "]"}], "[", 
    RowBox[{
     RowBox[{"V", ":", 
      RowBox[{
       RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Lambda]_", 
       "]"}]}], ",", "FundamentalBasis", ",", "\[Kappa]_"}], "]"}], "/;", 
   RowBox[{
    RowBox[{"UnitVectorQ", "[", "\[Lambda]", "]"}], "\[And]", 
    RowBox[{"ShortDominantRootQ", "[", 
     RowBox[{"\[CapitalGamma]", ",", "\[Lambda]"}], "]"}]}]}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A", 
     "]"}], "[", 
    RowBox[{"V", ",", "FundamentalBasis", ",", "\[Kappa]"}], "]"}], "=", 
   RowBox[{
    RowBox[{
     RowBox[{"ChangeOfBasisMatrix", "[", "\[CapitalGamma]", "]"}], "[", 
     RowBox[{"V", ",", "FundamentalBasis", ",", "ShortRootBasis", ",", 
      RowBox[{"\[Kappa]", "+", 
       RowBox[{
        RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", "A", 
        "]"}]}]}], "]"}], ".", 
    RowBox[{
     RowBox[{
      RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A", 
      "]"}], "[", 
     RowBox[{"V", ",", "ShortRootBasis", ",", "\[Kappa]"}], "]"}], ".", 
    RowBox[{
     RowBox[{"ChangeOfBasisMatrix", "[", "\[CapitalGamma]", "]"}], "[", 
     RowBox[{
     "V", ",", "ShortRootBasis", ",", "FundamentalBasis", ",", "\[Kappa]"}], 
     "]"}]}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.409501882832997*^9, 3.409501892797325*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"IrrepContainmentRules", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["B", "n_"], "]"}], "[", "\[Lambda]_", "]"}], "/;", 
       RowBox[{"MemberQ", "[", 
        RowBox[{
         RowBox[{"Take", "[", 
          RowBox[{
           RowBox[{"IdentityMatrix", "[", "n", "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"2", ",", 
             RowBox[{"-", "2"}]}], "}"}]}], "]"}], ",", "\[Lambda]"}], 
        "]"}]}], "\[RuleDelayed]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["B", "n"], "]"}], "[", 
        RowBox[{"RotateLeft", "[", "\[Lambda]", "]"}], "]"}], 
       "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["B", "n"], "]"}], "[", 
        RowBox[{"UnitVector", "[", 
         RowBox[{"n", ",", "1"}], "]"}], "]"}]}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["C", "n_"], "]"}], "[", "\[Lambda]_", "]"}], "/;", 
       RowBox[{"MemberQ", "[", 
        RowBox[{
         RowBox[{"Drop", "[", 
          RowBox[{
           RowBox[{"IdentityMatrix", "[", "n", "]"}], ",", "2"}], "]"}], ",", 
         "\[Lambda]"}], "]"}]}], "\[RuleDelayed]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["C", "n"], "]"}], "[", 
        RowBox[{"RotateLeft", "[", "\[Lambda]", "]"}], "]"}], 
       "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["C", "n"], "]"}], "[", 
        RowBox[{"UnitVector", "[", 
         RowBox[{"n", ",", "1"}], "]"}], "]"}]}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["D", "n_"], "]"}], "[", "\[Lambda]_", "]"}], "/;", 
       RowBox[{"MemberQ", "[", 
        RowBox[{
         RowBox[{"Take", "[", 
          RowBox[{
           RowBox[{"IdentityMatrix", "[", "n", "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", 
             RowBox[{"-", "3"}]}], "}"}]}], "]"}], ",", "\[Lambda]"}], 
        "]"}]}], "\[RuleDelayed]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"EvenQ", "[", 
         RowBox[{"n", "-", 
          RowBox[{
           RowBox[{"Position", "[", 
            RowBox[{
             RowBox[{"IdentityMatrix", "[", "n", "]"}], ",", "\[Lambda]"}], 
            "]"}], "\[LeftDoubleBracket]", 
           RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}]}], "]"}], ",", 
        RowBox[{
         RowBox[{
          RowBox[{"Irrep", "[", 
           SubscriptBox["D", "n"], "]"}], "[", 
          RowBox[{"UnitVector", "[", 
           RowBox[{"n", ",", "n"}], "]"}], "]"}], "\[CircleTimes]", 
         RowBox[{
          RowBox[{"Irrep", "[", 
           SubscriptBox["D", "n"], "]"}], "[", 
          RowBox[{"UnitVector", "[", 
           RowBox[{"n", ",", "n"}], "]"}], "]"}]}], ",", 
        RowBox[{
         RowBox[{
          RowBox[{"Irrep", "[", 
           SubscriptBox["D", "n"], "]"}], "[", 
          RowBox[{"UnitVector", "[", 
           RowBox[{"n", ",", 
            RowBox[{"n", "-", "1"}]}], "]"}], "]"}], "\[CircleTimes]", 
         RowBox[{
          RowBox[{"Irrep", "[", 
           SubscriptBox["D", "n"], "]"}], "[", 
          RowBox[{"UnitVector", "[", 
           RowBox[{"n", ",", "n"}], "]"}], "]"}]}]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["G", "2"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "1"}], "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["G", "2"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "0"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["F", "4"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["F", "4"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["F", "4"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["F", "4"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["F", "4"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["F", "4"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["E", "6"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "6"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["E", "6"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "6"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "6"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "1"}], 
         "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["E", "6"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "6"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "1"}], 
         "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["E", "6"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "1", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "6"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "6"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "7"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "7"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "1"}], 
         "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "7"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "7"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "7"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0"}], 
        "}"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "7"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
         "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", 
         "0"}], "}"}], "]"}], "->", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "1"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0"}], "}"}], "]"}], "->", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "1"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0"}], "}"}], "]"}], "->", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "1"}], "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "0", ",", 
         "0"}], "}"}], "]"}], "->", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0"}], "}"}], "]"}], "->", 
      SuperscriptBox[
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}], 
       RowBox[{"\[CircleTimes]", "2"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", 
         "0"}], "}"}], "]"}], "->", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Irrep", "[", 
        SubscriptBox["\[ExponentialE]", "8"], "]"}], "[", 
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", 
         "0"}], "}"}], "]"}], "->", 
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "0", ",", "0", ",", "1", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}], "\[CircleTimes]", 
       RowBox[{
        RowBox[{"Irrep", "[", 
         SubscriptBox["E", "8"], "]"}], "[", 
        RowBox[{"{", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
           "0"}], "}"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Lambda]_", 
        "]"}], "/;", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"!", 
          RowBox[{"UnitVectorQ", "[", "\[Lambda]", "]"}]}], "\[And]", 
         RowBox[{"!", 
          RowBox[{"ZeroVectorQ", "[", "\[Lambda]", "]"}]}]}], ")"}]}], 
      "\[RuleDelayed]", 
      RowBox[{"Module", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"n", "=", 
           RowBox[{"Rank", "[", "\[CapitalGamma]", "]"}]}], ",", "pos", ",", 
          "\[Mu]"}], "}"}], ",", 
        RowBox[{
         RowBox[{"pos", "=", 
          RowBox[{
           RowBox[{"Position", "[", 
            RowBox[{"\[Lambda]", ",", 
             RowBox[{"_", "?", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"#", "\[NotEqual]", "0"}], "&"}], ")"}]}]}], "]"}], 
           "\[LeftDoubleBracket]", 
           RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}]}], ";", 
         RowBox[{"\[Mu]", "=", 
          RowBox[{"UnitVector", "[", 
           RowBox[{"n", ",", "pos"}], "]"}]}], ";", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"Irrep", "[", "\[CapitalGamma]", "]"}], "[", 
           RowBox[{"\[Lambda]", "-", "\[Mu]"}], "]"}], "\[CircleTimes]", 
          RowBox[{
           RowBox[{"Irrep", "[", "\[CapitalGamma]", "]"}], "[", "\[Mu]", 
           "]"}]}]}]}], "]"}]}]}], "\[IndentingNewLine]", "}"}]}], 
  ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4201110099718018`*^9, 3.4201110220691967`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MatrixPresentation", "[", 
     SubscriptBox["A", "1"], "]"}], "[", 
    RowBox[{"Z", ":", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"SuperPlus", "[", 
        SubscriptBox["X", "1"], "]"}], "|", 
       RowBox[{"SuperMinus", "[", 
        SubscriptBox["X", "1"], "]"}]}], ")"}]}], "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
    RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"fastMatrixPresentation", "[", 
     SubscriptBox["A", "1"], "]"}], "[", "Z", "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"{", "\[Lambda]", "}"}], "]"}], ",", "FundamentalBasis", ",", 
    RowBox[{"{", "\[Mu]", "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.4189409086956034`*^9, 3.418940984003891*^9}, {
  3.4189411880272627`*^9, 3.4189411967397904`*^9}, {3.4189415183322177`*^9, 
  3.4189415261634784`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]_", "]"}], "[", 
     RowBox[{"A", ":", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"SuperPlus", "[", 
         SubscriptBox["X", "_"], "]"}], "|", 
        RowBox[{"SuperMinus", "[", 
         SubscriptBox["X", "_"], "]"}]}], ")"}]}], "]"}], "[", 
    RowBox[{
     RowBox[{"V", ":", 
      RowBox[{
       RowBox[{"Irrep", "[", "\[CapitalGamma]_", "]"}], "[", "\[Mu]_", 
       "]"}]}], ",", "FundamentalBasis", ",", "\[Lambda]_"}], "]"}], "/;", 
   RowBox[{
    RowBox[{
     RowBox[{"\[Not]", 
      RowBox[{"ZeroVectorQ", "[", "\[Mu]", "]"}]}], "\[And]", 
     RowBox[{"\[Not]", 
      RowBox[{"UnitVectorQ", "[", "\[Mu]", "]"}]}]}], "\[Or]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Not]", 
       RowBox[{"ShortDominantRootQ", "[", 
        RowBox[{"\[CapitalGamma]", ",", "\[Mu]"}], "]"}]}], "\[And]", 
      RowBox[{"\[Not]", 
       RowBox[{"MinusculeRepresentationQ", "[", 
        RowBox[{"\[CapitalGamma]", ",", "V"}], "]"}]}]}], ")"}]}]}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"W", "=", 
       RowBox[{"V", "/.", "IrrepContainmentRules"}]}], ",", "p", ",", 
      "result"}], "}"}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", "A",
         "]"}], "[", 
       RowBox[{"V", ",", "FundamentalBasis", ",", "\[Lambda]"}], "]"}], "=", 
      "\[IndentingNewLine]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"DebugPrintHeld", "[", 
         RowBox[{"\"\<Calculating \>\"", ",", 
          RowBox[{
           RowBox[{
            RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
            "A", "]"}], "[", 
           RowBox[{"V", ",", "FundamentalBasis", ",", "\[Lambda]"}], "]"}], 
          ",", " ", "\"\< by looking at \>\"", ",", "V", ",", 
          "\"\< as a subrep of \>\"", ",", 
          RowBox[{"Evaluate", "[", "W", "]"}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"W", "\[Equal]", "V"}], ",", 
          RowBox[{
           RowBox[{"Print", "[", 
            RowBox[{
            "\"\<Warning, couldn't work out how to find matrix presentations \
for \>\"", ",", "V"}], "]"}], ";", 
           RowBox[{"Return", "[", "$Failed", "]"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"p", "=", 
         RowBox[{
          RowBox[{"Position", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"DecomposeRepresentation", "[", "\[CapitalGamma]", "]"}],
              "[", "W", "]"}], ",", "V"}], "]"}], "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"result", "=", 
         RowBox[{"Simplify", "[", 
          RowBox[{"Together", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"DirectSumProjection", "[", "\[CapitalGamma]", "]"}], 
             "[", 
             RowBox[{
              RowBox[{
               RowBox[{
               "DecomposeRepresentation", "[", "\[CapitalGamma]", "]"}], "[", 
               "W", "]"}], ",", "p", ",", 
              RowBox[{"\[Lambda]", "+", 
               RowBox[{
                RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
                "A", "]"}]}]}], "]"}], ".", 
            RowBox[{
             RowBox[{
             "InverseDirectSumDecomposition", "[", "\[CapitalGamma]", "]"}], 
             "[", 
             RowBox[{"W", ",", "FundamentalBasis", ",", 
              RowBox[{"\[Lambda]", "+", 
               RowBox[{
                RowBox[{"OperatorWeight", "[", "\[CapitalGamma]", "]"}], "[", 
                "A", "]"}]}]}], "]"}], ".", 
            RowBox[{
             RowBox[{
              RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], 
              "[", "A", "]"}], "[", 
             RowBox[{"W", ",", "FundamentalBasis", ",", "\[Lambda]"}], "]"}], 
            ".", 
            RowBox[{
             RowBox[{"DirectSumDecomposition", "[", "\[CapitalGamma]", "]"}], 
             "[", 
             RowBox[{"W", ",", "FundamentalBasis", ",", "\[Lambda]"}], "]"}], 
            ".", 
            RowBox[{
             RowBox[{"DirectSumInclusion", "[", "\[CapitalGamma]", "]"}], "[", 
             RowBox[{
              RowBox[{
               RowBox[{
               "DecomposeRepresentation", "[", "\[CapitalGamma]", "]"}], "[", 
               "W", "]"}], ",", "p", ",", "\[Lambda]"}], "]"}]}], "]"}], 
          "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"DebugPrintHeld", "[", 
         RowBox[{"\"\<Finished calculating \>\"", ",", 
          RowBox[{
           RowBox[{
            RowBox[{"MatrixPresentation", "[", "\[CapitalGamma]", "]"}], "[", 
            "A", "]"}], "[", 
           RowBox[{"V", ",", "FundamentalBasis", ",", "\[Lambda]"}], "]"}]}], 
         "]"}], ";", "\[IndentingNewLine]", "result"}], ")"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"ReportSavedMatrixPresentation", "[", "\[CapitalGamma]", "]"}],
        "[", "A", "]"}], "[", 
      RowBox[{"V", ",", "FundamentalBasis", ",", "\[Lambda]"}], "]"}], ";", 
     "\[IndentingNewLine]", "result"}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{3.4095019092810273`*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"numberOfSavedMatrixPresentations", "[", "_", "]"}], ":=", 
  "0"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"autopackagingMatrixPresentations", "=", "True"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"ReportSavedMatrixPresentation", "[", "\[CapitalGamma]_", "]"}], 
    "[", "A_", "]"}], "[", 
   RowBox[{"V_", ",", "\[Beta]_", ",", "\[Lambda]_"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"++", 
      RowBox[{
      "numberOfSavedMatrixPresentations", "[", "\[CapitalGamma]", "]"}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"autopackagingMatrixPresentations", "\[And]", 
        RowBox[{
         RowBox[{"Mod", "[", 
          RowBox[{
           RowBox[{
           "numberOfSavedMatrixPresentations", "[", "\[CapitalGamma]", "]"}], 
           ",", "10"}], "]"}], "\[Equal]", "0"}]}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"PackageMatrixPresentations", "[", "\[CapitalGamma]", "]"}]}], 
      "\[IndentingNewLine]", "]"}], ";"}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"fastMatrixPresentation", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"SuperPlus", "[", 
      SubscriptBox["X", "1"], "]"}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", 
       SubscriptBox["A", "1"], "]"}], "[", 
      RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
     RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], "/;", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "\[Lambda]"}], "\[LessEqual]", "\[Mu]", "<", "\[Lambda]"}], 
    "\[And]", 
    RowBox[{"EvenQ", "[", 
     RowBox[{"\[Lambda]", "-", "\[Mu]"}], "]"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"1", ",", "1", ",", 
    RowBox[{"{", 
     RowBox[{"{", 
      RowBox[{"Sum", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"qInteger", "[", "\[Nu]", "]"}], "[", "q", "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"\[Nu]", ",", 
          RowBox[{"Max", "[", 
           RowBox[{
            RowBox[{"\[Mu]", "+", "2"}], ",", 
            RowBox[{"-", "\[Mu]"}]}], "]"}], ",", "\[Lambda]", ",", "2"}], 
         "}"}]}], "]"}], "}"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.418940534076928*^9, 3.4189406508047743`*^9}, {
  3.4189411994937506`*^9, 3.4189412044809217`*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"fastMatrixPresentation", "[", 
     SubscriptBox["A", "1"], "]"}], "[", 
    RowBox[{"SuperPlus", "[", 
     SubscriptBox["X", "1"], "]"}], "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
    RowBox[{"{", "\[Lambda]_", "}"}]}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", "1", ",", 
    RowBox[{"{", "}"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"fastMatrixPresentation", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"SuperPlus", "[", 
      SubscriptBox["X", "1"], "]"}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", 
       SubscriptBox["A", "1"], "]"}], "[", 
      RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
     RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], "/;", 
   RowBox[{"\[Mu]", "==", 
    RowBox[{
     RowBox[{"-", "\[Lambda]"}], "-", "2"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"1", ",", "0", ",", 
    RowBox[{"{", 
     RowBox[{"{", "}"}], "}"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"fastMatrixPresentation", "[", 
     SubscriptBox["A", "1"], "]"}], "[", 
    RowBox[{"SuperPlus", "[", 
     SubscriptBox["X", "1"], "]"}], "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
    RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", "0"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.418940682370163*^9, 3.41894073482559*^9}, {
  3.4189412088472*^9, 3.4189412118114624`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"fastMatrixPresentation", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"SuperMinus", "[", 
      SubscriptBox["X", "1"], "]"}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", 
       SubscriptBox["A", "1"], "]"}], "[", 
      RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
     RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], "/;", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "\[Lambda]"}], "<", "\[Mu]", "\[LessEqual]", "\[Lambda]"}], 
    "\[And]", 
    RowBox[{"EvenQ", "[", 
     RowBox[{"\[Lambda]", "-", "\[Mu]"}], "]"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"1", ",", "1", ",", 
    RowBox[{"{", 
     RowBox[{"{", "1", "}"}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.418940534076928*^9, 3.4189406508047743`*^9}, {
  3.418940794241026*^9, 3.418940817454405*^9}, {3.418941213714198*^9, 
  3.418941216418086*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"fastMatrixPresentation", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"SuperMinus", "[", 
      SubscriptBox["X", "1"], "]"}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", 
       SubscriptBox["A", "1"], "]"}], "[", 
      RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
     RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], "/;", 
   RowBox[{"\[Lambda]", "\[Equal]", 
    RowBox[{"-", "\[Mu]"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", "1", ",", 
    RowBox[{"{", "}"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"fastMatrixPresentation", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"SuperMinus", "[", 
      SubscriptBox["X", "1"], "]"}], "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Irrep", "[", 
       SubscriptBox["A", "1"], "]"}], "[", 
      RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
     RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], "/;", 
   RowBox[{"\[Mu]", "==", 
    RowBox[{"\[Lambda]", "+", "2"}]}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"1", ",", "0", ",", 
    RowBox[{"{", 
     RowBox[{"{", "}"}], "}"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"fastMatrixPresentation", "[", 
     SubscriptBox["A", "1"], "]"}], "[", 
    RowBox[{"SuperMinus", "[", 
     SubscriptBox["X", "1"], "]"}], "]"}], "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Irrep", "[", 
      SubscriptBox["A", "1"], "]"}], "[", 
     RowBox[{"{", "\[Lambda]_", "}"}], "]"}], ",", "FundamentalBasis", ",", 
    RowBox[{"{", "\[Mu]_", "}"}]}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", "0"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.418940682370163*^9, 3.41894073482559*^9}, {
  3.418940796884827*^9, 3.418940834378741*^9}, {3.418941223017576*^9, 
  3.418941228174992*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"End", "[", "]"}], ";"}]], "Input",
 InitializationCell->True]
}, Open  ]],

Cell[CellGroupData[{

Cell["End of package", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"EndPackage", "[", "]"}], ";"}]], "Input",
 InitializationCell->True]
}, Closed]]
}, Open  ]]
}, Open  ]]
},
AutoGeneratedPackage->Automatic,
WindowSize->{776, 649},
WindowMargins->{{Automatic, 128}, {Automatic, 9}},
Visible->True,
DockedCells->(FrontEndExecute[{
   FrontEnd`NotebookApply[
    FrontEnd`InputNotebook[], #, Placeholder]}]& ),
ShowSelection->True,
InputAliases->{"intt" -> RowBox[{"\[Integral]", 
     RowBox[{"\[SelectionPlaceholder]", 
       RowBox[{"\[DifferentialD]", "\[Placeholder]"}]}]}], "dintt" -> RowBox[{
     SubsuperscriptBox[
     "\[Integral]", "\[SelectionPlaceholder]", "\[Placeholder]"], 
     RowBox[{"\[Placeholder]", 
       RowBox[{"\[DifferentialD]", "\[Placeholder]"}]}]}], "sumt" -> RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"\[SelectionPlaceholder]", "=", "\[Placeholder]"}], 
      "\[Placeholder]"], "\[Placeholder]"}], "prodt" -> RowBox[{
     UnderoverscriptBox["\[Product]", 
      RowBox[{"\[SelectionPlaceholder]", "=", "\[Placeholder]"}], 
      "\[Placeholder]"], "\[Placeholder]"}], "dt" -> RowBox[{
     SubscriptBox["\[PartialD]", "\[Placeholder]"], " ", 
     "\[SelectionPlaceholder]"}], "notation" -> RowBox[{"Notation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        " ", "\[DoubleLongLeftRightArrow]", " ", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "notation>" -> 
  RowBox[{"Notation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        " ", "\[DoubleLongRightArrow]", " ", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "notation<" -> 
  RowBox[{"Notation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        " ", "\[DoubleLongLeftArrow]", " ", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "symb" -> RowBox[{"Symbolize", "[", 
     TagBox[
     "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"], 
     "]"}], "infixnotation" -> RowBox[{"InfixNotation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        ",", "\[Placeholder]"}], "]"}], "addia" -> 
  RowBox[{"AddInputAlias", "[", 
     RowBox[{"\"\[Placeholder]\"", "\[Rule]", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "pattwraper" -> 
  TagBox["\[Placeholder]", NotationPatternTag, TagStyle -> 
    "NotationPatternWrapperStyle"], "madeboxeswraper" -> 
  TagBox["\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
    "NotationMadeBoxesWrapperStyle"]},
FrontEndVersion->"10.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (September 9, \
2014)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Title"], 
      Cell[
      "Modify the definitions below to change the default appearance of all \
cells in a given style.  Make modifications to any definition using commands \
in the Format menu.", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {
           "Color" -> RGBColor[0.269993, 0.308507, 0.6]}, 
          CellLabelMargins -> {{12, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          CellLabelMargins -> {{24, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 12], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellLabelMargins -> {{8, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 8], 
         Cell[
          StyleData[All, "SlideShow"], PageWidth -> WindowWidth, 
          ScrollingOptions -> {
           "PagewiseDisplay" -> True, "VerticalScrollRange" -> Fit}, 
          ShowCellBracket -> False, ScriptMinSize -> 9], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, 
          CellLabelMargins -> {{2, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 5, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "The options defined for the style below will be used at the \
Notebook level.", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, CellFrameLabelMargins -> 
          6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, Inherited}, {10, 30}}, 
             CellGroupingRules -> {"TitleGrouping", 0}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LineSpacing -> {1, 11}, LanguageCategory -> 
             "NaturalLanguage", CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 36, FontWeight -> "Bold"], 
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{27, 10}, {10, 30}}, LineSpacing -> {1, 0}, 
             FontSize -> 44], 
            Cell[
             StyleData["Title", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, FontSize -> 20], 
            Cell[
             StyleData["Title", "SlideShow"]], 
            Cell[
             StyleData["Title", "Printout"], 
             CellMargins -> {{2, 10}, {12, 30}}, FontSize -> 24]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, Inherited}, {20, 2}},
              CellGroupingRules -> {"TitleGrouping", 10}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 24], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{27, 10}, {20, 2}}, LineSpacing -> {1, 0}, 
             FontSize -> 36], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 4}}, FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "SlideShow"]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{2, 10}, {12, 8}}, FontSize -> 18]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], 
             CellMargins -> {{27, Inherited}, {8, 2}}, 
             CellGroupingRules -> {"TitleGrouping", 20}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 16], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{24, 10}, {20, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {8, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "SlideShow"]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{2, 10}, {12, 8}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{0, 0}, {0, 1}}, 
             CellMargins -> {{27, Inherited}, {8, 34}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 4, DefaultNewInlineCellStyle -> 
             "None", InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LineSpacing -> {1, 2}, LanguageCategory -> 
             "NaturalLanguage", CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 20, FontWeight -> "Bold"], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{40, 10}, {11, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{18, Inherited}, {6, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "SlideShow"]], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{2, 0}, {7, 22}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> "\[FilledSmallSquare]", 
             CellMargins -> {{60, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 14, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{36, 10}, {11, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{16, Inherited}, {6, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "SlideShow"]], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {7, 22}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> 
             "\[FilledSmallSquare]", 
             CellMargins -> {{60, Inherited}, {2, 10}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subsubsection", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{34, 10}, {11, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{17, Inherited}, {6, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "SlideShow"]], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{9, 0}, {7, 14}}, FontSize -> 11]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{60, 10}, {7, 7}}, 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LineSpacing -> {1, 3}, CounterIncrements -> 
             "Text"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{24, 10}, {10, 10}}, LineSpacing -> {1, 5}, 
             FontSize -> 16], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {6, 6}},
              LineSpacing -> {1, 1}, FontSize -> 11], 
            Cell[
             StyleData["Text", "SlideShow"]], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{2, 2}, {6, 6}}, 
             TextJustification -> 0.5, Hyphenation -> True, FontSize -> 10]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{60, 10}, {6, 6}}, 
             DefaultNewInlineCellStyle -> "None", LineSpacing -> {1, 3}, 
             LanguageCategory -> "NaturalLanguage", CounterIncrements -> 
             "SmallText", FontFamily -> "Helvetica", FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{24, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {5, 5}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "SlideShow"]], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{2, 2}, {5, 5}}, TextJustification -> 0.5, 
             Hyphenation -> True, FontSize -> 7]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names. Some attributes for these styles are actually set in FormatType Styles \
(in the last section of this stylesheet). ", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellMargins -> {{66, 10}, {5, 7}}, 
             Evaluatable -> True, CellGroupingRules -> "InputGrouping", 
             CellHorizontalScrolling -> True, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, DefaultFormatType -> 
             DefaultInputFormatType, "TwoByteSyntaxCharacterAutoReplacement" -> 
             True, HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> "Mathematica", 
             FormatType -> InputForm, ShowStringCharacters -> True, 
             NumberMarks -> True, LinebreakAdjustments -> {0.85, 2, 10, 0, 1},
              CounterIncrements -> "Input", FontWeight -> "Bold"], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{72, Inherited}, {8, 10}}, LineSpacing -> {1, 0},
              FontSize -> 16], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {2, 3}}, FontSize -> 11], 
            Cell[
             StyleData["Input", "SlideShow"]], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{39, 0}, {4, 6}},
              LinebreakAdjustments -> {0.85, 2, 10, 1, 1}, FontSize -> 9]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["InputOnly"], CellMargins -> {{66, 10}, {7, 7}}, 
             Evaluatable -> True, CellGroupingRules -> "InputGrouping", 
             CellHorizontalScrolling -> True, DefaultFormatType -> 
             DefaultInputFormatType, "TwoByteSyntaxCharacterAutoReplacement" -> 
             True, HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> "Mathematica", 
             FormatType -> InputForm, ShowStringCharacters -> True, 
             NumberMarks -> True, LinebreakAdjustments -> {0.85, 2, 10, 0, 1},
              CounterIncrements -> "Input", StyleMenuListing -> None, 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["InputOnly", "Presentation"], 
             CellMargins -> {{72, Inherited}, {8, 10}}, LineSpacing -> {1, 0},
              FontSize -> 16], 
            Cell[
             StyleData["InputOnly", "Condensed"], 
             CellMargins -> {{40, 10}, {2, 3}}, FontSize -> 11], 
            Cell[
             StyleData["InputOnly", "SlideShow"]], 
            Cell[
             StyleData["InputOnly", "Printout"], 
             CellMargins -> {{39, 0}, {4, 6}}, 
             LinebreakAdjustments -> {0.85, 2, 10, 1, 1}, FontSize -> 9]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellMargins -> {{66, 10}, {7, 5}}, 
             CellEditDuplicate -> True, CellGroupingRules -> "OutputGrouping",
              CellHorizontalScrolling -> True, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, DefaultFormatType -> 
             DefaultOutputFormatType, "TwoByteSyntaxCharacterAutoReplacement" -> 
             True, HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> None, FormatType -> 
             InputForm, CounterIncrements -> "Output"], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{72, Inherited}, {10, 8}}, LineSpacing -> {1, 0},
              FontSize -> 16], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{41, Inherited}, {3, 2}}, FontSize -> 11], 
            Cell[
             StyleData["Output", "SlideShow"]], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{39, 0}, {6, 4}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{66, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> None, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontFamily -> "Helvetica", FontSize -> 10, FontColor -> 
             RGBColor[0.6, 0.100008, 0.100008]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{72, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}, FontSize -> 16], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{41, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 11], 
            Cell[
             StyleData["Message", "SlideShow"]], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{39, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 7, FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{66, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> None, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{72, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}, FontSize -> 16], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{41, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 11], 
            Cell[
             StyleData["Print", "SlideShow"]], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{39, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 8]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{4, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, LanguageCategory -> 
             None, FormatType -> InputForm, CounterIncrements -> "Graphics", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             StyleMenuListing -> None, FontFamily -> "Courier", FontSize -> 
             10], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             ImageMargins -> {{62, Inherited}, {Inherited, 0}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             ImageMargins -> {{38, Inherited}, {Inherited, 0}}, Magnification -> 
             0.6], 
            Cell[
             StyleData["Graphics", "SlideShow"]], 
            Cell[
             StyleData["Graphics", "Printout"], 
             ImageMargins -> {{30, Inherited}, {Inherited, 0}}, Magnification -> 
             0.8]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], LanguageCategory -> None, 
             StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 
             9, FontColor -> RGBColor[0.269993, 0.308507, 0.6]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 9], 
            Cell[
             StyleData["CellLabel", "SlideShow"]], 
            Cell[
             StyleData["CellLabel", "Printout"], FontFamily -> "Courier", 
             FontSize -> 8, FontSlant -> "Italic", FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["FrameLabel"], LanguageCategory -> None, 
             StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 
             9], 
            Cell[
             StyleData["FrameLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["FrameLabel", "Condensed"], FontSize -> 9], 
            Cell[
             StyleData["FrameLabel", "SlideShow"]], 
            Cell[
             StyleData["FrameLabel", "Printout"], FontFamily -> "Courier", 
             FontSize -> 8, FontSlant -> "Italic", FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Inline Formatting", "Section"], 
         Cell[
         "These styles are for modifying individual words or letters in a \
cell exclusive of the cell tag.", "Text"], 
         Cell[
          StyleData["RM"], StyleMenuListing -> None, FontWeight -> "Plain", 
          FontSlant -> "Plain"], 
         Cell[
          StyleData["BF"], StyleMenuListing -> None, FontWeight -> "Bold"], 
         Cell[
          StyleData["IT"], StyleMenuListing -> None, FontSlant -> "Italic"], 
         Cell[
          StyleData["TR"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Plain", FontSlant -> "Plain"], 
         Cell[
          StyleData["TI"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Plain", FontSlant -> "Italic"], 
         Cell[
          StyleData["TB"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Bold", FontSlant -> "Plain"], 
         Cell[
          StyleData["TBI"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Bold", FontSlant -> "Italic"], 
         Cell[
          StyleData["MR"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Plain", FontSlant -> "Plain"], 
         Cell[
          StyleData["MO"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Plain", FontSlant -> "Italic"], 
         Cell[
          StyleData["MB"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Bold", FontSlant -> "Plain"], 
         Cell[
          StyleData["MBO"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Bold", FontSlant -> "Italic"], 
         Cell[
          StyleData["SR"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Plain", FontSlant -> "Plain"], 
         Cell[
          StyleData["SO"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Plain", FontSlant -> "Italic"], 
         Cell[
          StyleData["SB"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Bold", FontSlant -> "Plain"], 
         Cell[
          StyleData["SBO"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Bold", FontSlant -> "Italic"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SO10"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 10, FontWeight -> "Plain", FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["SO10", "Printout"], StyleMenuListing -> None, 
             FontFamily -> "Helvetica", FontSize -> 7, FontWeight -> "Plain", 
             FontSlant -> "Italic"]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Inert"], StyleMenuListing -> None, Background -> 
             RGBColor[0.870588, 0.905882, 0.972549]], 
            Cell[
             StyleData["Inert", "Printout"], StyleMenuListing -> None, 
             Background -> GrayLevel[1]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["InlineFormula"], CellMargins -> {{10, 4}, {0, 8}}, 
             CellHorizontalScrolling -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Formula", ScriptLevel -> 1, SingleLetterItalics -> True], 
            Cell[
             StyleData["InlineFormula", "Presentation"], 
             CellMargins -> {{24, 10}, {10, 10}}, LineSpacing -> {1, 5}, 
             FontSize -> 16], 
            Cell[
             StyleData["InlineFormula", "Condensed"], 
             CellMargins -> {{8, 10}, {6, 6}}, LineSpacing -> {1, 1}, 
             FontSize -> 11], 
            Cell[
             StyleData["InlineFormula", "SlideShow"]], 
            Cell[
             StyleData["InlineFormula", "Printout"], 
             CellMargins -> {{2, 0}, {6, 6}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], 
             CellMargins -> {{60, Inherited}, {Inherited, Inherited}}, 
             CellHorizontalScrolling -> True, DefaultFormatType -> 
             DefaultInputFormatType, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Formula", ScriptLevel -> 0, SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             LineSpacing -> {1, 5}, FontSize -> 16], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], LineSpacing -> {1, 1}, 
             FontSize -> 11], 
            Cell[
             StyleData["DisplayFormula", "SlideShow"]], 
            Cell[
             StyleData["DisplayFormula", "Printout"], FontSize -> 10]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
             CellMargins -> {{60, 4}, {0, 8}}, CellHorizontalScrolling -> 
             True, Hyphenation -> False, LanguageCategory -> "Formula", 
             ScriptLevel -> 1, FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{24, 10}, {10, 10}}, LineSpacing -> {1, 5}, 
             FontSize -> 16], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {6, 6}}, LineSpacing -> {1, 1}, 
             FontSize -> 11], 
            Cell[
             StyleData["Program", "SlideShow"]], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{2, 0}, {6, 6}}, FontSize -> 9]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Outline Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline1"], CellMargins -> {{60, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, 
             ParagraphIndent -> -38, CounterIncrements -> "Outline1", 
             CounterAssignments -> {{"Outline2", 0}, {"Outline3", 0}, {
               "Outline4", 0}}, FontSize -> 18, FontWeight -> "Bold", 
             CounterBoxOptions -> {
              CounterFunction :> 
               Utilities`Notation`Private`CapitalRomanNumeral}], 
            Cell[
             StyleData["Outline1", "SlideShow"]], 
            Cell[
             StyleData["Outline1", "Printout"], 
             CounterBoxOptions -> {
              CounterFunction :> 
               Utilities`Notation`Private`CapitalRomanNumeral}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline2"], CellMargins -> {{90, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 60}, 
             ParagraphIndent -> -27, CounterIncrements -> "Outline2", 
             CounterAssignments -> {{"Outline3", 0}, {"Outline4", 0}}, 
             FontSize -> 15, FontWeight -> "Bold", 
             CounterBoxOptions -> {CounterFunction :> (Part[
                 CharacterRange["A", "Z"], #]& )}], 
            Cell[
             StyleData["Outline2", "SlideShow"]], 
            Cell[
             StyleData["Outline2", "Printout"], 
             CounterBoxOptions -> {CounterFunction :> (Part[
                 CharacterRange["A", "Z"], #]& )}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline3"], CellMargins -> {{120, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 70}, 
             ParagraphIndent -> -21, CounterIncrements -> "Outline3", 
             CounterAssignments -> {{"Outline4", 0}}, FontSize -> 12, 
             CounterBoxOptions -> {CounterFunction :> Identity}], 
            Cell[
             StyleData["Outline3", "SlideShow"]], 
            Cell[
             StyleData["Outline3", "Printout"], 
             CounterBoxOptions -> {CounterFunction :> Identity}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline4"], CellMargins -> {{150, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 80}, 
             ParagraphIndent -> -18, CounterIncrements -> "Outline4", 
             FontSize -> 10, CounterBoxOptions -> {CounterFunction :> (Part[
                 CharacterRange["a", "z"], #]& )}], 
            Cell[
             StyleData["Outline4", "SlideShow"]], 
            Cell[
             StyleData["Outline4", "Printout"]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["Hyperlink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["Hyperlink", "SlideShow"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
         "The following styles are for linking automatically to the on-line \
help system.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["MainBookLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["MainBookLink", "SlideShow"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["AddOnsLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["AddOnsLink", "SlideShow"]], 
            Cell[
             StyleData["AddOnsLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuide", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["RefGuideLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["RefGuideLink", "SlideShow"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLinkText"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuide", #]}]& )}], 
            Cell[
             StyleData["RefGuideLinkText", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["RefGuideLinkText", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["RefGuideLinkText", "SlideShow"]], 
            Cell[
             StyleData["RefGuideLinkText", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["GettingStartedLink", "SlideShow"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DemosLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["Demos", #]}]& )}], 
            Cell[
             StyleData["DemosLink", "SlideShow"]], 
            Cell[
             StyleData["DemosLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["TourLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["Tour", #]}]& )}], 
            Cell[
             StyleData["TourLink", "SlideShow"]], 
            Cell[
             StyleData["TourLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"], FontSize -> 
             16], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["OtherInformationLink", "SlideShow"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MasterIndexLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MasterIndex", #]}]& )}], 
            Cell[
             StyleData["MasterIndexLink", "SlideShow"]], 
            Cell[
             StyleData["MasterIndexLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          DefaultNewInlineCellStyle -> "None", LanguageCategory -> 
          "NaturalLanguage", StyleMenuListing -> None, FontSize -> 10, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          DefaultNewInlineCellStyle -> "None", LanguageCategory -> 
          "NaturalLanguage", StyleMenuListing -> None, FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Times", FontSize -> 10]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Palette Styles", "Section"], 
         Cell[
         "The cells below define styles that define standard ButtonFunctions, \
for use in palette buttons.", "Text"], 
         Cell[
          StyleData["Paste"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, Placeholder]}]& )}], 
         Cell[
          StyleData["Evaluate"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionEvaluate[
                FrontEnd`InputNotebook[], All]}]& )}], 
         Cell[
          StyleData["EvaluateCell"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionMove[
                FrontEnd`InputNotebook[], All, Cell, 1], 
               FrontEnd`SelectionEvaluateCreateCell[
                FrontEnd`InputNotebook[], All]}]& )}], 
         Cell[
          StyleData["CopyEvaluate"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`SelectionCreateCell[
                FrontEnd`InputNotebook[], All], 
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionEvaluate[
                FrontEnd`InputNotebook[], All]}]& )}], 
         Cell[
          StyleData["CopyEvaluateCell"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`SelectionCreateCell[
                FrontEnd`InputNotebook[], All], 
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionEvaluateCreateCell[
                FrontEnd`InputNotebook[], All]}]& )}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Placeholder Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making placeholder objects \
in palette templates.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Placeholder"], Placeholder -> True, StyleMenuListing -> 
             None, FontSlant -> "Italic", FontColor -> 
             RGBColor[0.890623, 0.864698, 0.384756], 
             TagBoxOptions -> {
              Editable -> False, Selectable -> False, StripWrapperBoxes -> 
               False}], 
            Cell[
             StyleData["Placeholder", "Presentation"]], 
            Cell[
             StyleData["Placeholder", "Condensed"]], 
            Cell[
             StyleData["Placeholder", "SlideShow"]], 
            Cell[
             StyleData["Placeholder", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["PrimaryPlaceholder"], StyleMenuListing -> None, 
             DrawHighlighted -> True, FontSlant -> "Italic", Background -> 
             RGBColor[0.912505, 0.891798, 0.507774], 
             TagBoxOptions -> {
              Editable -> False, Selectable -> False, StripWrapperBoxes -> 
               False}], 
            Cell[
             StyleData["PrimaryPlaceholder", "Presentation"]], 
            Cell[
             StyleData["PrimaryPlaceholder", "Condensed"]], 
            Cell[
             StyleData["PrimaryPlaceholder", "SlideShow"]], 
            Cell[
             StyleData["PrimaryPlaceholder", "Printout"]]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["FormatType Styles", "Section"], 
         Cell[
         "The cells below define styles that are mixed in with the styles of \
most cells.  If a cell's FormatType matches the name of one of the styles \
defined below, then that style is applied between the cell's style and its \
own options. This is particularly true of Input and Output.", "Text"], 
         Cell[
          StyleData["CellExpression"], PageWidth -> Infinity, 
          CellMargins -> {{6, Inherited}, {Inherited, Inherited}}, 
          ShowCellLabel -> False, ShowSpecialCharacters -> False, 
          AllowInlineCells -> False, Hyphenation -> False, 
          AutoItalicWords -> {}, StyleMenuListing -> None, FontFamily -> 
          "Courier", FontSize -> 12, Background -> GrayLevel[1]], 
         Cell[
          StyleData["InputForm"], InputAutoReplacements -> {}, 
          AllowInlineCells -> False, Hyphenation -> False, StyleMenuListing -> 
          None, FontFamily -> "Courier"], 
         Cell[
          StyleData["OutputForm"], PageWidth -> Infinity, TextAlignment -> 
          Left, LineSpacing -> {0.6, 1}, StyleMenuListing -> None, FontFamily -> 
          "Courier"], 
         Cell[
          StyleData["StandardForm"], 
          InputAutoReplacements -> {
           "->" -> "\[Rule]", ":>" -> "\[RuleDelayed]", "<=" -> 
            "\[LessEqual]", ">=" -> "\[GreaterEqual]", "!=" -> "\[NotEqual]", 
            "==" -> "\[Equal]", Inherited}, 
          "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          LineSpacing -> {1.25, 0}, StyleMenuListing -> None, FontFamily -> 
          "Courier"], 
         Cell[
          StyleData["TraditionalForm"], 
          InputAutoReplacements -> {
           "->" -> "\[Rule]", ":>" -> "\[RuleDelayed]", "<=" -> 
            "\[LessEqual]", ">=" -> "\[GreaterEqual]", "!=" -> "\[NotEqual]", 
            "==" -> "\[Equal]", Inherited}, 
          "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          LineSpacing -> {1.25, 0}, SingleLetterItalics -> True, 
          TraditionalFunctionNotation -> True, DelimiterMatching -> None, 
          StyleMenuListing -> None], 
         Cell[
         "The style defined below is mixed in to any cell that is in an \
inline cell within another.", "Text"], 
         Cell[
          StyleData["InlineCell"], LanguageCategory -> "Formula", ScriptLevel -> 
          1, StyleMenuListing -> None], 
         Cell[
          StyleData["InlineCellEditing"], StyleMenuListing -> None, 
          Background -> RGBColor[0.964706, 0.929412, 0.839216]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Automatic Styles", "Section"], 
         Cell[
         "The cells below define styles that are used to affect the display \
of certain types of objects in typeset expressions.  For example, \
\"UnmatchedBracket\" style defines how unmatched bracket, curly bracket, and \
parenthesis characters are displayed (typically by coloring them to make them \
stand out).", "Text"], 
         Cell[
          StyleData["UnmatchedBracket"], StyleMenuListing -> None, FontColor -> 
          RGBColor[0.760006, 0.330007, 0.8]], 
         Cell[
          StyleData["Completions"], StyleMenuListing -> None, FontFamily -> 
          "Courier"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles from HelpBrowser", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MathCaption"], CellFrame -> {{0, 0}, {0, 0.5}}, 
             CellMargins -> {{66, 12}, {2, 24}}, PageBreakBelow -> False, 
             CellFrameMargins -> {{8, 8}, {8, 2}}, CellFrameColor -> 
             GrayLevel[0.700008], CellFrameLabelMargins -> 4, 
             LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 8}, 
             StyleMenuListing -> None, FontColor -> GrayLevel[0.2]], 
            Cell[
             StyleData["MathCaption", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["MathCaption", "SlideShow"]], 
            Cell[
             StyleData["MathCaption", "Printout"], 
             CellMargins -> {{39, 0}, {0, 14}}, Hyphenation -> True, FontSize -> 
             9, FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ObjectName"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["ObjectName", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["ObjectName", "SlideShow"]], 
            Cell[
             StyleData["ObjectName", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Usage"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["Usage", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["Usage", "SlideShow"]], 
            Cell[
             StyleData["Usage", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Notes"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["Notes", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["Notes", "SlideShow"]], 
            Cell[
             StyleData["Notes", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["InlineOutput"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["InlineOutput", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["InlineOutput", "SlideShow"]], 
            Cell[
             StyleData["InlineOutput", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell["Emphasis Boxes and Pictures", "Subsection"], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["Box"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["Box", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["Box", "SlideShow"]], 
               Cell[
                StyleData["Box", "Printout"], CellMargins -> {{2, 0}, {0, 8}},
                 FontSize -> 10, Background -> GrayLevel[0.900008]]}, 
              Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DoubleBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Top}}, "RowsIndexed" -> {}}, 
                  GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[1.4]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DoubleBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["DoubleBox", "SlideShow"]], 
               Cell[
                StyleData["DoubleBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["1ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["1ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["1ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["1ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 SingleLetterItalics -> False, LineIndent -> 0, 
                StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.67]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["2ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnEvenBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.46]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnEvenBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnEvenBox", "SlideShow"]], 
               Cell[
                StyleData["2ColumnEvenBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnSmallBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {Right, {Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.35]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[1.0499999999999998`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnSmallBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnSmallBox", "SlideShow"]], 
               Cell[
                StyleData["2ColumnSmallBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["3ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.32]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["3ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["3ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["3ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["3ColumnSmallBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {Right, Center, {Left}}, 
                    "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
                    "RowsIndexed" -> {}}, GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.24]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[1.0499999999999998`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["3ColumnSmallBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["3ColumnSmallBox", "SlideShow"]], 
               Cell[
                StyleData["3ColumnSmallBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["4ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 SingleLetterItalics -> False, LineIndent -> 0, 
                StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.13], 
                    Scaled[0.35], 
                    Scaled[0.13], {
                    Scaled[0.35]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["4ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["4ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["4ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["5ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.202]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}},
                     "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["5ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["5ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["5ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["6ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.12], 
                    Scaled[0.22], 
                    Scaled[0.12], 
                    Scaled[0.12], 
                    Scaled[0.22], {
                    Scaled[0.12]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["6ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["6ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["6ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["FramedBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["FramedBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["FramedBox", "SlideShow"]], 
               Cell[
                StyleData["FramedBox", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.4], {
                    Scaled[0.6]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["DefinitionBox", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox3Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.35], 
                    Scaled[0.2], {
                    Scaled[0.45]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox3Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox3Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox3Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox4Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.15], 
                    Scaled[0.35], 
                    Scaled[0.15], {
                    Scaled[0.35]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox4Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox4Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox4Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox5Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.2]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox5Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox5Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox5Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox6Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.13], 
                    Scaled[0.24], 
                    Scaled[0.13], 
                    Scaled[0.13], 
                    Scaled[0.24], {
                    Scaled[0.13]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox6Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox6Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox6Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["TopBox"], CellFrame -> {{0.5, 0.5}, {0, 0.5}}, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakBelow -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.62]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["TopBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["TopBox", "SlideShow"]], 
               Cell[
                StyleData["TopBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, Background -> GrayLevel[1]]},
               Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["MiddleBox"], CellFrame -> {{0.5, 0.5}, {0, 0}}, 
                CellMargins -> {{27, 12}, {0, -7}}, CellHorizontalScrolling -> 
                True, PageBreakAbove -> False, PageBreakBelow -> False, 
                CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941], 
                AutoIndent -> False, AutoSpacing -> False, LineIndent -> 0, 
                StyleMenuListing -> None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.62]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["MiddleBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["MiddleBox", "SlideShow"]], 
               Cell[
                StyleData["MiddleBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 2}}, Background -> GrayLevel[1]]},
               Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["BottomBox"], CellFrame -> {{0.5, 0.5}, {0.5, 0}}, 
                CellMargins -> {{27, 12}, {0, -7}}, CellHorizontalScrolling -> 
                True, PageBreakAbove -> False, PageBreakBelow -> False, 
                CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941], 
                AutoIndent -> False, AutoSpacing -> False, LineIndent -> 0, 
                StyleMenuListing -> None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.62]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["BottomBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["BottomBox", "SlideShow"]], 
               Cell[
                StyleData["BottomBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, -5}}, FontSize -> 10, Background -> 
                GrayLevel[1], 
                GridBoxOptions -> {
                 GridBoxItemSize -> {
                   "Columns" -> {{All}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{2.2}}, "RowsIndexed" -> {}}}]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["TopSpanBox"], CellFrame -> {{0.5, 0.5}, {0, 0.5}}, 
                CellMargins -> {{27, 12}, {-2, 8}}, CellHorizontalScrolling -> 
                True, PageBreakBelow -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.9], {
                    Scaled[0.03]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["TopSpanBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["TopSpanBox", "SlideShow"]], 
               Cell[
                StyleData["TopSpanBox", "Printout"], 
                CellMargins -> {{2, 0}, {-2, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["MiddleSpanBox"], CellFrame -> {{0.5, 0.5}, {0, 0}},
                 CellMargins -> {{27, 12}, {0, 0}}, CellHorizontalScrolling -> 
                True, PageBreakAbove -> False, PageBreakBelow -> False, 
                CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941], 
                AutoIndent -> False, AutoSpacing -> False, LineIndent -> 0, 
                StyleMenuListing -> None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.9], {
                    Scaled[0.03]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["MiddleSpanBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["MiddleSpanBox", "SlideShow"]], 
               Cell[
                StyleData["MiddleSpanBox", "Printout"], 
                CellMargins -> {{2, 0}, {-5, 0}}, FontSize -> 10, Background -> 
                GrayLevel[1], 
                GridBoxOptions -> {
                 GridBoxItemSize -> {
                   "Columns" -> {{All}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{1.8}}, "RowsIndexed" -> {}}}]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["Picture"], 
                CellMargins -> {{27, Inherited}, {4, 4}}, CellGroupingRules -> 
                "GraphicsGrouping", CellHorizontalScrolling -> True, 
                StyleMenuListing -> None], 
               Cell[
                StyleData["Picture", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["Picture", "SlideShow"]], 
               Cell[
                StyleData["Picture", "Printout"], 
                CellMargins -> {{2, Inherited}, {4, 4}}, Magnification -> 
                0.65]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["OpenCloseItemizedPicture"], 
                CellMargins -> {{88, 4}, {4, 4}}, CellGroupingRules -> 
                "GraphicsGrouping", CellHorizontalScrolling -> True, 
                StyleMenuListing -> None, 
                PrivateCellOptions -> {"DefaultCellGroupOpen" -> False}], 
               Cell[
                StyleData["OpenCloseItemizedPicture", "Presentation"], 
                FontSize -> 18], 
               Cell[
                StyleData["OpenCloseItemizedPicture", "SlideShow"]], 
               Cell[
                StyleData["OpenCloseItemizedPicture", "Printout"], 
                CellMargins -> {{76, 2}, {0, 0}}, 
                CellElementSpacings -> {
                 "CellMinHeight" -> 1, "ClosedCellHeight" -> 0}, CellOpen -> 
                False]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["ItemizedPicture"], 
                CellMargins -> {{88, 4}, {4, 4}}, CellGroupingRules -> 
                "GraphicsGrouping", CellHorizontalScrolling -> True, 
                StyleMenuListing -> None], 
               Cell[
                StyleData["ItemizedPicture", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["ItemizedPicture", "SlideShow"]], 
               Cell[
                StyleData["ItemizedPicture", "Printout"], 
                CellMargins -> {{77, 2}, {4, -4}}, Magnification -> 0.5]}, 
              Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["ListGraphic"], CellMargins -> {{88, 4}, {4, 4}}, 
                CellGroupingRules -> "GraphicsGrouping", 
                CellHorizontalScrolling -> True, StyleMenuListing -> None], 
               Cell[
                StyleData["ListGraphic", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["ListGraphic", "SlideShow"]], 
               Cell[
                StyleData["ListGraphic", "Printout"], 
                CellMargins -> {{77, 2}, {4, -4}}, Magnification -> 0.5]}, 
              Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["ListNoteBox"], CellFrame -> 0.5, 
                CellMargins -> {{88, 12}, {8, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {FrameMargins -> {{1, 1}, {1.5, 1.5}}}, 
                GridBoxOptions -> {GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["ListNoteBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["ListNoteBox", "SlideShow"]], 
               Cell[
                StyleData["ListNoteBox", "Printout"], 
                CellMargins -> {{77, 4}, {6, 2}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["PictureGroup"], CellMargins -> {{41, 4}, {0, 4}}, 
                CellGroupingRules -> "GraphicsGrouping", 
                CellHorizontalScrolling -> True, StyleMenuListing -> None], 
               Cell[
                StyleData["PictureGroup", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["PictureGroup", "SlideShow"]], 
               Cell[
                StyleData["PictureGroup", "Printout"], 
                CellMargins -> {{76, 2}, {0, 0}}, 
                CellElementSpacings -> {
                 "CellMinHeight" -> 1, "ClosedCellHeight" -> 0}, CellOpen -> 
                False]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["Sound"], ShowCellBracket -> True, 
                CellMargins -> {{27, Inherited}, {0, 8}}, StyleMenuListing -> 
                None], 
               Cell[
                StyleData["Sound", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["Sound", "SlideShow"]], 
               Cell[
                StyleData["Sound", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10]}, Closed]]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell["Tables", "Subsection"], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnTable"], CellMargins -> {{35, 4}, {0, 8}}, 
                CellHorizontalScrolling -> True, LineIndent -> 0, 
                StyleMenuListing -> None, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.34], {
                    Scaled[0.64]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnTable", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnTable", "SlideShow"]], 
               Cell[
                StyleData["2ColumnTable", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnEvenTable"], 
                CellMargins -> {{35, 4}, {0, 8}}, CellHorizontalScrolling -> 
                True, LineIndent -> 0, StyleMenuListing -> None, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.49]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnEvenTable", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["2ColumnEvenTable", "SlideShow"]], 
               Cell[
                StyleData["2ColumnEvenTable", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["3ColumnTable"], CellMargins -> {{35, 4}, {0, 8}}, 
                CellHorizontalScrolling -> True, LineIndent -> 0, 
                StyleMenuListing -> None, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.28], 
                    Scaled[0.28], {
                    Scaled[0.43]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["3ColumnTable", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["3ColumnTable", "SlideShow"]], 
               Cell[
                StyleData["3ColumnTable", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9]}, Closed]]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Slide Show Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideShowNavigationBar"], Editable -> False, 
             CellFrame -> True, CellMargins -> {{0, 0}, {3, 3}}, 
             CellElementSpacings -> {"CellMinHeight" -> 0.8125}, 
             CellGroupingRules -> {"SectionGrouping", 30}, CellFrameMargins -> 
             False, CellFrameColor -> GrayLevel[1], CellFrameLabelMargins -> 
             False, TextAlignment -> Center, CounterIncrements -> 
             "SlideShowNavigationBar", StyleMenuListing -> None, FontSize -> 
             10, Magnification -> 1, Background -> GrayLevel[0.8], 
             GridBoxOptions -> {
              BaselinePosition -> Center, 
               GridBoxAlignment -> {
                "Columns" -> {
                  Center, Center, Center, Center, Center, Center, Right, {
                   Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
                 "RowsIndexed" -> {}}, 
               GridBoxItemSize -> {
                "Columns" -> {3.5, 3.5, 3.5, 3.5, 13, 5, {4}}, 
                 "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                 "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}}], 
            Cell[
             StyleData["SlideShowNavigationBar", "Presentation"]], 
            Cell[
             StyleData["SlideShowNavigationBar", "SlideShow"], Deletable -> 
             False, ShowCellBracket -> False, 
             CellMargins -> {{-1, -1}, {-1, -1}}, PageBreakAbove -> True, 
             CellFrameMargins -> {{1, 1}, {0, 0}}], 
            Cell[
             StyleData["SlideShowNavigationBar", "Printout"], 
             CellMargins -> {{18, 4}, {4, 4}}, LineSpacing -> {1, 3}, 
             FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideShowSection"], CellFrame -> {{0, 0}, {0, 0.5}}, 
             CellMargins -> {{0, 0}, {10, 0}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CellFrameMargins -> {{12, 4}, {6, 12}}, 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 
             18, FontWeight -> "Plain", FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.408011, 0.440726, 0.8]], 
            Cell[
             StyleData["SlideShowSection", "Presentation"], 
             CellFrameMargins -> {{20, 10}, {10, 18}}, FontSize -> 27], 
            Cell[
             StyleData["SlideShowSection", "SlideShow"], ShowCellBracket -> 
             False, PageBreakAbove -> True], 
            Cell[
             StyleData["SlideShowSection", "Printout"], 
             CellMargins -> {{18, 30}, {0, 30}}, CellFrameMargins -> 5, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideHyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontSize -> 26, FontColor -> 
             GrayLevel[0.400015], 
             ButtonBoxOptions -> {
              Active -> True, ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonMargins -> 0.5, 
               ButtonMinHeight -> 0.85, ButtonNote -> None}], 
            Cell[
             StyleData["SlideHyperlink", "Presentation"], 
             CellMargins -> {{14, 10}, {6, 12}}, FontSize -> 36], 
            Cell[
             StyleData["SlideHyperlink", "SlideShow"]], 
            Cell[
             StyleData["SlideHyperlink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideTOCLink"], 
             CellMargins -> {{24, Inherited}, {Inherited, Inherited}}, 
             StyleMenuListing -> None, ButtonStyleMenuListing -> Automatic, 
             FontFamily -> "Helvetica", 
             ButtonBoxOptions -> {
              Active -> True, ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonMargins -> 1.5, 
               ButtonNote -> ButtonData}], 
            Cell[
             StyleData["SlideTOCLink", "Presentation"], 
             CellMargins -> {{35, 10}, {8, 8}}, FontSize -> 18], 
            Cell[
             StyleData["SlideTOCLink", "SlideShow"]], 
            Cell[
             StyleData["SlideTOCLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideTOC"], CellDingbat -> "\[Bullet]", 
             CellMargins -> {{18, Inherited}, {Inherited, Inherited}}, 
             StyleMenuListing -> None, FontFamily -> "Helvetica"], 
            Cell[
             StyleData["SlideTOC", "Presentation"], 
             CellMargins -> {{25, 10}, {10, 5}}, FontSize -> 18], 
            Cell[
             StyleData["SlideTOC", "SlideShow"], FontSize -> 14], 
            Cell[
             StyleData["SlideTOC", "Printout"], FontSize -> 10, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
      "Notation Package Styles", "Section", CellTags -> "NotationPackage"], 
      Cell[
      "The cells below define certain styles needed by the Notation package. \
These styles serve to make visible otherwise invisible tagboxes.", "Text", 
       CellTags -> "NotationPackage"], 
      Cell[
       StyleData["NotationTemplateStyle"], StyleMenuListing -> None, 
       Background -> RGBColor[1, 1, 0.850004], 
       TagBoxOptions -> {SyntaxForm -> "symbol"}, CellTags -> 
       "NotationPackage"], 
      Cell[
       StyleData["NotationPatternWrapperStyle"], StyleMenuListing -> None, 
       Background -> RGBColor[1, 0.900008, 0.979995], 
       TagBoxOptions -> {SyntaxForm -> "symbol"}, CellTags -> 
       "NotationPackage"], 
      Cell[
       StyleData["NotationMadeBoxesWrapperStyle"], StyleMenuListing -> None, 
       Background -> RGBColor[0.900008, 0.889998, 1], 
       TagBoxOptions -> {SyntaxForm -> "symbol"}, CellTags -> 
       "NotationPackage"]}, Closed]]}, Visible -> False, FrontEndVersion -> 
  "10.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (September 9, 2014)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 74, 1, 47, "Subtitle",
 InitializationCell->True],
Cell[644, 25, 128, 4, 41, "Text",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[797, 33, 58, 1, 68, "Section",
 InitializationCell->True],
Cell[858, 36, 160, 4, 26, "Text",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[1055, 45, 60, 1, 68, "Section",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[1140, 50, 38, 0, 35, "Subsection"],
Cell[1181, 52, 45, 0, 26, "Text"],
Cell[1229, 54, 685, 16, 90, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[1951, 75, 36, 0, 35, "Subsection"],
Cell[1990, 77, 150, 5, 26, "Input",
 InitializationCell->True],
Cell[2143, 84, 149, 5, 26, "Input",
 InitializationCell->True],
Cell[2295, 91, 87, 2, 26, "Input",
 InitializationCell->True],
Cell[2385, 95, 140, 4, 26, "Input",
 InitializationCell->True],
Cell[2528, 101, 152, 5, 26, "Input",
 InitializationCell->True],
Cell[2683, 108, 103, 2, 26, "Input",
 InitializationCell->True],
Cell[2789, 112, 89, 2, 26, "Input",
 InitializationCell->True],
Cell[2881, 116, 84, 2, 25, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[3002, 123, 31, 0, 35, "Subsection"],
Cell[3036, 125, 119, 3, 26, "Input",
 InitializationCell->True],
Cell[3158, 130, 101, 3, 26, "Input",
 InitializationCell->True],
Cell[3262, 135, 572, 16, 26, "Input",
 InitializationCell->True],
Cell[3837, 153, 790, 22, 58, "Input",
 InitializationCell->True],
Cell[4630, 177, 776, 22, 42, "Input",
 InitializationCell->True],
Cell[5409, 201, 545, 16, 42, "Input",
 InitializationCell->True],
Cell[5957, 219, 404, 11, 26, "Input",
 InitializationCell->True],
Cell[6364, 232, 520, 14, 26, "Input",
 InitializationCell->True],
Cell[6887, 248, 873, 28, 58, "Input",
 InitializationCell->True],
Cell[7763, 278, 1529, 42, 84, "Input",
 InitializationCell->True],
Cell[9295, 322, 1614, 43, 77, "Input",
 InitializationCell->True],
Cell[10912, 367, 1061, 28, 61, "Input",
 InitializationCell->True],
Cell[11976, 397, 942, 26, 58, "Input",
 InitializationCell->True],
Cell[12921, 425, 2120, 55, 138, "Input",
 InitializationCell->True],
Cell[15044, 482, 2599, 66, 122, "Input",
 InitializationCell->True],
Cell[17646, 550, 914, 26, 58, "Input",
 InitializationCell->True],
Cell[18563, 578, 549, 16, 26, "Input",
 InitializationCell->True],
Cell[19115, 596, 153, 3, 26, "Text"],
Cell[19271, 601, 623, 18, 42, "Input",
 InitializationCell->True],
Cell[19897, 621, 1280, 34, 74, "Input",
 InitializationCell->True],
Cell[21180, 657, 1046, 28, 58, "Input",
 InitializationCell->True],
Cell[22229, 687, 729, 18, 58, "Input",
 InitializationCell->True],
Cell[22961, 707, 415, 16, 29, "Text",
 InitializationCell->True],
Cell[23379, 725, 1994, 49, 154, "Input",
 InitializationCell->True],
Cell[25376, 776, 910, 25, 58, "Input",
 InitializationCell->True],
Cell[26289, 803, 638, 18, 42, "Input",
 InitializationCell->True],
Cell[26930, 823, 108, 3, 26, "Text"],
Cell[27041, 828, 572, 16, 26, "Input"],
Cell[27616, 846, 40, 0, 26, "Text"],
Cell[27659, 848, 2547, 60, 170, "Input",
 InitializationCell->True],
Cell[30209, 910, 147, 3, 26, "Text"],
Cell[30359, 915, 750, 21, 42, "Input",
 InitializationCell->True],
Cell[31112, 938, 939, 27, 74, "Input",
 InitializationCell->True],
Cell[32054, 967, 353, 11, 26, "Input",
 InitializationCell->True],
Cell[32410, 980, 520, 15, 42, "Input",
 InitializationCell->True],
Cell[32933, 997, 173, 6, 26, "Input",
 InitializationCell->True],
Cell[33109, 1005, 358, 10, 42, "Input",
 InitializationCell->True],
Cell[33470, 1017, 229, 6, 26, "Input",
 InitializationCell->True],
Cell[33702, 1025, 357, 11, 24, "Input",
 InitializationCell->True],
Cell[34062, 1038, 588, 17, 40, "Input",
 InitializationCell->True],
Cell[34653, 1057, 177, 6, 24, "Input",
 InitializationCell->True],
Cell[34833, 1065, 366, 10, 40, "Input",
 InitializationCell->True],
Cell[35202, 1077, 237, 6, 24, "Input",
 InitializationCell->True],
Cell[35442, 1085, 166, 4, 24, "Input",
 InitializationCell->True],
Cell[35611, 1091, 1156, 31, 130, "Input",
 InitializationCell->True],
Cell[36770, 1124, 279, 8, 24, "Input",
 InitializationCell->True],
Cell[37052, 1134, 3116, 74, 220, "Input",
 InitializationCell->True],
Cell[40171, 1210, 270, 8, 28, "Input",
 InitializationCell->True],
Cell[40444, 1220, 781, 19, 49, "Input",
 InitializationCell->True],
Cell[41228, 1241, 1556, 40, 129, "Input",
 InitializationCell->True],
Cell[42787, 1283, 596, 16, 49, "Input",
 InitializationCell->True],
Cell[43386, 1301, 1363, 36, 109, "Input",
 InitializationCell->True],
Cell[44752, 1339, 548, 15, 49, "Input",
 InitializationCell->True],
Cell[45303, 1356, 398, 9, 49, "Input",
 InitializationCell->True],
Cell[45704, 1367, 1392, 35, 109, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[47121, 1406, 7362, 198, 431, "Input",
 InitializationCell->True],
Cell[54486, 1606, 378, 7, 36, "Message"]
}, Open  ]],
Cell[54879, 1616, 1788, 47, 129, "Input",
 InitializationCell->True],
Cell[56670, 1665, 18119, 506, 833, "Input",
 InitializationCell->True],
Cell[74792, 2173, 1111, 31, 69, "Input"],
Cell[75906, 2206, 5512, 140, 489, "Input",
 InitializationCell->True],
Cell[81421, 2348, 139, 4, 28, "Input",
 InitializationCell->True],
Cell[81563, 2354, 128, 3, 28, "Input",
 InitializationCell->True],
Cell[81694, 2359, 1023, 28, 129, "Input",
 InitializationCell->True],
Cell[82720, 2389, 1301, 37, 49, "Input"],
Cell[84024, 2428, 1851, 54, 129, "Input"],
Cell[85878, 2484, 965, 27, 49, "Input"],
Cell[86846, 2513, 1977, 57, 129, "Input"],
Cell[88826, 2572, 96, 3, 28, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[88959, 2580, 36, 0, 36, "Subsection"],
Cell[88998, 2582, 103, 3, 70, "Input",
 InitializationCell->True]
}, Closed]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
