(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[      4340,        160]
NotebookOptionsPosition[      3374,        124]
NotebookOutlinePosition[      3749,        140]
CellTagsIndexPosition[      3706,        137]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["\<\
A simple implementation of intersecting subspaces specified by a basis\
\>", "Subsubsection"],

Cell[BoxData[
 RowBox[{"BeginPackage", "[", 
  "\"\<QuantumGroups`Utilities`IntersectSubspaces`\>\"", "]"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"IntersectSubspaces", "::", "usage"}], "=", 
   "\"\<IntersectSubspaces[n,\!\(\*SubscriptBox[\(\[Beta]\), \
\(1\)]\),\!\(\*SubscriptBox[\(\[Beta]\), \(2\)]\),...] gives a basis for the \
intersection of the subspaces of \!\(\*SuperscriptBox[\(A\), \(\(n\)\(\\\ \
\)\)]\)spanned by the \!\(\*SubscriptBox[\(\[Beta]\), \(i\)]\).\>\""}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"Begin", "[", "\"\<`Private`\>\"", "]"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"{", "}"}]}], "]"}], ":=", 
  RowBox[{"IdentityMatrix", "[", "n", "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"{", "basis_", "}"}]}], "]"}], ":=", "basis"}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", "}"}], ",", "_"}], "}"}]}], "]"}], ":=", 
  RowBox[{"{", "}"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", 
      RowBox[{"{", "}"}]}], "}"}]}], "]"}], ":=", 
  RowBox[{"{", "}"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"{", 
     RowBox[{"basis1_", ",", "basis2_"}], "}"}]}], "]"}], ":=", 
  RowBox[{"NullSpace", "[", 
   RowBox[{"Join", "@@", 
    RowBox[{"(", 
     RowBox[{"NullSpace", "/@", 
      RowBox[{"{", 
       RowBox[{"basis1", ",", "basis2"}], "}"}]}], ")"}]}], "]"}]}]], "Input",\

 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"{", 
     RowBox[{"basis1_", ",", "basis2_", ",", "bases__"}], "}"}]}], "]"}], ":=",
   
  RowBox[{"IntersectSubspaces", "[", 
   RowBox[{"n", ",", 
    RowBox[{"{", 
     RowBox[{"basis1", ",", 
      RowBox[{"IntersectSubspaces", "[", 
       RowBox[{"n", ",", 
        RowBox[{"{", 
         RowBox[{"basis2", ",", "bases"}], "}"}]}], "]"}]}], "}"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"End", "[", "]"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"EndPackage", "[", "]"}], ";"}]], "Input",
 InitializationCell->True]
}, Open  ]]
},
AutoGeneratedPackage->Automatic,
WindowSize->{495, 593},
WindowMargins->{{336, Automatic}, {30, Automatic}},
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (June 19, 2007)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[590, 23, 103, 2, 45, "Subsubsection"],
Cell[696, 27, 145, 3, 52, "Input",
 InitializationCell->True],
Cell[844, 32, 427, 9, 92, "Input",
 InitializationCell->True],
Cell[1274, 43, 119, 3, 31, "Input",
 InitializationCell->True],
Cell[1396, 48, 208, 6, 31, "Input",
 InitializationCell->True],
Cell[1607, 56, 181, 5, 31, "Input",
 InitializationCell->True],
Cell[1791, 63, 452, 15, 52, "Input",
 InitializationCell->True],
Cell[2246, 80, 401, 13, 52, "Input",
 InitializationCell->True],
Cell[2650, 95, 503, 16, 72, "Input",
 InitializationCell->True],
Cell[3156, 113, 96, 3, 31, "Input",
 InitializationCell->True],
Cell[3255, 118, 103, 3, 31, "Input",
 InitializationCell->True]
}, Open  ]]
}
]
*)

(* End of internal cache information *)

