(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    162711,       3641]
NotebookOptionsPosition[     45265,       1194]
NotebookOutlinePosition[    160662,       3576]
CellTagsIndexPosition[    160619,       3573]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["KnotTheory`UniversalKh` package", "Subtitle",
 InitializationCell->True],

Cell["\<\
A subpackage for KnotTheory`, providing access to the \"universal mode\" of \
Jeremy Green's JavaKh program, and a program written by Scott Morrison to \
decompose complexes over Q[t] into indecomposable direct summands.
May 27, 2007, Scott Morrison\
\>", "Text",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"BeginPackage", "[", 
   RowBox[{"\"\<KnotTheory`UniversalKh`\>\"", ",", 
    RowBox[{"{", 
     RowBox[{
     "\"\<KnotTheory`\>\"", ",", 
      "\"\<QuantumGroups`Utilities`MatrixWrapper`\>\""}], "}"}]}], "]"}], 
  ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4294630944324737`*^9, 3.4294630956041584`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"UniversalKh", "::", "about"}], "=", 
   "\"\<Universal Khovanov homology over Q[t] is calculated using Jeremy \
Green's JavaKh program, interpreted by a wrapper written by Dror Bar-Natan, \
and decomposed into direct summands using a program of Scott Morrison and \
Alexander Shumakovitch.\>\""}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"UniversalKh", "::", "usage"}], "=", 
   "\"\<UniversalKh[K] computes the universal Khovanov homology over Q. \
(Probably broken for links, at present.) See also KhC and KhE for more about \
the output.\>\""}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"sInvariant", "::", "usage"}], "=", 
   "\"\<sInvariant[K] computes the s-invariant of a knot K, using the \
UniversalKh program. (Probably broken for links, at present.)\>\""}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"KhReduced", "::", "usage"}], "=", 
   "\"\<KhReduced[K][q,t] gives the reduced Khovanov homology of the knot K, \
using the UniversalKh program.\>\""}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"KhE", "::", "usage"}], "=", 
  "\"\<KhE denotes a free generator in Khovanov homology, corresponding to an \
exceptional pair. See ?UniversalKh for more information.\>\""}], "\
\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"KhC", "::", "usage"}], "=", 
  "\"\<KhC denotes a torsion generator in Khovanov homology, with the \
differential in KhC[n] being the n-th power of the punctured torus. Thus \
KhC[1] corresponds to a knight's move pair. See ?UniversalKh for more \
information.\>\""}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.5316774507583227`*^9, 3.531677459643817*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Begin", "[", "\"\<`Private`\>\"", "]"}]], "Input",
 InitializationCell->True],

Cell[BoxData["\<\"KnotTheory`UniversalKh`Private`\"\>"], "Output",
 CellChangeTimes->{{3.531676441083721*^9, 3.5316764428426523`*^9}, 
   3.531676577676794*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"q", "=", "Global`q"}], ";", 
  RowBox[{"t", "=", "Global`t"}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"KhN", "[", 
   RowBox[{"pd_PD", ",", "options___"}], "]"}], " ", ":=", " ", 
  RowBox[{
   RowBox[{"KhN", "[", 
    RowBox[{"pd", ",", "options"}], "]"}], "=", 
   RowBox[{"Module", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
      "n", ",", "pd1", ",", "  ", "f", ",", " ", "cl", ",", " ", "out", ",", 
       "kh", ",", "saveContext", ",", "saveContextPath", ",", 
       "JavaKhDirectory", ",", "jarDirectory", ",", "classDirectory", ",", 
       "classpath", ",", 
       RowBox[{"new", "=", "True"}], ",", "javaoptions"}], "}"}], ",", 
     "\[IndentingNewLine]", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"javaoptions", "=", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"JavaOptions", "/.", 
          RowBox[{"{", "options", "}"}]}], "/.", 
         RowBox[{"Options", "[", "Kh", "]"}]}], ")"}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"n", "=", 
       RowBox[{"Max", " ", "@@", " ", 
        RowBox[{"(", 
         RowBox[{"Max", " ", "@@@", " ", "pd"}], ")"}]}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"pd1", " ", "=", " ", 
       RowBox[{"pd", " ", "/.", " ", 
        RowBox[{"{", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"X", "[", 
            RowBox[{"n", ",", " ", "i_", ",", " ", "1", ",", " ", "j_"}], 
            "]"}], " ", "\[RuleDelayed]", " ", 
           RowBox[{"X", "[", 
            RowBox[{"n", ",", " ", "i", ",", " ", 
             RowBox[{"n", "+", "1"}], ",", " ", "j"}], "]"}]}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"X", "[", 
            RowBox[{"i_", ",", " ", "1", ",", " ", "j_", ",", " ", "n"}], 
            "]"}], " ", "\[RuleDelayed]", " ", 
           RowBox[{"X", "[", 
            RowBox[{"i", ",", " ", 
             RowBox[{"n", "+", "1"}], ",", " ", "j", ",", " ", "n"}], "]"}]}],
           ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"X", "[", 
            RowBox[{"1", ",", " ", "j_", ",", " ", "n", ",", " ", "i_"}], 
            "]"}], " ", "\[RuleDelayed]", " ", 
           RowBox[{"X", "[", 
            RowBox[{
             RowBox[{"n", "+", "1"}], ",", " ", "j", ",", " ", "n", ",", " ", 
             "i"}], "]"}]}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"X", "[", 
            RowBox[{"j_", ",", " ", "n", ",", " ", "i_", ",", " ", "1"}], 
            "]"}], " ", "\[RuleDelayed]", " ", 
           RowBox[{"X", "[", 
            RowBox[{"j", ",", " ", "n", ",", " ", "i", ",", " ", 
             RowBox[{"n", "+", "1"}]}], "]"}]}]}], "\[IndentingNewLine]", 
         "}"}]}]}], ";", "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"Print", "[", 
       RowBox[{"\"\<pd1->\>\"", ",", "pd1"}], "]"}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"new", "=", "True"}], ";", " ", 
      RowBox[{"(*", " ", 
       RowBox[{
        RowBox[{
        "This", " ", "is", " ", "just", " ", "an", " ", "option", " ", "for", 
         " ", "Scott"}], ",", " ", 
        RowBox[{"to", " ", "allow", " ", "comparing", " ", "against", " ", 
         RowBox[{"Jeremy", "'"}], "s", " ", "program", " ", "before", " ", 
         "butchering", " ", 
         RowBox[{"it", "."}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{"new", ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"JavaKhDirectory", "=", 
          RowBox[{"ToFileName", "[", 
           RowBox[{
            RowBox[{"KnotTheoryDirectory", "[", "]"}], ",", 
            "\"\<JavaKh-v2\>\""}], "]"}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"classpath", "=", 
          RowBox[{"KnotTheory`FastKh`JavaKhv2ClassPath", "[", "]"}]}], ";"}], 
        ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"JavaKhDirectory", "=", 
          RowBox[{"ToFileName", "[", 
           RowBox[{
            RowBox[{"KnotTheoryDirectory", "[", "]"}], ",", 
            "\"\<JavaKh-v1\>\""}], "]"}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"classpath", "=", 
          RowBox[{"KnotTheory`FastKh`JavaKhv1ClassPath", "[", "]"}]}], ";"}], 
        ","}], "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"SetDirectory", "[", "JavaKhDirectory", "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"f", "=", 
       RowBox[{"OpenWrite", "[", 
        RowBox[{"\"\<pd\>\"", ",", 
         RowBox[{"PageWidth", "\[Rule]", "Infinity"}]}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"f", ",", 
        RowBox[{"ToString", "[", "pd1", "]"}]}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"Close", "[", "f", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"cl", "=", 
       RowBox[{"StringJoin", "[", 
        RowBox[{
        "\"\<!java -classpath \\\"\>\"", ",", "classpath", ",", 
         "\"\<\\\" \>\"", ",", "javaoptions", ",", 
         "\"\< org.katlas.JavaKh.JavaKh -U -Q < pd 2> JavaKh.log\>\""}], 
        "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"f", "=", 
       RowBox[{"OpenRead", "[", "cl", "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"out", "=", 
       RowBox[{"Read", "[", 
        RowBox[{"f", ",", "Expression"}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"Close", "[", "f", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"out", "\[Equal]", "EndOfFile"}], ",", 
        RowBox[{
         RowBox[{
         "Print", "[", 
          "\"\<Something went wrong running JavaKh; nothing was returned. The \
command line was: \>\"", "]"}], ";", 
         RowBox[{"Print", "[", "cl", "]"}], ";", 
         RowBox[{
         "Print", "[", 
          "\"\<There may have been an error log produced by Java: \>\"", 
          "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"FilePrint", "[", "\"\<JavaKh.log\>\"", "]"}], ";", 
         RowBox[{"Return", "[", "$Failed", "]"}]}]}], "]"}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"ResetDirectory", "[", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"out", "=", 
       RowBox[{"StringReplace", "[", 
        RowBox[{"out", ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"\"\<q\>\"", "\[Rule]", "\"\<#1\>\""}], ",", 
           RowBox[{"\"\<t\>\"", "\[Rule]", "\"\<#2\>\""}]}], "}"}]}], "]"}]}],
       ";", "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{"ToExpression", " ", "is", " ", 
        RowBox[{"dangerous", "!"}], " ", "We", " ", "have", " ", "to", " ", 
        "fiddle", " ", "with", " ", "the", " ", "$Context", " ", 
        RowBox[{"here", "."}]}], " ", "*)"}], "\[IndentingNewLine]", 
      RowBox[{"saveContext", "=", "$Context"}], ";", "\[IndentingNewLine]", 
      RowBox[{"saveContextPath", "=", "$ContextPath"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"$Context", "=", "\"\<KnotTheory`UniversalKh`Private`\>\""}], 
      ";", "\[IndentingNewLine]", 
      RowBox[{"$ContextPath", "=", 
       RowBox[{"{", "\"\<KnotTheory`UniversalKh`Private`\>\"", "}"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"kh", "=", 
       RowBox[{
        RowBox[{"ToExpression", "[", 
         RowBox[{"out", "<>", "\"\<&\>\""}], "]"}], "[", 
        RowBox[{"q", ",", "t"}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"$Context", "=", "saveContext"}], ";", "\[IndentingNewLine]", 
      RowBox[{"$ContextPath", "=", "saveContextPath"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"Print", "[", 
       RowBox[{"\"\<kh->\>\"", ",", "kh"}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"minr", "=", 
       RowBox[{"Exponent", "[", 
        RowBox[{"kh", ",", " ", "t", ",", " ", "Min"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"maxr", "=", 
       RowBox[{"Exponent", "[", 
        RowBox[{"kh", ",", " ", "t", ",", " ", "Max"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"obs", " ", "=", " ", 
       RowBox[{"Expand", "[", 
        RowBox[{
         RowBox[{"kh", " ", "/.", " ", 
          RowBox[{"h", " ", "\[Rule]", " ", "0"}]}], " ", "/.", " ", 
         RowBox[{
          RowBox[{"M", "[", 
           RowBox[{"_", ",", " ", "n_", ",", " ", "___"}], "]"}], "  ", 
          "\[RuleDelayed]", " ", 
          RowBox[{"Plus", " ", "@@", " ", 
           RowBox[{"Array", "[", 
            RowBox[{"Arc", ",", " ", "n"}], "]"}]}]}]}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"obs", " ", "=", " ", 
       RowBox[{
        RowBox[{"obs", " ", "/.", " ", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"q", "^", "j_."}], ")"}], "*", 
           RowBox[{"Arc", "[", "i_", "]"}]}], " ", "\[RuleDelayed]", " ", 
          RowBox[{"Arc", "[", 
           RowBox[{"j", ",", " ", "i"}], "]"}]}]}], " ", "/.", " ", 
        RowBox[{
         RowBox[{"Arc", "[", "i_", "]"}], " ", "\[RuleDelayed]", " ", 
         RowBox[{"Arc", "[", 
          RowBox[{"0", ",", " ", "i"}], "]"}]}]}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"mos", "=", " ", 
       RowBox[{"Expand", "[", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
          RowBox[{"h", "*", "kh"}], " ", "/.", " ", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{
             RowBox[{"M", "[", 
              RowBox[{"0", ",", "_"}], "]"}], "\[Rule]", " ", "0"}], ",", " ", 
            RowBox[{
             RowBox[{"M", "[", 
              RowBox[{"_", ",", " ", "0"}], "]"}], " ", "\[Rule]", " ", "0"}],
             ",", " ", 
            RowBox[{"h", "\[Rule]", " ", "H"}]}], "}"}]}], 
         "\[IndentingNewLine]", "/.", " ", 
         RowBox[{
          RowBox[{"M", "[", 
           RowBox[{"m_", ",", " ", "n_", ",", " ", "cs___"}], "]"}], " ", 
          "\[RuleDelayed]", " ", 
          RowBox[{"Plus", " ", "@@", " ", 
           RowBox[{"Flatten", "[", 
            RowBox[{"MapIndexed", "[", "\[IndentingNewLine]", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{"#1", "*", 
                 RowBox[{"Curtain", "@@", 
                  RowBox[{"Reverse", "[", "#2", "]"}]}]}], ")"}], "&"}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{"Partition", "[", 
               RowBox[{
                RowBox[{"{", "cs", "}"}], ",", " ", "n"}], "]"}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{"{", "2", "}"}]}], "\[IndentingNewLine]", "]"}], 
            "]"}]}]}]}], "\[IndentingNewLine]", "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"mos", " ", "=", " ", 
       RowBox[{
        RowBox[{"mos", " ", "/.", " ", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"q", "^", "j_."}], ")"}], "*", 
           RowBox[{"Curtain", "[", 
            RowBox[{"k_", ",", " ", "l_"}], "]"}]}], " ", "\[RuleDelayed]", 
          " ", 
          RowBox[{"Curtain", "[", 
           RowBox[{"j", ",", " ", "k", ",", " ", "l"}], "]"}]}]}], " ", "/.", 
        " ", 
        RowBox[{
         RowBox[{"Curtain", "[", 
          RowBox[{"k_", ",", " ", "l_"}], "]"}], " ", "\[RuleDelayed]", " ", 
         RowBox[{"Curtain", "[", 
          RowBox[{"0", ",", " ", "k", ",", " ", "l"}], "]"}]}]}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"mos", " ", "=", " ", 
       RowBox[{"mos", " ", "/.", " ", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"H", "^", "g_."}], ")"}], "*", 
          RowBox[{"Curtain", "[", 
           RowBox[{"j_", ",", " ", "k_", ",", " ", "l_"}], "]"}]}], " ", 
         "\[RuleDelayed]", " ", 
         RowBox[{
          RowBox[{"H", "^", 
           RowBox[{"(", 
            RowBox[{"g", "-", "1"}], ")"}]}], 
          RowBox[{"Curtain", "[", 
           RowBox[{"j", ",", " ", "k", ",", " ", 
            RowBox[{"j", "+", 
             RowBox[{"2", 
              RowBox[{"(", 
               RowBox[{"g", "-", "1"}], ")"}]}]}], ",", " ", "l"}], 
           "]"}]}]}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"Komplex", " ", "@@", " ", 
       RowBox[{"Table", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"r", ",", " ", 
           RowBox[{"Coefficient", "[", 
            RowBox[{"obs", ",", " ", "t", ",", " ", "r"}], "]"}], ",", "  ", 
           RowBox[{"Coefficient", "[", 
            RowBox[{"mos", ",", " ", "t", ",", " ", "r"}], "]"}]}], "}"}], 
         ",", " ", 
         RowBox[{"{", 
          RowBox[{"r", ",", " ", "minr", ",", " ", "maxr"}], "}"}]}], 
        "]"}]}]}]}], "\[IndentingNewLine]", "]"}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.427643628589238*^9, 3.4276437155943456`*^9}, {
   3.4276437530281725`*^9, 3.427643769081256*^9}, {3.428678796617854*^9, 
   3.4286788322891474`*^9}, {3.428772729146197*^9, 3.4287727372478466`*^9}, {
   3.428772808460245*^9, 3.4287728173630466`*^9}, {3.4294625662930465`*^9, 
   3.429462727855362*^9}, {3.429462774322178*^9, 3.4294627757542367`*^9}, 
   3.4294767618352337`*^9, {3.4294774006538095`*^9, 3.429477402346243*^9}, {
   3.429477723457979*^9, 3.4294777241089153`*^9}, {3.4294777703854575`*^9, 
   3.429477805435858*^9}, {3.4360452186769667`*^9, 3.4360452268587313`*^9}, {
   3.436045264132328*^9, 3.4360453534808044`*^9}, {3.4360454205672703`*^9, 
   3.4360454211681347`*^9}, {3.440870891579186*^9, 3.440870893700836*^9}, {
   3.4791418096834803`*^9, 3.479141816359941*^9}, {3.479141855708859*^9, 
   3.479141872932847*^9}, {3.4791421579553137`*^9, 3.479142162302726*^9}, {
   3.479142343614765*^9, 3.479142417755108*^9}, {3.509841572160133*^9, 
   3.509841573087188*^9}, {3.531676413295127*^9, 3.531676436409396*^9}, {
   3.531676567052416*^9, 3.531676575931653*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"ElementaryMatrix", "[", 
   RowBox[{"m_", ",", "n_", ",", "i_", ",", "j_"}], "]"}], ":=", 
  RowBox[{"ElementaryMatrix", "[", 
   RowBox[{"m", ",", "n", ",", "i", ",", "j", ",", "1"}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ElementaryMatrix", "[", 
    RowBox[{"m_", ",", "n_", ",", "i_", ",", "j_", ",", "z_"}], "]"}], "/;", 
   RowBox[{
    RowBox[{"1", "\[LessEqual]", "i", "\[LessEqual]", "m"}], "\[And]", 
    RowBox[{"1", "\[LessEqual]", "j", "\[LessEqual]", "n"}]}]}], ":=", 
  "\[IndentingNewLine]", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "data", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"data", "=", 
      RowBox[{"Table", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", "m", "}"}], ",", 
        RowBox[{"{", "n", "}"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"data", "\[LeftDoubleBracket]", 
       RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "=", "z"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"Matrix", "[", "data", "]"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"GradingsList", "[", 
   RowBox[{"k", ":", 
    RowBox[{"Komplex", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"n_", ",", "_", ",", "_"}], "}"}], ",", "___"}], "]"}]}], 
   "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{"n", ",", 
    RowBox[{
     RowBox[{
      RowBox[{"Cases", "[", 
       RowBox[{
        RowBox[{"{", "#", "}"}], ",", 
        RowBox[{
         RowBox[{"Arc", "[", 
          RowBox[{"m_", ",", "_"}], "]"}], "\[RuleDelayed]", "m"}], ",", 
        "2"}], "]"}], "&"}], "/@", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"List", "@@", "k"}], ")"}], "\[LeftDoubleBracket]", 
      RowBox[{"All", ",", "2"}], "\[RightDoubleBracket]"}]}]}], 
   "}"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"AllMatrices", "[", 
   RowBox[{"k", ":", 
    RowBox[{"Komplex", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"n_", ",", "_", ",", "_"}], "}"}], ",", "___"}], "]"}]}], 
   "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{"n", ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"gradings", "=", 
          RowBox[{
           RowBox[{"GradingsList", "[", "k", "]"}], "\[LeftDoubleBracket]", 
           "2", "\[RightDoubleBracket]"}]}], ",", "dimensions", ",", 
         "matrix"}], "}"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"dimensions", "=", 
         RowBox[{"Length", "/@", "gradings"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"Table", "[", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"matrix", "=", 
            RowBox[{"ZeroesMatrix", "[", 
             RowBox[{
              RowBox[{"dimensions", "\[LeftDoubleBracket]", 
               RowBox[{"i", "+", "1"}], "\[RightDoubleBracket]"}], ",", 
              RowBox[{
              "dimensions", "\[LeftDoubleBracket]", "i", 
               "\[RightDoubleBracket]"}]}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"matrix", " ", "=", 
            RowBox[{"matrix", "+", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"k", "\[LeftDoubleBracket]", 
                RowBox[{"i", ",", "3"}], "\[RightDoubleBracket]"}], "/.", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"Curtain", "[", 
                  RowBox[{"q1_", ",", "m1_", ",", "q2_", ",", "m2_"}], "]"}], 
                 "\[RuleDelayed]", 
                 RowBox[{"ElementaryMatrix", "[", 
                  RowBox[{
                   RowBox[{"dimensions", "\[LeftDoubleBracket]", 
                    RowBox[{"i", "+", "1"}], "\[RightDoubleBracket]"}], ",", 
                   RowBox[{
                   "dimensions", "\[LeftDoubleBracket]", "i", 
                    "\[RightDoubleBracket]"}], ",", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"Position", "[", 
                    RowBox[{
                    RowBox[{"gradings", "\[LeftDoubleBracket]", 
                    RowBox[{"i", "+", "1"}], "\[RightDoubleBracket]"}], ",", 
                    "q2"}], "]"}], "\[LeftDoubleBracket]", 
                    RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}], "+", 
                    "m2", "-", "1"}], ",", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"Position", "[", 
                    RowBox[{
                    RowBox[{
                    "gradings", "\[LeftDoubleBracket]", "i", 
                    "\[RightDoubleBracket]"}], ",", "q1"}], "]"}], 
                    "\[LeftDoubleBracket]", 
                    RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}], "+", 
                    "m1", "-", "1"}]}], "]"}]}], ")"}]}], ")"}]}]}]}], 
          "\[IndentingNewLine]", ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "1", ",", 
            RowBox[{
             RowBox[{"Length", "[", "k", "]"}], "-", "1"}]}], "}"}]}], 
         "]"}]}]}], "\[IndentingNewLine]", "]"}], "/.", 
     RowBox[{"{", 
      RowBox[{"H", "\[Rule]", "T"}], "}"}]}]}], "\[IndentingNewLine]", 
   "}"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4294637480423183`*^9, 3.4294637562040544`*^9}, {
   3.429463844551091*^9, 3.4294638601535263`*^9}, {3.4294639556408305`*^9, 
   3.429463985353555*^9}, 3.429464051518696*^9, {3.6055452927092376`*^9, 
   3.6055452930972595`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"ZeroVector", "[", "n_", "]"}], ":=", 
  RowBox[{"Table", "[", 
   RowBox[{"0", ",", 
    RowBox[{"{", "n", "}"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"Matrix", "/:", 
  RowBox[{
   RowBox[{"\[Alpha]_", " ", 
    RowBox[{"Matrix", "[", 
     RowBox[{"j_", ",", "k_", ",", "data_"}], "]"}]}], "/;", 
   RowBox[{"(", 
    RowBox[{"NumberQ", "[", 
     RowBox[{"\[Alpha]", "/.", 
      RowBox[{"T", "\[Rule]", "3.14159`"}]}], "]"}], ")"}]}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"j", ",", "k", ",", 
    RowBox[{"\[Alpha]", " ", "data"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"FirstRow", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"1", ",", "c", ",", 
    RowBox[{"{", 
     RowBox[{"First", "[", "data", "]"}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FirstRow", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"0", ",", "c_", ",", "_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", "c"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FirstColumn", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "1", ",", 
    RowBox[{
     RowBox[{
      RowBox[{"{", 
       RowBox[{"First", "[", "#", "]"}], "}"}], "&"}], "/@", "data"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FirstColumn", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "0", ",", "_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RestColumns", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", 
    RowBox[{"c", "-", "1"}], ",", 
    RowBox[{"Rest", "/@", "data"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RestColumns", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", 
     RowBox[{"0", "|", "1"}], ",", "_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "0"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RestRows", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{
    RowBox[{"r", "-", "1"}], ",", "c", ",", 
    RowBox[{"Rest", "[", "data", "]"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RestRows", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{
     RowBox[{"0", "|", "1"}], ",", "c_", ",", "_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"0", ",", "c"}], "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"RotateRows", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "c", ",", 
    RowBox[{"RotateLeft", "[", "data", "]"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RotateColumns", "[", 
   RowBox[{"Matrix", "[", 
    RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "c", ",", 
    RowBox[{"RotateLeft", "/@", "data"}]}], "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"RotateRows", "[", 
   RowBox[{
    RowBox[{"Matrix", "[", 
     RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], ",", "n_"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "c", ",", 
    RowBox[{"RotateLeft", "[", 
     RowBox[{"data", ",", "n"}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RotateColumns", "[", 
   RowBox[{
    RowBox[{"Matrix", "[", 
     RowBox[{"r_", ",", "c_", ",", "data_"}], "]"}], ",", "n_"}], "]"}], ":=", 
  RowBox[{"Matrix", "[", 
   RowBox[{"r", ",", "c", ",", 
    RowBox[{
     RowBox[{
      RowBox[{"RotateLeft", "[", 
       RowBox[{"#", ",", "n"}], "]"}], "&"}], "/@", "data"}]}], 
   "]"}]}]}], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"UniversalKhTimingData", "=", 
   RowBox[{"{", "}"}]}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"twist", "[", 
   RowBox[{
   "\[Alpha]_", ",", "k_", ",", "\[Lambda]_", ",", "\[Mu]_", ",", "\[Nu]_"}], 
   "]"}], ":=", 
  RowBox[{"\[Nu]", "-", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "/", "\[Alpha]"}], ")"}], 
    RowBox[{"T", "^", 
     RowBox[{"(", 
      RowBox[{"-", "k"}], ")"}]}], 
    RowBox[{"\[Mu]", ".", "\[Lambda]"}]}]}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UniversalKh", "[", 
   RowBox[{
    RowBox[{"K", ":", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"Knot", "|", "Link", "|", "TorusKnot"}], ")"}], "[", 
       RowBox[{"_Integer", ",", "__"}], "]"}], ")"}]}], ",", "options___"}], 
   "]"}], ":=", 
  RowBox[{
   RowBox[{"UniversalKh", "[", 
    RowBox[{"K", ",", "options"}], "]"}], "=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"khn", ",", "result", ",", "components", ",", "factor"}], "}"}],
      ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"CreditMessage", "[", 
       RowBox[{"UniversalKh", "::", "about"}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Length", "[", 
          RowBox[{"Skeleton", "[", "K", "]"}], "]"}], ">", "1"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
         "Print", "[", 
          "\"\<Warning: UniversalKh is currently *broken* for links. It may \
be a simple matter of dividing the coefficient of KhE by (q+q^{-1}), but we \
haven't identified the bug.\>\"", "]"}], ";"}]}], "\[IndentingNewLine]", 
       "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"khn", "=", 
       RowBox[{"KhN", "[", 
        RowBox[{
         RowBox[{"PD", "[", "K", "]"}], ",", "options"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"result", "=", 
       RowBox[{"AbsoluteTiming", "[", 
        RowBox[{"DecomposeComplex", "[", 
         RowBox[{
          RowBox[{"GradingsList", "[", "khn", "]"}], ",", 
          RowBox[{"AllMatrices", "[", "khn", "]"}]}], "]"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"AppendTo", "[", 
       RowBox[{"UniversalKhTimingData", ",", 
        RowBox[{"{", 
         RowBox[{"K", ",", 
          RowBox[{
           RowBox[{
           "result", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "/.", 
           RowBox[{"Second", "\[Rule]", "1"}]}]}], "}"}]}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{
      "result", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}]}], 
    "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"UniversalKh", "[", 
   RowBox[{"d_PD", ",", "options___"}], "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"khn", "=", 
      RowBox[{"KhN", "[", 
       RowBox[{"d", ",", "options"}], "]"}]}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{"DecomposeComplex", "[", 
     RowBox[{
      RowBox[{"GradingsList", "[", "khn", "]"}], ",", 
      RowBox[{"AllMatrices", "[", "khn", "]"}]}], "]"}]}], 
   "\[IndentingNewLine]", "]"}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.479142440708848*^9, 3.479142501937842*^9}, {
  3.47914261083459*^9, 3.479142610917959*^9}, {3.50984154327634*^9, 
  3.509841544945129*^9}, {3.531677484117547*^9, 3.531677537816416*^9}, {
  3.6055453087241535`*^9, 3.605545310608261*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"DecomposeComplex", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"g0_", ",", "gradings0_"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"g0_", ",", "matrices0_List"}], "}"}]}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"g", "=", "g0"}], ",", 
      RowBox[{"gradings", "=", "gradings0"}], ",", 
      RowBox[{"matrices", "=", "matrices0"}], ",", 
      RowBox[{"result", "=", "0"}], ",", "matrix", ",", "objects", ",", 
      "exponents", ",", "i", ",", "j", ",", "k", ",", "\[Alpha]", ",", 
      "\[Lambda]", ",", "\[Mu]", ",", "\[Nu]"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"While", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Length", "[", "matrices", "]"}], ">", "0"}], ",", 
       RowBox[{
        RowBox[{"objects", "=", 
         RowBox[{
         "gradings", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}],
         ";", 
        RowBox[{"matrix", "=", 
         RowBox[{
         "matrices", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}],
         ";", "\[IndentingNewLine]", 
        RowBox[{"While", "[", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"exponents", "=", 
             RowBox[{"DeleteCases", "[", 
              RowBox[{
               RowBox[{"Union", "[", 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"Exponent", "[", 
                    RowBox[{"#1", ",", "T"}], "]"}], "&"}], ")"}], "/@", 
                 RowBox[{"Flatten", "[", 
                  RowBox[{"MatrixData", "[", "matrix", "]"}], "]"}]}], "]"}], 
               ",", 
               RowBox[{"-", "\[Infinity]"}]}], "]"}]}], ")"}], "\[NotEqual]", 
           RowBox[{"{", "}"}]}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"k", "=", 
            RowBox[{"First", "[", "exponents", "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"k", "\[Equal]", "0"}], ",", "\[IndentingNewLine]", 
             RowBox[{
              RowBox[{
              "Print", "[", 
               "\"\<Found an isomorphism I wasn't expecting!\>\"", "]"}], ";",
               "\[IndentingNewLine]", 
              RowBox[{"Print", "[", 
               RowBox[{"\"\<Result so far: \>\"", ",", "result"}], "]"}], ";",
               "\[IndentingNewLine]", 
              RowBox[{"Print", "[", 
               RowBox[{
               "\"\<Remaining objects at this height: \>\"", ",", "objects"}],
                "]"}], ";", "\[IndentingNewLine]", 
              RowBox[{"Print", "[", 
               RowBox[{
               "\"\<Remaining matrices at this height: \>\"", ",", "matrix"}],
                "]"}], ";", "\[IndentingNewLine]", 
              RowBox[{"Print", "[", 
               RowBox[{"\"\<Other gradings: \>\"", ",", "gradings"}], "]"}], 
              ";", "\[IndentingNewLine]", 
              RowBox[{"Print", "[", 
               RowBox[{"\"\<Other matrices: \>\"", ",", "matrices"}], "]"}], 
              ";", "\[IndentingNewLine]", 
              RowBox[{"Return", "[", "$Failed", "]"}]}]}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"i", ",", "j"}], "}"}], "=", 
            RowBox[{
             RowBox[{"Position", "[", 
              RowBox[{
               RowBox[{"MatrixData", "[", "matrix", "]"}], ",", 
               RowBox[{"e_", "/;", 
                RowBox[{
                 RowBox[{"Exponent", "[", 
                  RowBox[{"e", ",", "T"}], "]"}], "\[Equal]", "k"}]}], ",", 
               "2", ",", "1"}], "]"}], "\[LeftDoubleBracket]", "1", 
             "\[RightDoubleBracket]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"objects", "=", 
            RowBox[{"RotateLeft", "[", 
             RowBox[{"objects", ",", 
              RowBox[{"j", "-", "1"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
            "gradings", "\[LeftDoubleBracket]", "2", 
             "\[RightDoubleBracket]"}], "=", 
            RowBox[{"RotateLeft", "[", 
             RowBox[{
              RowBox[{
              "gradings", "\[LeftDoubleBracket]", "2", 
               "\[RightDoubleBracket]"}], ",", 
              RowBox[{"i", "-", "1"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"matrix", "=", 
            RowBox[{"RotateRows", "[", 
             RowBox[{"matrix", ",", 
              RowBox[{"i", "-", "1"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"matrix", "=", 
            RowBox[{"RotateColumns", "[", 
             RowBox[{"matrix", ",", 
              RowBox[{"j", "-", "1"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"Length", "[", "matrices", "]"}], ">", "1"}], ",", 
             RowBox[{
              RowBox[{
              "matrices", "\[LeftDoubleBracket]", "2", 
               "\[RightDoubleBracket]"}], "=", 
              RowBox[{"RotateColumns", "[", 
               RowBox[{
                RowBox[{
                "matrices", "\[LeftDoubleBracket]", "2", 
                 "\[RightDoubleBracket]"}], ",", 
                RowBox[{"i", "-", "1"}]}], "]"}]}]}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"\[Alpha]", "=", 
            FractionBox[
             RowBox[{"matrix", "\[LeftDoubleBracket]", 
              RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}], 
             SuperscriptBox["T", "k"]]}], ";", "\[IndentingNewLine]", 
           RowBox[{"\[Lambda]", "=", 
            RowBox[{"RestColumns", "[", 
             RowBox[{"FirstRow", "[", "matrix", "]"}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"\[Mu]", "=", 
            RowBox[{"RestRows", "[", 
             RowBox[{"FirstColumn", "[", "matrix", "]"}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"\[Nu]", "=", 
            RowBox[{"RestRows", "[", 
             RowBox[{"RestColumns", "[", "matrix", "]"}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"matrix", "=", 
            RowBox[{"twist", "[", 
             RowBox[{
             "\[Alpha]", ",", "k", ",", "\[Lambda]", ",", "\[Mu]", ",", 
              "\[Nu]"}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"Length", "[", "matrices", "]"}], ">", "1"}], ",", 
             RowBox[{
              RowBox[{
              "matrices", "\[LeftDoubleBracket]", "2", 
               "\[RightDoubleBracket]"}], "=", 
              RowBox[{"RestColumns", "[", 
               RowBox[{
               "matrices", "\[LeftDoubleBracket]", "2", 
                "\[RightDoubleBracket]"}], "]"}]}]}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"result", "+=", 
            RowBox[{
             SuperscriptBox["t", 
              RowBox[{"g", "+", "1"}]], " ", 
             SuperscriptBox["q", 
              RowBox[{
               RowBox[{"2", " ", "k"}], "+", 
               RowBox[{
               "objects", "\[LeftDoubleBracket]", "1", 
                "\[RightDoubleBracket]"}]}]], " ", 
             RowBox[{"KhC", "[", "k", "]"}]}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"objects", "=", 
            RowBox[{"Rest", "[", "objects", "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
            "gradings", "\[LeftDoubleBracket]", "2", 
             "\[RightDoubleBracket]"}], "=", 
            RowBox[{"Rest", "[", 
             RowBox[{
             "gradings", "\[LeftDoubleBracket]", "2", 
              "\[RightDoubleBracket]"}], "]"}]}], ";"}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"!", 
           RowBox[{"ZeroMatrixQ", "[", "matrix", "]"}]}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
           "Print", "[", "\"\<I was expecting the matrix to be zero now.\>\"",
             "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{"Print", "[", 
            RowBox[{"\"\<Result so far: \>\"", ",", "result"}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"Print", "[", 
            RowBox[{
            "\"\<Remaining objects at this height: \>\"", ",", "objects"}], 
            "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{"Print", "[", 
            RowBox[{
            "\"\<Remaining matrices at this height: \>\"", ",", "matrix"}], 
            "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{"Print", "[", 
            RowBox[{"\"\<Other gradings: \>\"", ",", "gradings"}], "]"}], ";",
            "\[IndentingNewLine]", 
           RowBox[{"Print", "[", 
            RowBox[{"\"\<Other matrices: \>\"", ",", "matrices"}], "]"}], ";",
            "\[IndentingNewLine]", 
           RowBox[{"Return", "[", "$Failed", "]"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"result", "+=", 
         RowBox[{"KhE", " ", 
          RowBox[{"Plus", "@@", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["t", "g"], " ", 
               SuperscriptBox["q", "#1"]}], "&"}], ")"}], "/@", 
            "objects"}]}]}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"matrices", "=", 
         RowBox[{"Rest", "[", "matrices", "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"gradings", "=", 
         RowBox[{"Rest", "[", "gradings", "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"g", "++"}], ";"}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"result", "+=", 
      RowBox[{"KhE", " ", 
       RowBox[{"Plus", "@@", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox["t", "g"], " ", 
            SuperscriptBox["q", "#1"]}], "&"}], ")"}], "/@", 
         RowBox[{
         "gradings", "\[LeftDoubleBracket]", "1", 
          "\[RightDoubleBracket]"}]}]}]}]}], ";", "result"}]}], 
   "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4294770001879683`*^9, 3.429477029480088*^9}, {
   3.429477064009739*^9, 3.4294770835678625`*^9}, {3.429477133639862*^9, 
   3.4294772491159086`*^9}, 3.429477300489781*^9, {3.479138072286153*^9, 
   3.479138085658778*^9}, {3.479138149924748*^9, 3.4791382759059963`*^9}, {
   3.47913836110141*^9, 3.479138377016591*^9}, {3.479138425463806*^9, 
   3.479138426453793*^9}, 3.531676027694219*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"sInvariant", "[", "K_", "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"ukh", "=", 
      RowBox[{"UniversalKh", "[", "K", "]"}]}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Length", "[", 
        RowBox[{"Position", "[", 
         RowBox[{"ukh", ",", "KhE"}], "]"}], "]"}], "\[Equal]", "1"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"Replace", "[", 
       RowBox[{
        RowBox[{"ukh", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"_KhC", "\[RuleDelayed]", "0"}], ",", 
           RowBox[{"KhE", "\[Rule]", "1"}]}], "}"}]}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{
           SuperscriptBox["q", "s_."], "\[RuleDelayed]", "s"}], ",", 
          RowBox[{"1", "\[Rule]", "0"}]}], "}"}]}], "]"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"ukh", "/.", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"_KhC", "\[RuleDelayed]", "0"}], ",", 
         RowBox[{"KhE", "\[Rule]", "1"}]}], "}"}]}]}], "\[IndentingNewLine]", 
     "]"}]}], "\[IndentingNewLine]", "]"}]}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Alpha]0rules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"KhE", "\[Rule]", 
      RowBox[{"q", "+", 
       SuperscriptBox["q", 
        RowBox[{"-", "1"}]]}]}], ",", 
     RowBox[{
      RowBox[{"KhC", "[", "1", "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{
        SuperscriptBox["t", 
         RowBox[{"-", "1"}]], 
        SuperscriptBox["q", 
         RowBox[{"-", "3"}]]}], "+", " ", 
       SuperscriptBox["q", "1"]}]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"KhC", "[", "n_", "]"}], "/;", 
       RowBox[{"n", "\[GreaterEqual]", "2"}]}], "\[RuleDelayed]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"q", "+", 
         SuperscriptBox["q", 
          RowBox[{"-", "1"}]]}], ")"}], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox["t", 
           RowBox[{"-", "1"}]], " ", 
          SuperscriptBox["q", 
           RowBox[{
            RowBox[{"-", "2"}], "n"}]]}], "+", "1"}], ")"}]}]}]}], "}"}]}], 
  ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"reducedRules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"KhE", "\[Rule]", 
      SuperscriptBox["q", 
       RowBox[{"-", "1"}]]}], ",", 
     RowBox[{
      RowBox[{"KhC", "[", "n_", "]"}], "\[RuleDelayed]", 
      RowBox[{
       RowBox[{
        SuperscriptBox["t", 
         RowBox[{"-", "1"}]], " ", 
        SuperscriptBox["q", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], "n"}], "-", "1"}]]}], "+", 
       SuperscriptBox["q", 
        RowBox[{"-", "1"}]]}]}]}], "}"}]}], ";"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"KhReduced", "[", "K_", "]"}], ":=", 
  RowBox[{"Function", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"q", ",", "t"}], "}"}], ",", 
    RowBox[{"Evaluate", "[", 
     RowBox[{"Expand", "[", 
      RowBox[{
       RowBox[{"UniversalKh", "[", "K", "]"}], "/.", "reducedRules"}], "]"}], 
     "]"}]}], "]"}]}]], "Input",
 InitializationCell->True],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"End", "[", "]"}]], "Input",
 InitializationCell->True],

Cell[BoxData["\<\"KnotTheory`UniversalKh`Private`\"\>"], "Output",
 CellChangeTimes->{{3.5316764421428432`*^9, 3.531676443885693*^9}, 
   3.531676578868841*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"EndPackage", "[", "]"}]], "Input",
 InitializationCell->True]
}, Open  ]]
},
AutoGeneratedPackage->Automatic,
WindowSize->{1904, 985},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
ShowSelection->True,
InputAliases->{"intt" -> RowBox[{"\[Integral]", 
     RowBox[{"\[SelectionPlaceholder]", 
       RowBox[{"\[DifferentialD]", "\[Placeholder]"}]}]}], "dintt" -> RowBox[{
     SubsuperscriptBox[
     "\[Integral]", "\[SelectionPlaceholder]", "\[Placeholder]"], 
     RowBox[{"\[Placeholder]", 
       RowBox[{"\[DifferentialD]", "\[Placeholder]"}]}]}], "sumt" -> RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"\[SelectionPlaceholder]", "=", "\[Placeholder]"}], 
      "\[Placeholder]"], "\[Placeholder]"}], "prodt" -> RowBox[{
     UnderoverscriptBox["\[Product]", 
      RowBox[{"\[SelectionPlaceholder]", "=", "\[Placeholder]"}], 
      "\[Placeholder]"], "\[Placeholder]"}], "dt" -> RowBox[{
     SubscriptBox["\[PartialD]", "\[Placeholder]"], " ", 
     "\[SelectionPlaceholder]"}], "notation" -> RowBox[{"Notation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        " ", "\[DoubleLongLeftRightArrow]", " ", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "notation>" -> 
  RowBox[{"Notation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        " ", "\[DoubleLongRightArrow]", " ", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "notation<" -> 
  RowBox[{"Notation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        " ", "\[DoubleLongLeftArrow]", " ", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "symb" -> RowBox[{"Symbolize", "[", 
     TagBox[
     "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"], 
     "]"}], "infixnotation" -> RowBox[{"InfixNotation", "[", 
     RowBox[{
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> "NotationTemplateStyle"],
        ",", "\[Placeholder]"}], "]"}], "addia" -> 
  RowBox[{"AddInputAlias", "[", 
     RowBox[{"\"\[Placeholder]\"", "\[Rule]", 
       TagBox[
       "\[Placeholder]", NotationBoxTag, TagStyle -> 
        "NotationTemplateStyle"]}], "]"}], "pattwraper" -> 
  TagBox["\[Placeholder]", NotationPatternTag, TagStyle -> 
    "NotationPatternWrapperStyle"], "madeboxeswraper" -> 
  TagBox["\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
    "NotationMadeBoxesWrapperStyle"]},
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Title"], 
      Cell[
      "Modify the definitions below to change the default appearance of all \
cells in a given style.  Make modifications to any definition using commands \
in the Format menu.", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {
           "Color" -> RGBColor[0.269993, 0.308507, 0.6]}, 
          CellLabelMargins -> {{12, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          CellLabelMargins -> {{24, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 12], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellLabelMargins -> {{8, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 8], 
         Cell[
          StyleData[All, "SlideShow"], PageWidth -> WindowWidth, 
          ScrollingOptions -> {
           "PagewiseDisplay" -> True, "VerticalScrollRange" -> Fit}, 
          ShowCellBracket -> False, ScriptMinSize -> 9], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, 
          CellLabelMargins -> {{2, Inherited}, {Inherited, Inherited}}, 
          ScriptMinSize -> 5, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "The options defined for the style below will be used at the \
Notebook level.", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, CellFrameLabelMargins -> 
          6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, Inherited}, {10, 30}}, 
             CellGroupingRules -> {"TitleGrouping", 0}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LineSpacing -> {1, 11}, LanguageCategory -> 
             "NaturalLanguage", CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 36, FontWeight -> "Bold"], 
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{27, 10}, {10, 30}}, LineSpacing -> {1, 0}, 
             FontSize -> 44], 
            Cell[
             StyleData["Title", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, FontSize -> 20], 
            Cell[
             StyleData["Title", "SlideShow"]], 
            Cell[
             StyleData["Title", "Printout"], 
             CellMargins -> {{2, 10}, {12, 30}}, FontSize -> 24]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, Inherited}, {20, 2}},
              CellGroupingRules -> {"TitleGrouping", 10}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 24], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{27, 10}, {20, 2}}, LineSpacing -> {1, 0}, 
             FontSize -> 36], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 4}}, FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "SlideShow"]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{2, 10}, {12, 8}}, FontSize -> 18]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], 
             CellMargins -> {{27, Inherited}, {8, 2}}, 
             CellGroupingRules -> {"TitleGrouping", 20}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 16], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{24, 10}, {20, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {8, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "SlideShow"]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{2, 10}, {12, 8}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{0, 0}, {0, 1}}, 
             CellMargins -> {{27, Inherited}, {8, 34}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 4, DefaultNewInlineCellStyle -> 
             "None", InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LineSpacing -> {1, 2}, LanguageCategory -> 
             "NaturalLanguage", CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 20, FontWeight -> "Bold"], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{40, 10}, {11, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{18, Inherited}, {6, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "SlideShow"]], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{2, 0}, {7, 22}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> "\[FilledSmallSquare]", 
             CellMargins -> {{60, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 14, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{36, 10}, {11, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{16, Inherited}, {6, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "SlideShow"]], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {7, 22}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> 
             "\[FilledSmallSquare]", 
             CellMargins -> {{60, Inherited}, {2, 10}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, DefaultNewInlineCellStyle -> "None", 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LanguageCategory -> "NaturalLanguage", 
             CounterIncrements -> "Subsubsection", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{34, 10}, {11, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{17, Inherited}, {6, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "SlideShow"]], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{9, 0}, {7, 14}}, FontSize -> 11]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{60, 10}, {7, 7}}, 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, LineSpacing -> {1, 3}, CounterIncrements -> 
             "Text"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{24, 10}, {10, 10}}, LineSpacing -> {1, 5}, 
             FontSize -> 16], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {6, 6}},
              LineSpacing -> {1, 1}, FontSize -> 11], 
            Cell[
             StyleData["Text", "SlideShow"]], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{2, 2}, {6, 6}}, 
             TextJustification -> 0.5, Hyphenation -> True, FontSize -> 10]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{60, 10}, {6, 6}}, 
             DefaultNewInlineCellStyle -> "None", LineSpacing -> {1, 3}, 
             LanguageCategory -> "NaturalLanguage", CounterIncrements -> 
             "SmallText", FontFamily -> "Helvetica", FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{24, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {5, 5}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "SlideShow"]], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{2, 2}, {5, 5}}, TextJustification -> 0.5, 
             Hyphenation -> True, FontSize -> 7]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names. Some attributes for these styles are actually set in FormatType Styles \
(in the last section of this stylesheet). ", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellMargins -> {{66, 10}, {5, 7}}, 
             Evaluatable -> True, CellGroupingRules -> "InputGrouping", 
             CellHorizontalScrolling -> True, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, DefaultFormatType -> 
             DefaultInputFormatType, "TwoByteSyntaxCharacterAutoReplacement" -> 
             True, HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> "Mathematica", 
             FormatType -> InputForm, ShowStringCharacters -> True, 
             NumberMarks -> True, LinebreakAdjustments -> {0.85, 2, 10, 0, 1},
              CounterIncrements -> "Input", FontWeight -> "Bold"], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{72, Inherited}, {8, 10}}, LineSpacing -> {1, 0},
              FontSize -> 16], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {2, 3}}, FontSize -> 11], 
            Cell[
             StyleData["Input", "SlideShow"]], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{39, 0}, {4, 6}},
              LinebreakAdjustments -> {0.85, 2, 10, 1, 1}, FontSize -> 9]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["InputOnly"], CellMargins -> {{66, 10}, {7, 7}}, 
             Evaluatable -> True, CellGroupingRules -> "InputGrouping", 
             CellHorizontalScrolling -> True, DefaultFormatType -> 
             DefaultInputFormatType, "TwoByteSyntaxCharacterAutoReplacement" -> 
             True, HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> "Mathematica", 
             FormatType -> InputForm, ShowStringCharacters -> True, 
             NumberMarks -> True, LinebreakAdjustments -> {0.85, 2, 10, 0, 1},
              CounterIncrements -> "Input", StyleMenuListing -> None, 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["InputOnly", "Presentation"], 
             CellMargins -> {{72, Inherited}, {8, 10}}, LineSpacing -> {1, 0},
              FontSize -> 16], 
            Cell[
             StyleData["InputOnly", "Condensed"], 
             CellMargins -> {{40, 10}, {2, 3}}, FontSize -> 11], 
            Cell[
             StyleData["InputOnly", "SlideShow"]], 
            Cell[
             StyleData["InputOnly", "Printout"], 
             CellMargins -> {{39, 0}, {4, 6}}, 
             LinebreakAdjustments -> {0.85, 2, 10, 1, 1}, FontSize -> 9]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellMargins -> {{66, 10}, {7, 5}}, 
             CellEditDuplicate -> True, CellGroupingRules -> "OutputGrouping",
              CellHorizontalScrolling -> True, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, DefaultFormatType -> 
             DefaultOutputFormatType, "TwoByteSyntaxCharacterAutoReplacement" -> 
             True, HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> None, FormatType -> 
             InputForm, CounterIncrements -> "Output"], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{72, Inherited}, {10, 8}}, LineSpacing -> {1, 0},
              FontSize -> 16], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{41, Inherited}, {3, 2}}, FontSize -> 11], 
            Cell[
             StyleData["Output", "SlideShow"]], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{39, 0}, {6, 4}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{66, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> None, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontFamily -> "Helvetica", FontSize -> 10, FontColor -> 
             RGBColor[0.6, 0.100008, 0.100008]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{72, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}, FontSize -> 16], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{41, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 11], 
            Cell[
             StyleData["Message", "SlideShow"]], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{39, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 7, FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{66, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, 
             AutoItalicWords -> {}, LanguageCategory -> None, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{72, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}, FontSize -> 16], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{41, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 11], 
            Cell[
             StyleData["Print", "SlideShow"]], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{39, Inherited}, {Inherited, Inherited}}, 
             FontSize -> 8]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{4, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, LanguageCategory -> 
             None, FormatType -> InputForm, CounterIncrements -> "Graphics", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             StyleMenuListing -> None, FontFamily -> "Courier", FontSize -> 
             10], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             ImageMargins -> {{62, Inherited}, {Inherited, 0}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             ImageMargins -> {{38, Inherited}, {Inherited, 0}}, Magnification -> 
             0.6], 
            Cell[
             StyleData["Graphics", "SlideShow"]], 
            Cell[
             StyleData["Graphics", "Printout"], 
             ImageMargins -> {{30, Inherited}, {Inherited, 0}}, Magnification -> 
             0.8]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], LanguageCategory -> None, 
             StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 
             9, FontColor -> RGBColor[0.269993, 0.308507, 0.6]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 9], 
            Cell[
             StyleData["CellLabel", "SlideShow"]], 
            Cell[
             StyleData["CellLabel", "Printout"], FontFamily -> "Courier", 
             FontSize -> 8, FontSlant -> "Italic", FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["FrameLabel"], LanguageCategory -> None, 
             StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 
             9], 
            Cell[
             StyleData["FrameLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["FrameLabel", "Condensed"], FontSize -> 9], 
            Cell[
             StyleData["FrameLabel", "SlideShow"]], 
            Cell[
             StyleData["FrameLabel", "Printout"], FontFamily -> "Courier", 
             FontSize -> 8, FontSlant -> "Italic", FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Inline Formatting", "Section"], 
         Cell[
         "These styles are for modifying individual words or letters in a \
cell exclusive of the cell tag.", "Text"], 
         Cell[
          StyleData["RM"], StyleMenuListing -> None, FontWeight -> "Plain", 
          FontSlant -> "Plain"], 
         Cell[
          StyleData["BF"], StyleMenuListing -> None, FontWeight -> "Bold"], 
         Cell[
          StyleData["IT"], StyleMenuListing -> None, FontSlant -> "Italic"], 
         Cell[
          StyleData["TR"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Plain", FontSlant -> "Plain"], 
         Cell[
          StyleData["TI"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Plain", FontSlant -> "Italic"], 
         Cell[
          StyleData["TB"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Bold", FontSlant -> "Plain"], 
         Cell[
          StyleData["TBI"], StyleMenuListing -> None, FontFamily -> "Times", 
          FontWeight -> "Bold", FontSlant -> "Italic"], 
         Cell[
          StyleData["MR"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Plain", FontSlant -> "Plain"], 
         Cell[
          StyleData["MO"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Plain", FontSlant -> "Italic"], 
         Cell[
          StyleData["MB"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Bold", FontSlant -> "Plain"], 
         Cell[
          StyleData["MBO"], "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          HyphenationOptions -> {"HyphenationCharacter" -> "\[Continuation]"},
           StyleMenuListing -> None, FontFamily -> "Courier", FontWeight -> 
          "Bold", FontSlant -> "Italic"], 
         Cell[
          StyleData["SR"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Plain", FontSlant -> "Plain"], 
         Cell[
          StyleData["SO"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Plain", FontSlant -> "Italic"], 
         Cell[
          StyleData["SB"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Bold", FontSlant -> "Plain"], 
         Cell[
          StyleData["SBO"], StyleMenuListing -> None, FontFamily -> 
          "Helvetica", FontWeight -> "Bold", FontSlant -> "Italic"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SO10"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 10, FontWeight -> "Plain", FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["SO10", "Printout"], StyleMenuListing -> None, 
             FontFamily -> "Helvetica", FontSize -> 7, FontWeight -> "Plain", 
             FontSlant -> "Italic"]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Inert"], StyleMenuListing -> None, Background -> 
             RGBColor[0.870588, 0.905882, 0.972549]], 
            Cell[
             StyleData["Inert", "Printout"], StyleMenuListing -> None, 
             Background -> GrayLevel[1]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["InlineFormula"], CellMargins -> {{10, 4}, {0, 8}}, 
             CellHorizontalScrolling -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Formula", ScriptLevel -> 1, SingleLetterItalics -> True], 
            Cell[
             StyleData["InlineFormula", "Presentation"], 
             CellMargins -> {{24, 10}, {10, 10}}, LineSpacing -> {1, 5}, 
             FontSize -> 16], 
            Cell[
             StyleData["InlineFormula", "Condensed"], 
             CellMargins -> {{8, 10}, {6, 6}}, LineSpacing -> {1, 1}, 
             FontSize -> 11], 
            Cell[
             StyleData["InlineFormula", "SlideShow"]], 
            Cell[
             StyleData["InlineFormula", "Printout"], 
             CellMargins -> {{2, 0}, {6, 6}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], 
             CellMargins -> {{60, Inherited}, {Inherited, Inherited}}, 
             CellHorizontalScrolling -> True, DefaultFormatType -> 
             DefaultInputFormatType, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Formula", ScriptLevel -> 0, SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             LineSpacing -> {1, 5}, FontSize -> 16], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], LineSpacing -> {1, 1}, 
             FontSize -> 11], 
            Cell[
             StyleData["DisplayFormula", "SlideShow"]], 
            Cell[
             StyleData["DisplayFormula", "Printout"], FontSize -> 10]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
             CellMargins -> {{60, 4}, {0, 8}}, CellHorizontalScrolling -> 
             True, Hyphenation -> False, LanguageCategory -> "Formula", 
             ScriptLevel -> 1, FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{24, 10}, {10, 10}}, LineSpacing -> {1, 5}, 
             FontSize -> 16], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {6, 6}}, LineSpacing -> {1, 1}, 
             FontSize -> 11], 
            Cell[
             StyleData["Program", "SlideShow"]], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{2, 0}, {6, 6}}, FontSize -> 9]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Outline Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline1"], CellMargins -> {{60, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, 
             ParagraphIndent -> -38, CounterIncrements -> "Outline1", 
             CounterAssignments -> {{"Outline2", 0}, {"Outline3", 0}, {
               "Outline4", 0}}, FontSize -> 18, FontWeight -> "Bold", 
             CounterBoxOptions -> {
              CounterFunction :> 
               Utilities`Notation`Private`CapitalRomanNumeral}], 
            Cell[
             StyleData["Outline1", "SlideShow"]], 
            Cell[
             StyleData["Outline1", "Printout"], 
             CounterBoxOptions -> {
              CounterFunction :> 
               Utilities`Notation`Private`CapitalRomanNumeral}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline2"], CellMargins -> {{90, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 60}, 
             ParagraphIndent -> -27, CounterIncrements -> "Outline2", 
             CounterAssignments -> {{"Outline3", 0}, {"Outline4", 0}}, 
             FontSize -> 15, FontWeight -> "Bold", 
             CounterBoxOptions -> {CounterFunction :> (Part[
                 CharacterRange["A", "Z"], #]& )}], 
            Cell[
             StyleData["Outline2", "SlideShow"]], 
            Cell[
             StyleData["Outline2", "Printout"], 
             CounterBoxOptions -> {CounterFunction :> (Part[
                 CharacterRange["A", "Z"], #]& )}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline3"], CellMargins -> {{120, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 70}, 
             ParagraphIndent -> -21, CounterIncrements -> "Outline3", 
             CounterAssignments -> {{"Outline4", 0}}, FontSize -> 12, 
             CounterBoxOptions -> {CounterFunction :> Identity}], 
            Cell[
             StyleData["Outline3", "SlideShow"]], 
            Cell[
             StyleData["Outline3", "Printout"], 
             CounterBoxOptions -> {CounterFunction :> Identity}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Outline4"], CellMargins -> {{150, 10}, {7, 7}}, 
             CellGroupingRules -> {"SectionGrouping", 80}, 
             ParagraphIndent -> -18, CounterIncrements -> "Outline4", 
             FontSize -> 10, CounterBoxOptions -> {CounterFunction :> (Part[
                 CharacterRange["a", "z"], #]& )}], 
            Cell[
             StyleData["Outline4", "SlideShow"]], 
            Cell[
             StyleData["Outline4", "Printout"]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["Hyperlink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["Hyperlink", "SlideShow"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
         "The following styles are for linking automatically to the on-line \
help system.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["MainBookLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["MainBookLink", "SlideShow"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["AddOnsLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["AddOnsLink", "SlideShow"]], 
            Cell[
             StyleData["AddOnsLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuide", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["RefGuideLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["RefGuideLink", "SlideShow"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLinkText"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuide", #]}]& )}], 
            Cell[
             StyleData["RefGuideLinkText", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["RefGuideLinkText", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["RefGuideLinkText", "SlideShow"]], 
            Cell[
             StyleData["RefGuideLinkText", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"], FontSize -> 16], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["GettingStartedLink", "SlideShow"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DemosLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["Demos", #]}]& )}], 
            Cell[
             StyleData["DemosLink", "SlideShow"]], 
            Cell[
             StyleData["DemosLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["TourLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["Tour", #]}]& )}], 
            Cell[
             StyleData["TourLink", "SlideShow"]], 
            Cell[
             StyleData["TourLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"], FontSize -> 
             16], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"], FontSize -> 11], 
            Cell[
             StyleData["OtherInformationLink", "SlideShow"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MasterIndexLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> 
             RGBColor[0.269993, 0.308507, 0.6], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MasterIndex", #]}]& )}], 
            Cell[
             StyleData["MasterIndexLink", "SlideShow"]], 
            Cell[
             StyleData["MasterIndexLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          DefaultNewInlineCellStyle -> "None", LanguageCategory -> 
          "NaturalLanguage", StyleMenuListing -> None, FontSize -> 10, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          DefaultNewInlineCellStyle -> "None", LanguageCategory -> 
          "NaturalLanguage", StyleMenuListing -> None, FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Times", FontSize -> 10]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Palette Styles", "Section"], 
         Cell[
         "The cells below define styles that define standard ButtonFunctions, \
for use in palette buttons.", "Text"], 
         Cell[
          StyleData["Paste"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, Placeholder]}]& )}], 
         Cell[
          StyleData["Evaluate"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionEvaluate[
                FrontEnd`InputNotebook[], All]}]& )}], 
         Cell[
          StyleData["EvaluateCell"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionMove[
                FrontEnd`InputNotebook[], All, Cell, 1], 
               FrontEnd`SelectionEvaluateCreateCell[
                FrontEnd`InputNotebook[], All]}]& )}], 
         Cell[
          StyleData["CopyEvaluate"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`SelectionCreateCell[
                FrontEnd`InputNotebook[], All], 
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionEvaluate[
                FrontEnd`InputNotebook[], All]}]& )}], 
         Cell[
          StyleData["CopyEvaluateCell"], StyleMenuListing -> None, 
          ButtonStyleMenuListing -> Automatic, 
          ButtonBoxOptions -> {ButtonFunction :> (FrontEndExecute[{
               FrontEnd`SelectionCreateCell[
                FrontEnd`InputNotebook[], All], 
               FrontEnd`NotebookApply[
                FrontEnd`InputNotebook[], #, All], 
               FrontEnd`SelectionEvaluateCreateCell[
                FrontEnd`InputNotebook[], All]}]& )}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Placeholder Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making placeholder objects \
in palette templates.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Placeholder"], Placeholder -> True, StyleMenuListing -> 
             None, FontSlant -> "Italic", FontColor -> 
             RGBColor[0.890623, 0.864698, 0.384756], 
             TagBoxOptions -> {
              Editable -> False, Selectable -> False, StripWrapperBoxes -> 
               False}], 
            Cell[
             StyleData["Placeholder", "Presentation"]], 
            Cell[
             StyleData["Placeholder", "Condensed"]], 
            Cell[
             StyleData["Placeholder", "SlideShow"]], 
            Cell[
             StyleData["Placeholder", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["PrimaryPlaceholder"], StyleMenuListing -> None, 
             DrawHighlighted -> True, FontSlant -> "Italic", Background -> 
             RGBColor[0.912505, 0.891798, 0.507774], 
             TagBoxOptions -> {
              Editable -> False, Selectable -> False, StripWrapperBoxes -> 
               False}], 
            Cell[
             StyleData["PrimaryPlaceholder", "Presentation"]], 
            Cell[
             StyleData["PrimaryPlaceholder", "Condensed"]], 
            Cell[
             StyleData["PrimaryPlaceholder", "SlideShow"]], 
            Cell[
             StyleData["PrimaryPlaceholder", "Printout"]]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["FormatType Styles", "Section"], 
         Cell[
         "The cells below define styles that are mixed in with the styles of \
most cells.  If a cell's FormatType matches the name of one of the styles \
defined below, then that style is applied between the cell's style and its \
own options. This is particularly true of Input and Output.", "Text"], 
         Cell[
          StyleData["CellExpression"], PageWidth -> Infinity, 
          CellMargins -> {{6, Inherited}, {Inherited, Inherited}}, 
          ShowCellLabel -> False, ShowSpecialCharacters -> False, 
          AllowInlineCells -> False, Hyphenation -> False, 
          AutoItalicWords -> {}, StyleMenuListing -> None, FontFamily -> 
          "Courier", FontSize -> 12, Background -> GrayLevel[1]], 
         Cell[
          StyleData["InputForm"], InputAutoReplacements -> {}, 
          AllowInlineCells -> False, Hyphenation -> False, StyleMenuListing -> 
          None, FontFamily -> "Courier"], 
         Cell[
          StyleData["OutputForm"], PageWidth -> Infinity, TextAlignment -> 
          Left, LineSpacing -> {0.6, 1}, StyleMenuListing -> None, FontFamily -> 
          "Courier"], 
         Cell[
          StyleData["StandardForm"], 
          InputAutoReplacements -> {
           "->" -> "\[Rule]", ":>" -> "\[RuleDelayed]", "<=" -> 
            "\[LessEqual]", ">=" -> "\[GreaterEqual]", "!=" -> "\[NotEqual]", 
            "==" -> "\[Equal]", Inherited}, 
          "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          LineSpacing -> {1.25, 0}, StyleMenuListing -> None, FontFamily -> 
          "Courier"], 
         Cell[
          StyleData["TraditionalForm"], 
          InputAutoReplacements -> {
           "->" -> "\[Rule]", ":>" -> "\[RuleDelayed]", "<=" -> 
            "\[LessEqual]", ">=" -> "\[GreaterEqual]", "!=" -> "\[NotEqual]", 
            "==" -> "\[Equal]", Inherited}, 
          "TwoByteSyntaxCharacterAutoReplacement" -> True, 
          LineSpacing -> {1.25, 0}, SingleLetterItalics -> True, 
          TraditionalFunctionNotation -> True, DelimiterMatching -> None, 
          StyleMenuListing -> None], 
         Cell[
         "The style defined below is mixed in to any cell that is in an \
inline cell within another.", "Text"], 
         Cell[
          StyleData["InlineCell"], LanguageCategory -> "Formula", ScriptLevel -> 
          1, StyleMenuListing -> None], 
         Cell[
          StyleData["InlineCellEditing"], StyleMenuListing -> None, 
          Background -> RGBColor[0.964706, 0.929412, 0.839216]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Automatic Styles", "Section"], 
         Cell[
         "The cells below define styles that are used to affect the display \
of certain types of objects in typeset expressions.  For example, \
\"UnmatchedBracket\" style defines how unmatched bracket, curly bracket, and \
parenthesis characters are displayed (typically by coloring them to make them \
stand out).", "Text"], 
         Cell[
          StyleData["UnmatchedBracket"], StyleMenuListing -> None, FontColor -> 
          RGBColor[0.760006, 0.330007, 0.8]], 
         Cell[
          StyleData["Completions"], StyleMenuListing -> None, FontFamily -> 
          "Courier"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles from HelpBrowser", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MathCaption"], CellFrame -> {{0, 0}, {0, 0.5}}, 
             CellMargins -> {{66, 12}, {2, 24}}, PageBreakBelow -> False, 
             CellFrameMargins -> {{8, 8}, {8, 2}}, CellFrameColor -> 
             GrayLevel[0.700008], CellFrameLabelMargins -> 4, 
             LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 8}, 
             StyleMenuListing -> None, FontColor -> GrayLevel[0.2]], 
            Cell[
             StyleData["MathCaption", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["MathCaption", "SlideShow"]], 
            Cell[
             StyleData["MathCaption", "Printout"], 
             CellMargins -> {{39, 0}, {0, 14}}, Hyphenation -> True, FontSize -> 
             9, FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ObjectName"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["ObjectName", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["ObjectName", "SlideShow"]], 
            Cell[
             StyleData["ObjectName", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Usage"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["Usage", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["Usage", "SlideShow"]], 
            Cell[
             StyleData["Usage", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Notes"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["Notes", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["Notes", "SlideShow"]], 
            Cell[
             StyleData["Notes", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["InlineOutput"], ShowCellBracket -> True, 
             CellMargins -> {{66, 4}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, CellLabelAutoDelete -> False, 
             CellLabelMargins -> {{14, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, 
             ShowSpecialCharacters -> Automatic, 
             "TwoByteSyntaxCharacterAutoReplacement" -> True, 
             HyphenationOptions -> {
              "HyphenationCharacter" -> "\[Continuation]"}, LanguageCategory -> 
             "Mathematica", FormatType -> StandardForm, ShowStringCharacters -> 
             True, NumberMarks -> True, StyleMenuListing -> None, FontWeight -> 
             "Bold"], 
            Cell[
             StyleData["InlineOutput", "Presentation"], FontSize -> 18], 
            Cell[
             StyleData["InlineOutput", "SlideShow"]], 
            Cell[
             StyleData["InlineOutput", "Printout"], ShowCellBracket -> False, 
             CellMargins -> {{39, 0}, {6, 6}}, FontSize -> 9]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell["Emphasis Boxes and Pictures", "Subsection"], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["Box"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["Box", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["Box", "SlideShow"]], 
               Cell[
                StyleData["Box", "Printout"], CellMargins -> {{2, 0}, {0, 8}},
                 FontSize -> 10, Background -> GrayLevel[0.900008]]}, 
              Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DoubleBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Top}}, "RowsIndexed" -> {}}, 
                  GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[1.4]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DoubleBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["DoubleBox", "SlideShow"]], 
               Cell[
                StyleData["DoubleBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["1ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["1ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["1ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["1ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 SingleLetterItalics -> False, LineIndent -> 0, 
                StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.67]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["2ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnEvenBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.46]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnEvenBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnEvenBox", "SlideShow"]], 
               Cell[
                StyleData["2ColumnEvenBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnSmallBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {Right, {Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.35]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[1.0499999999999998`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnSmallBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnSmallBox", "SlideShow"]], 
               Cell[
                StyleData["2ColumnSmallBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["3ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.32]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["3ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["3ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["3ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["3ColumnSmallBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {Right, Center, {Left}}, 
                    "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
                    "RowsIndexed" -> {}}, GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.24]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[1.0499999999999998`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["3ColumnSmallBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["3ColumnSmallBox", "SlideShow"]], 
               Cell[
                StyleData["3ColumnSmallBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["4ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 SingleLetterItalics -> False, LineIndent -> 0, 
                StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.13], 
                    Scaled[0.35], 
                    Scaled[0.13], {
                    Scaled[0.35]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["4ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["4ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["4ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["5ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.202]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}},
                     "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["5ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["5ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["5ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["6ColumnBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {BoxFrame -> 0.5, FrameMargins -> True}, 
                GridBoxOptions -> {GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.12], 
                    Scaled[0.22], 
                    Scaled[0.12], 
                    Scaled[0.12], 
                    Scaled[0.22], {
                    Scaled[0.12]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["6ColumnBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["6ColumnBox", "SlideShow"]], 
               Cell[
                StyleData["6ColumnBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["FramedBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.5599999999999999]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["FramedBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["FramedBox", "SlideShow"]], 
               Cell[
                StyleData["FramedBox", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.4], {
                    Scaled[0.6]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["DefinitionBox", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox3Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.35], 
                    Scaled[0.2], {
                    Scaled[0.45]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox3Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox3Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox3Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox4Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.15], 
                    Scaled[0.35], 
                    Scaled[0.15], {
                    Scaled[0.35]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox4Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox4Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox4Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox5Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.2]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox5Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox5Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox5Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["DefinitionBox6Col"], CellFrame -> 0.5, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakWithin -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.13], 
                    Scaled[0.24], 
                    Scaled[0.13], 
                    Scaled[0.13], 
                    Scaled[0.24], {
                    Scaled[0.13]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["DefinitionBox6Col", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["DefinitionBox6Col", "SlideShow"]], 
               Cell[
                StyleData["DefinitionBox6Col", "Printout"], 
                CellMargins -> {{2, 4}, {0, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["TopBox"], CellFrame -> {{0.5, 0.5}, {0, 0.5}}, 
                CellMargins -> {{27, 12}, {0, 8}}, CellHorizontalScrolling -> 
                True, PageBreakBelow -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.62]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["TopBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["TopBox", "SlideShow"]], 
               Cell[
                StyleData["TopBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, Background -> GrayLevel[1]]},
               Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["MiddleBox"], CellFrame -> {{0.5, 0.5}, {0, 0}}, 
                CellMargins -> {{27, 12}, {0, -7}}, CellHorizontalScrolling -> 
                True, PageBreakAbove -> False, PageBreakBelow -> False, 
                CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941], 
                AutoIndent -> False, AutoSpacing -> False, LineIndent -> 0, 
                StyleMenuListing -> None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.62]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["MiddleBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["MiddleBox", "SlideShow"]], 
               Cell[
                StyleData["MiddleBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, 2}}, Background -> GrayLevel[1]]},
               Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["BottomBox"], CellFrame -> {{0.5, 0.5}, {0.5, 0}}, 
                CellMargins -> {{27, 12}, {0, -7}}, CellHorizontalScrolling -> 
                True, PageBreakAbove -> False, PageBreakBelow -> False, 
                CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941], 
                AutoIndent -> False, AutoSpacing -> False, LineIndent -> 0, 
                StyleMenuListing -> None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.31], {
                    Scaled[0.62]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["BottomBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["BottomBox", "SlideShow"]], 
               Cell[
                StyleData["BottomBox", "Printout"], 
                CellMargins -> {{2, 0}, {0, -5}}, FontSize -> 10, Background -> 
                GrayLevel[1], 
                GridBoxOptions -> {
                 GridBoxItemSize -> {
                   "Columns" -> {{All}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{2.2}}, "RowsIndexed" -> {}}}]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["TopSpanBox"], CellFrame -> {{0.5, 0.5}, {0, 0.5}}, 
                CellMargins -> {{27, 12}, {-2, 8}}, CellHorizontalScrolling -> 
                True, PageBreakBelow -> False, CellFrameColor -> 
                RGBColor[0.74902, 0.694118, 0.552941], AutoIndent -> False, 
                AutoSpacing -> False, LineIndent -> 0, StyleMenuListing -> 
                None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.9], {
                    Scaled[0.03]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["TopSpanBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["TopSpanBox", "SlideShow"]], 
               Cell[
                StyleData["TopSpanBox", "Printout"], 
                CellMargins -> {{2, 0}, {-2, 8}}, FontSize -> 10, Background -> 
                GrayLevel[1]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["MiddleSpanBox"], CellFrame -> {{0.5, 0.5}, {0, 0}},
                 CellMargins -> {{27, 12}, {0, 0}}, CellHorizontalScrolling -> 
                True, PageBreakAbove -> False, PageBreakBelow -> False, 
                CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941], 
                AutoIndent -> False, AutoSpacing -> False, LineIndent -> 0, 
                StyleMenuListing -> None, FontWeight -> "Plain", Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.9], {
                    Scaled[0.03]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.6]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["MiddleSpanBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["MiddleSpanBox", "SlideShow"]], 
               Cell[
                StyleData["MiddleSpanBox", "Printout"], 
                CellMargins -> {{2, 0}, {-5, 0}}, FontSize -> 10, Background -> 
                GrayLevel[1], 
                GridBoxOptions -> {
                 GridBoxItemSize -> {
                   "Columns" -> {{All}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{1.8}}, "RowsIndexed" -> {}}}]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["Picture"], 
                CellMargins -> {{27, Inherited}, {4, 4}}, CellGroupingRules -> 
                "GraphicsGrouping", CellHorizontalScrolling -> True, 
                StyleMenuListing -> None], 
               Cell[
                StyleData["Picture", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["Picture", "SlideShow"]], 
               Cell[
                StyleData["Picture", "Printout"], 
                CellMargins -> {{2, Inherited}, {4, 4}}, Magnification -> 
                0.65]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["OpenCloseItemizedPicture"], 
                CellMargins -> {{88, 4}, {4, 4}}, 
                PrivateCellOptions -> {"DefaultCellGroupOpen" -> False}, 
                CellGroupingRules -> "GraphicsGrouping", 
                CellHorizontalScrolling -> True, StyleMenuListing -> None], 
               Cell[
                StyleData["OpenCloseItemizedPicture", "Presentation"], 
                FontSize -> 18], 
               Cell[
                StyleData["OpenCloseItemizedPicture", "SlideShow"]], 
               Cell[
                StyleData["OpenCloseItemizedPicture", "Printout"], 
                CellMargins -> {{76, 2}, {0, 0}}, 
                CellElementSpacings -> {
                 "CellMinHeight" -> 1, "ClosedCellHeight" -> 0}, CellOpen -> 
                False]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["ItemizedPicture"], 
                CellMargins -> {{88, 4}, {4, 4}}, CellGroupingRules -> 
                "GraphicsGrouping", CellHorizontalScrolling -> True, 
                StyleMenuListing -> None], 
               Cell[
                StyleData["ItemizedPicture", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["ItemizedPicture", "SlideShow"]], 
               Cell[
                StyleData["ItemizedPicture", "Printout"], 
                CellMargins -> {{77, 2}, {4, -4}}, Magnification -> 0.5]}, 
              Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["ListGraphic"], CellMargins -> {{88, 4}, {4, 4}}, 
                CellGroupingRules -> "GraphicsGrouping", 
                CellHorizontalScrolling -> True, StyleMenuListing -> None], 
               Cell[
                StyleData["ListGraphic", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["ListGraphic", "SlideShow"]], 
               Cell[
                StyleData["ListGraphic", "Printout"], 
                CellMargins -> {{77, 2}, {4, -4}}, Magnification -> 0.5]}, 
              Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["ListNoteBox"], CellFrame -> 0.5, 
                CellMargins -> {{88, 12}, {8, 8}}, CellHorizontalScrolling -> 
                True, CellFrameColor -> RGBColor[0.74902, 0.694118, 0.552941],
                 LineIndent -> 0, StyleMenuListing -> None, Background -> 
                RGBColor[0.964706, 0.929412, 0.839216], 
                FrameBoxOptions -> {FrameMargins -> {{1, 1}, {1.5, 1.5}}}, 
                GridBoxOptions -> {GridBoxSpacings -> {"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.7]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["ListNoteBox", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["ListNoteBox", "SlideShow"]], 
               Cell[
                StyleData["ListNoteBox", "Printout"], 
                CellMargins -> {{77, 4}, {6, 2}}, FontSize -> 10, Background -> 
                GrayLevel[0.900008]]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["PictureGroup"], CellMargins -> {{41, 4}, {0, 4}}, 
                CellGroupingRules -> "GraphicsGrouping", 
                CellHorizontalScrolling -> True, StyleMenuListing -> None], 
               Cell[
                StyleData["PictureGroup", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["PictureGroup", "SlideShow"]], 
               Cell[
                StyleData["PictureGroup", "Printout"], 
                CellMargins -> {{76, 2}, {0, 0}}, 
                CellElementSpacings -> {
                 "CellMinHeight" -> 1, "ClosedCellHeight" -> 0}, CellOpen -> 
                False]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["Sound"], ShowCellBracket -> True, 
                CellMargins -> {{27, Inherited}, {0, 8}}, StyleMenuListing -> 
                None], 
               Cell[
                StyleData["Sound", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["Sound", "SlideShow"]], 
               Cell[
                StyleData["Sound", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 10]}, Closed]]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell["Tables", "Subsection"], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnTable"], CellMargins -> {{35, 4}, {0, 8}}, 
                CellHorizontalScrolling -> True, LineIndent -> 0, 
                StyleMenuListing -> None, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.34], {
                    Scaled[0.64]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnTable", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["2ColumnTable", "SlideShow"]], 
               Cell[
                StyleData["2ColumnTable", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["2ColumnEvenTable"], 
                CellMargins -> {{35, 4}, {0, 8}}, CellHorizontalScrolling -> 
                True, LineIndent -> 0, StyleMenuListing -> None, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {{
                    Scaled[0.49]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["2ColumnEvenTable", "Presentation"], FontSize -> 
                18], 
               Cell[
                StyleData["2ColumnEvenTable", "SlideShow"]], 
               Cell[
                StyleData["2ColumnEvenTable", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9]}, Closed]], 
            Cell[
             CellGroupData[{
               Cell[
                StyleData["3ColumnTable"], CellMargins -> {{35, 4}, {0, 8}}, 
                CellHorizontalScrolling -> True, LineIndent -> 0, 
                StyleMenuListing -> None, 
                GridBoxOptions -> {
                 GridBoxAlignment -> {
                   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}, 
                  GridBoxItemSize -> {"Columns" -> {
                    Scaled[0.28], 
                    Scaled[0.28], {
                    Scaled[0.43]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                    "RowsIndexed" -> {}}}], 
               Cell[
                StyleData["3ColumnTable", "Presentation"], FontSize -> 18], 
               Cell[
                StyleData["3ColumnTable", "SlideShow"]], 
               Cell[
                StyleData["3ColumnTable", "Printout"], 
                CellMargins -> {{2, 0}, {0, 8}}, FontSize -> 9]}, Closed]]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Slide Show Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideShowNavigationBar"], Editable -> False, 
             CellFrame -> True, CellMargins -> {{0, 0}, {3, 3}}, 
             CellElementSpacings -> {"CellMinHeight" -> 0.8125}, 
             CellGroupingRules -> {"SectionGrouping", 30}, CellFrameMargins -> 
             False, CellFrameColor -> GrayLevel[1], CellFrameLabelMargins -> 
             False, TextAlignment -> Center, CounterIncrements -> 
             "SlideShowNavigationBar", StyleMenuListing -> None, FontSize -> 
             10, Magnification -> 1, Background -> GrayLevel[0.8], 
             GridBoxOptions -> {
              BaselinePosition -> Center, 
               GridBoxAlignment -> {
                "Columns" -> {
                  Center, Center, Center, Center, Center, Center, Right, {
                   Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
                 "RowsIndexed" -> {}}, 
               GridBoxItemSize -> {
                "Columns" -> {3.5, 3.5, 3.5, 3.5, 13, 5, {4}}, 
                 "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
                 "RowsIndexed" -> {}}, GridBoxSpacings -> {"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}}], 
            Cell[
             StyleData["SlideShowNavigationBar", "Presentation"]], 
            Cell[
             StyleData["SlideShowNavigationBar", "SlideShow"], Deletable -> 
             False, ShowCellBracket -> False, 
             CellMargins -> {{-1, -1}, {-1, -1}}, PageBreakAbove -> True, 
             CellFrameMargins -> {{1, 1}, {0, 0}}], 
            Cell[
             StyleData["SlideShowNavigationBar", "Printout"], 
             CellMargins -> {{18, 4}, {4, 4}}, LineSpacing -> {1, 3}, 
             FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideShowSection"], CellFrame -> {{0, 0}, {0, 0.5}}, 
             CellMargins -> {{0, 0}, {10, 0}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CellFrameMargins -> {{12, 4}, {6, 12}}, 
             InputAutoReplacements -> {"TeX" -> StyleBox[
                 RowBox[{"T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "LaTeX" -> StyleBox[
                 RowBox[{"L", 
                   StyleBox[
                    AdjustmentBox[
                    "A", BoxMargins -> {{-0.36, -0.1}, {0, 0}}, 
                    BoxBaselineShift -> -0.2], FontSize -> Smaller], "T", 
                   AdjustmentBox[
                   "E", BoxMargins -> {{-0.075, -0.085}, {0, 0}}, 
                    BoxBaselineShift -> 0.5], "X"}]], "mma" -> "Mathematica", 
               "Mma" -> "Mathematica", "MMA" -> "Mathematica", 
               "gridMathematica" -> FormBox[
                 RowBox[{"grid", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               "webMathematica" -> FormBox[
                 RowBox[{"web", 
                   AdjustmentBox[
                    StyleBox["Mathematica", FontSlant -> "Italic"], 
                    BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
               Inherited}, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 
             18, FontWeight -> "Plain", FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.408011, 0.440726, 0.8]], 
            Cell[
             StyleData["SlideShowSection", "Presentation"], 
             CellFrameMargins -> {{20, 10}, {10, 18}}, FontSize -> 27], 
            Cell[
             StyleData["SlideShowSection", "SlideShow"], ShowCellBracket -> 
             False, PageBreakAbove -> True], 
            Cell[
             StyleData["SlideShowSection", "Printout"], 
             CellMargins -> {{18, 30}, {0, 30}}, CellFrameMargins -> 5, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideHyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontSize -> 26, FontColor -> 
             GrayLevel[0.400015], 
             ButtonBoxOptions -> {
              Active -> True, ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonMargins -> 0.5, 
               ButtonMinHeight -> 0.85, ButtonNote -> None}], 
            Cell[
             StyleData["SlideHyperlink", "Presentation"], 
             CellMargins -> {{14, 10}, {6, 12}}, FontSize -> 36], 
            Cell[
             StyleData["SlideHyperlink", "SlideShow"]], 
            Cell[
             StyleData["SlideHyperlink", "Printout"], FontSize -> 10, 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideTOCLink"], 
             CellMargins -> {{24, Inherited}, {Inherited, Inherited}}, 
             StyleMenuListing -> None, ButtonStyleMenuListing -> Automatic, 
             FontFamily -> "Helvetica", 
             ButtonBoxOptions -> {
              Active -> True, ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonMargins -> 1.5, 
               ButtonNote -> ButtonData}], 
            Cell[
             StyleData["SlideTOCLink", "Presentation"], 
             CellMargins -> {{35, 10}, {8, 8}}, FontSize -> 18], 
            Cell[
             StyleData["SlideTOCLink", "SlideShow"]], 
            Cell[
             StyleData["SlideTOCLink", "Printout"], 
             FontVariations -> {"Underline" -> False}, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SlideTOC"], CellDingbat -> "\[Bullet]", 
             CellMargins -> {{18, Inherited}, {Inherited, Inherited}}, 
             StyleMenuListing -> None, FontFamily -> "Helvetica"], 
            Cell[
             StyleData["SlideTOC", "Presentation"], 
             CellMargins -> {{25, 10}, {10, 5}}, FontSize -> 18], 
            Cell[
             StyleData["SlideTOC", "SlideShow"], FontSize -> 14], 
            Cell[
             StyleData["SlideTOC", "Printout"], FontSize -> 10, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
      "Notation Package Styles", "Section", CellTags -> "NotationPackage"], 
      Cell[
      "The cells below define certain styles needed by the Notation package. \
These styles serve to make visible otherwise invisible tagboxes.", "Text", 
       CellTags -> "NotationPackage"], 
      Cell[
       StyleData["NotationTemplateStyle"], StyleMenuListing -> None, 
       Background -> RGBColor[1, 1, 0.850004], 
       TagBoxOptions -> {SyntaxForm -> "symbol"}, CellTags -> 
       "NotationPackage"], 
      Cell[
       StyleData["NotationPatternWrapperStyle"], StyleMenuListing -> None, 
       Background -> RGBColor[1, 0.900008, 0.979995], 
       TagBoxOptions -> {SyntaxForm -> "symbol"}, CellTags -> 
       "NotationPackage"], 
      Cell[
       StyleData["NotationMadeBoxesWrapperStyle"], StyleMenuListing -> None, 
       Background -> RGBColor[0.900008, 0.889998, 1], 
       TagBoxOptions -> {SyntaxForm -> "symbol"}, CellTags -> 
       "NotationPackage"]}, Closed]]}, Visible -> False, FrontEndVersion -> 
  "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 78, 1, 49, "Subtitle",
 InitializationCell->True],
Cell[648, 25, 300, 6, 47, "Text",
 InitializationCell->True],
Cell[951, 33, 364, 10, 28, "Input",
 InitializationCell->True],
Cell[1318, 45, 388, 8, 49, "Input",
 InitializationCell->True],
Cell[1709, 55, 300, 7, 28, "Input",
 InitializationCell->True],
Cell[2012, 64, 271, 7, 28, "Input",
 InitializationCell->True],
Cell[2286, 73, 243, 6, 28, "Input",
 InitializationCell->True],
Cell[2532, 81, 631, 13, 49, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[3188, 98, 101, 2, 28, "Input",
 InitializationCell->True],
Cell[3292, 102, 160, 2, 27, "Output"]
}, Open  ]],
Cell[3467, 107, 141, 4, 28, "Input",
 InitializationCell->True],
Cell[3611, 113, 13935, 316, 1289, "Input",
 InitializationCell->True],
Cell[17549, 431, 265, 6, 28, "Input",
 InitializationCell->True],
Cell[17817, 439, 924, 24, 129, "Input",
 InitializationCell->True],
Cell[18744, 465, 753, 25, 28, "Input",
 InitializationCell->True],
Cell[19500, 492, 3690, 86, 189, "Input",
 InitializationCell->True],
Cell[23193, 580, 196, 6, 28, "Input",
 InitializationCell->True],
Cell[23392, 588, 462, 13, 28, "Input",
 InitializationCell->True],
Cell[23857, 603, 2159, 63, 169, "Input",
 InitializationCell->True],
Cell[26019, 668, 575, 16, 49, "Input",
 InitializationCell->True],
Cell[26597, 686, 736, 22, 49, "Input",
 InitializationCell->True],
Cell[27336, 710, 133, 4, 28, "Input",
 InitializationCell->True],
Cell[27472, 716, 418, 14, 28, "Input",
 InitializationCell->True],
Cell[27893, 732, 3009, 79, 269, "Input",
 InitializationCell->True],
Cell[30905, 813, 10723, 252, 890, "Input",
 InitializationCell->True],
Cell[41631, 1067, 1212, 35, 139, "Input",
 InitializationCell->True],
Cell[42846, 1104, 1057, 36, 30, "Input",
 InitializationCell->True],
Cell[43906, 1142, 589, 20, 30, "Input",
 InitializationCell->True],
Cell[44498, 1164, 385, 12, 28, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[44908, 1180, 78, 2, 28, "Input",
 InitializationCell->True],
Cell[44989, 1184, 160, 2, 27, "Output"]
}, Open  ]],
Cell[45164, 1189, 85, 2, 28, "Input",
 InitializationCell->True]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
